Package: readabs
Type: Package
Title: Download and Tidy Time Series Data from the Australian Bureau of
        Statistics
Version: 0.4.30
Authors@R: c(
           person("Matt", "Cowgill", role = c("aut", "cre"), email = "mattcowgill@gmail.com", comment = c(ORCID = "0000-0003-0422-3300")),
           person("Zoe", "Meers", role = "aut", email = "zoe.meers@sydney.edu.au"),
           person("Jaron", "Lee", role = "aut", email = "jaron.lee@sydney.edu.au"),
           person("David", "Diviny", role = "aut", email = "David.Diviny@nousgroup.com.au"),
           person("Hugh", "Parsonage", role = "ctb", email = "hugh.parsonage@gmail.com"),
           person("Kinto", "Behr", role = "ctb", email = "kinto.behr@gmail.com"),
           person("Angus", "Moore", role = "ctb"),
           person("Francis", "Markham", role = "ctb", comment = c(ORCID = "0000-0002-4266-2569"))
           )
Maintainer: Matt Cowgill <mattcowgill@gmail.com>
Description: Downloads, imports, and tidies time series data from the 
    Australian Bureau of Statistics <https://www.abs.gov.au/>.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 3.5)
Imports: readxl (>= 1.2.0), dplyr (>= 0.8.0), hutils (>= 1.5.0), fst,
        purrr (>= 1.0.0), tidyr (>= 1.0.0), stringi, tools, glue, httr,
        rvest, xml2, rlang, labelled
URL: https://github.com/mattcowgill/readabs
BugReports: https://github.com/mattcowgill/readabs/issues
VignetteBuilder: knitr
Suggests: knitr, rmarkdown, markdown, testthat (>= 2.1.0), ggplot2
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-17 09:59:27 UTC; mcowgill
Author: Matt Cowgill [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-0422-3300>),
  Zoe Meers [aut],
  Jaron Lee [aut],
  David Diviny [aut],
  Hugh Parsonage [ctb],
  Kinto Behr [ctb],
  Angus Moore [ctb],
  Francis Markham [ctb] (ORCID: <https://orcid.org/0000-0002-4266-2569>)
Repository: CRAN
Date/Publication: 2026-05-17 13:10:02 UTC
Built: R 4.5.2; ; 2026-05-17 13:58:08 UTC; unix

# Generated by roxygen2: do not edit by hand

export(check_latest_date)
export(download_abs_data_cube)
export(extract_abs_sheets)
export(get_available_files)
export(get_available_lfs_cubes)
export(read_abs)
export(read_abs_data)
export(read_abs_local)
export(read_abs_metadata)
export(read_abs_series)
export(read_abs_url)
export(read_api)
export(read_api_dataflows)
export(read_api_datastructure)
export(read_api_url)
export(read_awe)
export(read_cpi)
export(read_erp)
export(read_job_mobility)
export(read_lfs_datacube)
export(read_lfs_grossflows)
export(read_payrolls)
export(search_catalogues)
export(search_files)
export(separate_series)
export(show_available_catalogues)
export(show_available_files)
export(tidy_abs)
export(tidy_abs_list)
import(dplyr)
importFrom(dplyr,"%>%")
importFrom(dplyr,arrange)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,mutate)
importFrom(dplyr,mutate_at)
importFrom(dplyr,pull)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,summarise)
importFrom(dplyr,sym)
importFrom(dplyr,tibble)
importFrom(glue,glue)
importFrom(httr,GET)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,map_chr)
importFrom(purrr,map_dfr)
importFrom(purrr,set_names)
importFrom(purrr,walk)
importFrom(purrr,walk2)
importFrom(readxl,excel_sheets)
importFrom(readxl,read_excel)
importFrom(rlang,.data)
importFrom(rlang,.env)
importFrom(rvest,html_attr)
importFrom(rvest,html_nodes)
importFrom(rvest,html_text)
importFrom(stringi,stri_trim_both)
importFrom(tidyr,gather)
importFrom(tidyr,pivot_longer)
importFrom(tidyr,separate)
importFrom(tools,file_path_sans_ext)
importFrom(tools,toTitleCase)
importFrom(xml2,read_html)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# readabs <img src="man/figures/logo.png" align="right" height="139" />

<!-- badges: start -->

[![R build
status](https://github.com/mattcowgill/readabs/workflows/R-CMD-check/badge.svg)](https://github.com/mattcowgill/readabs/actions)
[![codecov
status](https://img.shields.io/codecov/c/github/mattcowgill/readabs.svg)](https://app.codecov.io/gh/MattCowgill/readabs)
[![CRAN
status](https://www.r-pkg.org/badges/version/readabs)](https://cran.r-project.org/package=readabs)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)
<!-- badges: end -->

## Overview

{readabs} helps you easily download, import, and tidy data from the
Australian Bureau of Statistics within R. This saves you time manually
downloading and tediously tidying data and allows you to spend more time
on your analysis.

## Installing {readabs}

Install the latest CRAN version of {readabs} with:

``` r
install.packages("readabs")
```

You can install the development version of {readabs} from GitHub with:

``` r
# if you don't have devtools installed, first run:
# install.packages("devtools")
devtools::install_github("mattcowgill/readabs")
```

## Using {readabs}

The ABS releases data in many different formats, through many different
dissemination channels.

The {readabs} contains functions for working with three different types
of ABS data:

- `read_abs()` and related functions downloads, imports, and tidies ABS
  time series data.
- `download_abs_data_cube()` and related functions find and download ABS
  data cubes, which are spreadsheets on the ABS website that are not in
  the standard time series format.
- `read_api()` and related functions find, filter, and import data from
  the [ABS.Stat](https://dataexplorer.abs.gov.au) API.

### ABS time series data

A key function in {readabs} is `read_abs()`, which downloads, imports,
and tidies time series data from the ABS website. **Note that
`read_abs()` only works with spreadsheets in the standard ABS time
series format.**

First we’ll load {readabs} and the {tidyverse}:

``` r
library(readabs)
library(tidyverse)
```

Now we’ll create one data frame that contains all the time series data
from the Wage Price Index, catalogue number 6345.0:

``` r
all_wpi <- read_abs("6345.0")
#> Finding URLs for tables corresponding to ABS catalogue 6345.0
#> Attempting to download files from catalogue 6345.0, Wage Price Index, Australia
#> Extracting data from downloaded spreadsheets
#> Tidying data from imported ABS spreadsheets
```

This is what it looks like:

``` r
str(all_wpi)
#> tibble [68,137 × 12] (S3: tbl_df/tbl/data.frame)
#>  $ table_no        : chr [1:68137] "634501" "634501" "634501" "634501" ...
#>  $ sheet_no        : chr [1:68137] "Data1" "Data1" "Data1" "Data1" ...
#>  $ table_title     : chr [1:68137] "Table 1. Total Hourly Rates of Pay Excluding Bonuses: Sector, Original, Seasonally Adjusted and Trend" "Table 1. Total Hourly Rates of Pay Excluding Bonuses: Sector, Original, Seasonally Adjusted and Trend" "Table 1. Total Hourly Rates of Pay Excluding Bonuses: Sector, Original, Seasonally Adjusted and Trend" "Table 1. Total Hourly Rates of Pay Excluding Bonuses: Sector, Original, Seasonally Adjusted and Trend" ...
#>  $ date            : Date[1:68137], format: "1997-09-01" "1997-09-01" ...
#>  $ series          : chr [1:68137] "Quarterly Index ;  Total hourly rates of pay excluding bonuses ;  Australia ;  Private ;  All industries ;" "Quarterly Index ;  Total hourly rates of pay excluding bonuses ;  Australia ;  Public ;  All industries ;" "Quarterly Index ;  Total hourly rates of pay excluding bonuses ;  Australia ;  Private and Public ;  All industries ;" "Quarterly Index ;  Total hourly rates of pay excluding bonuses ;  Australia ;  Private ;  All industries ;" ...
#>  $ value           : num [1:68137] 67.4 64.7 66.7 67.3 64.8 66.6 67.3 64.8 66.7 NA ...
#>  $ series_type     : chr [1:68137] "Original" "Original" "Original" "Seasonally Adjusted" ...
#>  $ data_type       : chr [1:68137] "INDEX" "INDEX" "INDEX" "INDEX" ...
#>  $ collection_month: chr [1:68137] "3" "3" "3" "3" ...
#>  $ frequency       : chr [1:68137] "Quarter" "Quarter" "Quarter" "Quarter" ...
#>  $ series_id       : chr [1:68137] "A2603039T" "A2603989W" "A2603609J" "A2713846W" ...
#>  $ unit            : chr [1:68137] "Index Numbers" "Index Numbers" "Index Numbers" "Index Numbers" ...
```

It only takes you a few lines of code to make a graph from your data:

``` r
all_wpi %>%
  filter(
    series == "Percentage Change From Corresponding Quarter of Previous Year ;  Australia ;  Total hourly rates of pay excluding bonuses ;  Private and Public ;  All industries ;",
    !is.na(value)
  ) %>%
  ggplot(aes(x = date, y = value, col = series_type)) +
  geom_line() +
  theme_minimal() +
  labs(y = "Annual wage growth (per cent)")
```

<img src="man/figures/README-all-in-one-example-1.png" width="672" />

In the example above we downloaded all the time series from a catalogue
number. This will often be overkill. If you know the data you need is in
a particular table, you can just get that table like this:

``` r
wpi_t1 <- read_abs("6345.0", tables = 1)
#> Warning in read_abs("6345.0", tables = 1): `tables` was providedyet
#> `check_local = TRUE` and fst files are available so `tables` will be ignored.
```

If you want multiple tables, but not the whole catalogue, that’s easy
too:

``` r
wpi_t1_t5 <- read_abs("6345.0", tables = c("1", "5a"))
#> Warning in read_abs("6345.0", tables = c("1", "5a")): `tables` was providedyet
#> `check_local = TRUE` and fst files are available so `tables` will be ignored.
```

For more examples, please see the vignette on working with time series
data (run `browseVignettes("readabs")`).

Some other functions that may come in handy when working with ABS time
series data:

- `read_abs_local()` imports and tidies time series data from ABS
  spreadsheets stored on a local drive. Thanks to Hugh Parsonage for
  contributing to this functionality.
- `separate_series()` splits the `series` column of a tidied ABS time
  series spreadsheet into multiple columns, reducing the manual
  wrangling that’s needed to work with the data. Thanks to David Diviny
  for writing this function.

#### Convenience functions for loading time series data

There are several functions that load specific ABS time series data:

- `read_cpi()` imports the Consumer Price Index numbers as a two-column
  tibble: `date` and `cpi`. This is useful for joining to other series
  to adjust data for changes in consumer prices.
- `read_awe()` returns a long time series of Average Weekly Earnings
  data.
- `read_job_mobility()` downloads, imports and tidies tables from the
  ABS Job Mobility dataset.

### ABS data cubes

The ABS (generally) releases time series data in a standard format,
which allows `read_abs()` to download, import and tidy it (see above).
But not all ABS data is time series data - the ABS also releases data as
‘data cubes’. These are all formatted in their own, unique way.

Unfortunately, because data cubes are all formatted in their own way,
there is no one function that can import tidy data cubes for you in the
same way that `read_abs()` works with all time series. But `{readabs}`
still has functions that can help. Thanks to David Diviny for writing
these functions.

The `download_abs_data_cube()` function can download an ABS data cube
for you. It works with any data cube on the ABS website. To use this
function, we need two things: a `catalogue_string` (the short name of
the release) and `cube`, a (unique fragment of) the filename within the
catalogue you wish to download.

For example, let’s say you wanted to download table 4 from *Weekly
Payroll Jobs and Wages in Australia*. We can find the catalogue name
like this:

``` r
search_catalogues("payroll")
#> # A tibble: 2 × 4
#>   heading sub_heading                                         catalogue    url  
#>   <chr>   <chr>                                               <chr>        <chr>
#> 1 Jobs    Weekly Payroll Jobs                                 weekly-payr… http…
#> 2 Jobs    Weekly Payroll Jobs and Wages in Australia, Interim weekly-payr… http…
```

Now we know that the string `"weekly-payroll-jobs"` is the
`catalogue_string` for this release. We can now see what files are
available to download from this catalogue:

``` r
show_available_files("weekly-payroll-jobs")
#> # A tibble: 7 × 3
#>   label                                                              file  url  
#>   <chr>                                                              <chr> <chr>
#> 1 Table 20: Payroll jobs - characteristics distributionsContains se… 6160… http…
#> 2 Table 4: Payroll jobs indexes                                      6160… http…
#> 3 Table 6: Industry subdivision - Payroll jobs indexes               6160… http…
#> 4 Table 7: Employer characteristics - Payroll jobs index             6160… http…
#> 5 Table 8: Jobholder characteristics - Payroll jobs index            6160… http…
#> 6 Table 9: Sector - Payroll jobs index                               6160… http…
#> 7 All data cubes                                                     6160… http…
```

We want Table 4, which has the filename `6160055001_DO004.xlsx`.

We can download the file as follows:

``` r
payrolls_t4_path <- download_abs_data_cube("weekly-payroll-jobs", "004")
#> File downloaded in /tmp/RtmpV2bAuK/6160055001_DO004.xlsx

payrolls_t4_path
#> [1] "/tmp/RtmpV2bAuK/6160055001_DO004.xlsx"
```

The `download_abs_data_cube()` function downloads the file and returns
the full file path to the saved file. You can then pipe that in to
another function:

``` r
payrolls_t4_path %>%
  readxl::read_excel(
    sheet = "Payroll jobs index",
    skip = 5
  )
```

#### Convenience functions for data cubes

As it happens, if you want the ABS Weekly Payrolls data, you don’t need
to use `download_abs_data_cube()` directly. Instead, there is a
convenience function available that downloads, imports, and tidies the
data for you:

``` r
read_payrolls()
```

There is also a convenience function available for data cube GM1 from
the monthly Labour Force data, which contains labour force gross flows:

``` r
read_lfs_grossflows()
```

### Finding and loading data from the ABS.Stat API

The ABS has created a new site to access its data, called the ABS Data
Explorer, also known as ABS.Stat. As at early 2023, this site is in Beta
mode. The site provides an API.

The {readabs} package includes functions to query the ABS.Stat API.
Thank you to Kinto Behr for writing these functions. The functions are:

- `read_api_dataflows()` lists available dataflows (roughly equivalent
  to ‘tables’)
- `read_api_datastructure()` lists variables within a particular
  dataflow and the levels of those variables, which you can use to
  filter the data server-side in an API query
- `read_api()` downloads data from the ABS.Stat API.

Let’s list available dataflows:

``` r
flows <- read_api_dataflows()
```

Say from this I am interested in the first dataflow, the projected
population of Aboriginal and Torres Strait Islander Australians. The id
for this dataflow is `"ABORIGINAL_POP_PROJ"`, which I can use to
download the data.

In this case, I could download the entire dataflow with:

``` r
read_api("ABORIGINAL_POP_PROJ")
#> # A tibble: 72,144 × 11
#>    measure         sex_abs age       asgs_2011 proj_series frequency time_period
#>    <chr+lbl>       <chr+l> <chr+lbl> <chr+lbl> <chr+lbl>   <chr+lbl>       <dbl>
#>  1 POP_PROJ [Proj… 3 [Per… A04 [0 -… 1 [New S… 3 [Series … A [Annua…        2016
#>  2 POP_PROJ [Proj… 3 [Per… A04 [0 -… 1 [New S… 3 [Series … A [Annua…        2017
#>  3 POP_PROJ [Proj… 3 [Per… A04 [0 -… 1 [New S… 3 [Series … A [Annua…        2018
#>  4 POP_PROJ [Proj… 3 [Per… A04 [0 -… 1 [New S… 3 [Series … A [Annua…        2019
#>  5 POP_PROJ [Proj… 3 [Per… A04 [0 -… 1 [New S… 3 [Series … A [Annua…        2020
#>  6 POP_PROJ [Proj… 3 [Per… A04 [0 -… 1 [New S… 3 [Series … A [Annua…        2021
#>  7 POP_PROJ [Proj… 3 [Per… A04 [0 -… 1 [New S… 3 [Series … A [Annua…        2022
#>  8 POP_PROJ [Proj… 3 [Per… A04 [0 -… 1 [New S… 3 [Series … A [Annua…        2023
#>  9 POP_PROJ [Proj… 3 [Per… A04 [0 -… 1 [New S… 3 [Series … A [Annua…        2024
#> 10 POP_PROJ [Proj… 3 [Per… A04 [0 -… 1 [New S… 3 [Series … A [Annua…        2025
#> # ℹ 72,134 more rows
#> # ℹ 4 more variables: obs_value <dbl>, unit_measure <chr+lbl>,
#> #   obs_status <chr+lbl>, obs_comment <dbl>
```

Let’s say I’m only interested in the population projections for males,
not females or all persons. In that case, I can filter the data on the
ABS server before downloading my query. I can use
`read_api_datastructure()` to help with this.

``` r
read_api_datastructure("ABORIGINAL_POP_PROJ")
#> # A tibble: 332 × 6
#>    role      var     position desc    code     label               
#>    <chr>     <chr>   <chr>    <chr>   <chr>    <chr>               
#>  1 dimension measure 1        Measure POP_PROJ Projected population
#>  2 dimension sex_abs 2        Sex     1        Males               
#>  3 dimension sex_abs 2        Sex     2        Females             
#>  4 dimension sex_abs 2        Sex     3        Persons             
#>  5 dimension age     3        Age     0        0                   
#>  6 dimension age     3        Age     0014     0 - 14              
#>  7 dimension age     3        Age     0015     0 - 15              
#>  8 dimension age     3        Age     0104     1 - 4               
#>  9 dimension age     3        Age     0514     5 - 14              
#> 10 dimension age     3        Age     1        1                   
#> # ℹ 322 more rows
```

From this, I can see that there’s a variable (`var`) called `sex_abs`,
which can take the value `1`, `2`, or `3`, corresponding to `Males`,
`Females` and `Persons`. If I only want to data for Males, I can obtain
this by supplying a datakey:

``` r
read_api("ABORIGINAL_POP_PROJ", datakey = list(sex_abs = 1))
#> # A tibble: 24,048 × 11
#>    measure         sex_abs age       asgs_2011 proj_series frequency time_period
#>    <chr+lbl>       <chr+l> <chr+lbl> <chr+lbl> <chr+lbl>   <chr+lbl>       <dbl>
#>  1 POP_PROJ [Proj… 1 [Mal… A04 [0 -… 1 [New S… 7 [Series … A [Annua…        2016
#>  2 POP_PROJ [Proj… 1 [Mal… A04 [0 -… 1 [New S… 7 [Series … A [Annua…        2017
#>  3 POP_PROJ [Proj… 1 [Mal… A04 [0 -… 1 [New S… 7 [Series … A [Annua…        2018
#>  4 POP_PROJ [Proj… 1 [Mal… A04 [0 -… 1 [New S… 7 [Series … A [Annua…        2019
#>  5 POP_PROJ [Proj… 1 [Mal… A04 [0 -… 1 [New S… 7 [Series … A [Annua…        2020
#>  6 POP_PROJ [Proj… 1 [Mal… A04 [0 -… 1 [New S… 7 [Series … A [Annua…        2021
#>  7 POP_PROJ [Proj… 1 [Mal… A04 [0 -… 1 [New S… 7 [Series … A [Annua…        2022
#>  8 POP_PROJ [Proj… 1 [Mal… A04 [0 -… 1 [New S… 7 [Series … A [Annua…        2023
#>  9 POP_PROJ [Proj… 1 [Mal… A04 [0 -… 1 [New S… 7 [Series … A [Annua…        2024
#> 10 POP_PROJ [Proj… 1 [Mal… A04 [0 -… 1 [New S… 7 [Series … A [Annua…        2025
#> # ℹ 24,038 more rows
#> # ℹ 4 more variables: obs_value <dbl>, unit_measure <chr+lbl>,
#> #   obs_status <chr+lbl>, obs_comment <dbl>
```

Note that in some cases, querying the API without filtering the data
will return an error, as the table will be too big. In this case, you
will need to supply a datakey that reduces the size of the data.

## Resolving network issues by manually setting the download method

Certain corporate networks restrict your ability to download files in an
R session. On some of these networks, the `"wininet"` method must be
used when downloading files. Users can now specify the method that will
be used to download files by setting the `"R_READABS_DL_METHOD"`
environment variable.

For example, the following code sets the environment variable for your
current session:

``` r
Sys.setenv("R_READABS_DL_METHOD" = "wininet")
```

You can add `R_READABS_DL_METHOD = "wininet"` to your .Renviron to have
this persist across sessions.

If you have other issues using `{readabs}` in your corporate
environment, I would appreciate you opening an issue on GitHub.

## Bug reports and feedback

GitHub issues containing error reports or feature requests are welcome.
Please try to make a [reprex](https://reprex.tidyverse.org) (a minimal,
reproducible example) if possible.

Alternatively you can email the package maintainer at mattcowgill at
gmail dot com.

## Disclaimer

The `{readabs}` package is not associated with the Australian Bureau of
Statistics. All data is provided subject to any restrictions and
licensing arrangements noted on the ABS website.

## Awesome Official Statistics Software

[![Mentioned in Awesome Official
Statistics](https://awesome.re/mentioned-badge.svg)](https://github.com/SNStatComp/awesome-official-statistics-software)

We’re pleased to be included in a [list of
software](https://github.com/SNStatComp/awesome-official-statistics-software)
that can be used to work with official statistics.

#' Internal functions for working with the fst cache
#' @name fst-utils
#' @noRd
#'
#' @param cat_no,path As in \code{\link{read_abs}}.
#'
#' @return For `catno2fst` the path to the `fst` file to be saved or read, given
#' `cat_no` and `path`.
#'
#' `fst_available` returns `TRUE` if and only if an appropriate `fst` file is
#' available.
#'
#' `ext2ext` changes the extension of the provided file to a file in the same
#' path but with the provided extension.
#'


catno2fst <- function(cat_no,
                      path = Sys.getenv("R_READABS_PATH", unset = tempdir())) {
  hutils::provide.file(
    file.path(
      path,
      "fst",
      paste0(
        gsub(".", "-", cat_no, fixed = TRUE),
        ".fst"
      )
    ),
    on_failure = stop(
      "`path = ", normalizePath(path,
        winslash = "/"
      ),
      "`, ",
      "but it was not possible to write to this directory."
    )
  )
}

fst_available <- function(cat_no,
                          path = Sys.getenv("R_READABS_PATH",
                            unset = tempdir()
                          )) {
  if (!requireNamespace("fst", quietly = TRUE) ||
    !dir.exists(path)) {
    return(FALSE)
  }

  if (!is.character(cat_no) ||
    length(cat_no) != 1L ||
    anyNA(cat_no) ||
    nchar(cat_no) < 6L) {
    return(FALSE)
  }

  file.fst <- catno2fst(cat_no, path)

  if (!file.exists(file.fst)) {
    return(FALSE) # nocov
  }

  # fst may be damaged. If it appears to be (i.e. fst metadata returns an error)
  #   return FALSE


  out <- tryCatch(inherits(fst::fst.metadata(file.fst), "fstmetadata"),
    error = function(e) FALSE,
    warning = function(e) FALSE
  )

  out
}

ext2ext <- function(file, new.ext) {
  paste0(tools::file_path_sans_ext(file), new.ext)
}

test_that("read_abs works out of the box", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  out6345_init <- read_abs(cat_no = "6345.0")
  skip_if_not(file.exists(catno2fst("6345.0")))
  out6345_next <- read_abs(cat_no = "6345.0")


  expect_equal(out6345_init, out6345_next)

  out6345_A2599619A <- read_abs(cat_no = "6345.0", series_id = "A2599619A")
  expect_equal(out6345_A2599619A, dplyr::filter(out6345_init, series_id == "A2599619A"))

  # warning handling --- must be after out6401_next
  expect_warning(read_abs(cat_no = "6345.0", series_id = "not a real series id"),
    regexp = "but was not present in the local table and will be ignored",
    fixed = TRUE
  )

  expect_warning(read_abs(cat_no = "6345.0", tables = "not all"),
    regexp = "tables.*will be ignored"
  )
})

test_that("fst-utils", {
  expect_false(fst_available(cat_no = NULL))
  expect_false(fst_available(cat_no = character(2)))
  expect_false(fst_available(cat_no = NA_character_))
  expect_false(fst_available(cat_no = ""))
  tempf.fst <- tempfile(fileext = ".fst")
  writeLines("9999-9", tempf.fst)
  expect_false(fst_available("9999.9", path = dirname(tempf.fst)))
  expect_equal(ext2ext("file.csv", ".txt"), "file.txt")
})

#' @param urls vector of URLs pointing to XML files
#' @return a tibble
#' @noRd
#' @importFrom dplyr filter select "%>%"

get_xml_dfs <- function(urls) {
  xml_files <- tempfile(
    pattern = as.character(seq_along(urls)),
    fileext = ".xml"
  )

  purrr::walk2(
    urls, xml_files,
    ~ dl_file(
      url = .x,
      destfile = .y
    )
  )

  xml_to_df <- function(file) {
    xml_series <- file %>%
      xml2::read_xml() %>%
      xml2::xml_find_all("//Series")

    child_list <- xml_series %>%
      purrr::map(xml2::xml_children)

    xml_names <- child_list %>%
      purrr::pluck(1) %>%
      xml2::xml_name(xml_series)

    child_list %>%
      purrr::map(xml2::xml_text) %>%
      purrr::map_dfr(.f = ~ purrr::set_names(.x, xml_names))
  }

  purrr::map_dfr(xml_files, xml_to_df)
}

.onAttach <- function(libname = find.package("readabs"), pkgname = "readabs") {
  if (is.na(Sys.getenv("R_READABS_PATH", unset = NA))) {
    packageStartupMessage(
      "Environment variable 'R_READABS_PATH' is unset. ",
      "Downloaded files will be saved in a temporary directory.\n",
      "You can set 'R_READABS_PATH' at any time. To set it for ",
      "the rest of this session, use\n\tSys.setenv(R_READABS_PATH = <path>)"
    )
  }
  invisible(NULL)
}

#' Search for a file within an ABS catalogue
#'
#' @param string String to search for among filenames in a catalogue
#' @param catalogue Name of catalogue
#' @param refresh logical; `FALSE` by default. When `TRUE`, will re-scrape the
#' list of files within the catalogue.
#' @export
#' @examples
#' \dontrun{
#' search_files("GM1", "labour-force-australia")
#' }
#'
search_files <- function(string, catalogue, refresh = FALSE) {
  files <- show_available_files(
    catalogue_string = catalogue,
    refresh = refresh
  )

  # Using dplyr::if_any() is fast and clean but requires dplyr >= 1.0.4
  if (getNamespaceVersion("dplyr") > "1.0.4") {
    out <- files %>%
      dplyr::filter(dplyr::if_any(
        .cols = dplyr::everything(),
        ~ grepl(string, .x,
          perl = TRUE, ignore.case = TRUE
        )
      ))
  } else {
    matches <- purrr::map_dfr(files,
      grepl,
      pattern = string, perl = TRUE, ignore.case = TRUE
    ) %>%
      rowSums()

    out <- dplyr::tibble(sum_true = matches) %>%
      dplyr::bind_cols(files) %>%
      dplyr::filter(.data$sum_true >= 1) %>%
      dplyr::select(-.data$sum_true)
  }

  out <- out %>%
    dplyr::mutate(url = as.character(url)) %>%
    dplyr::pull(file)

  return(out)
}

#' Separate the series column in a tidy ABS time series data frame
#'
#' Separate the 'series' column in a data frame (tibble)
#' downloaded using \code{read_abs()} into multiple columns using the ";"
#' separator.
#'
#' @param data A data frame (tibble) containing tidied data from the ABS time
#' series table(s).
#' @param column_names (optional) character vector. Supply a vector of column
#' names, such as \code{c("group_name", "variable","gender")}. If not
#' supplied, columns will be named "series_1" etc.
#' @param remove_totals logical. FALSE by default.
#' If set to TRUE, any series rows that contain the word "total"
#' will be removed.
#' @param remove_nas locical. FALSE by default. If set to TRUE, any rows
#' containining an NA in at least one of the separated series columns
#' will be removed.
#'
#' @return A data frame (tibble) containing the tidied data from the ABS time
#' series table(s).
#'
#' @export
#'
#' @examples
#' \dontrun{
#' wpi <- read_abs("6345.0", 1) %>%
#'   separate_series()
#' }
#'
#' @importFrom stringi stri_trim_both
#' @importFrom dplyr filter mutate_at sym
#' @importFrom tidyr separate
#' @importFrom purrr map_dfr

separate_series <- function(data,
                            column_names = NULL,
                            remove_totals = FALSE,
                            remove_nas = FALSE) {
  # satisfy R CMD check's "no visible binding" error
  series <- NULL
  original_series <- NULL

  # Create copy of series column to restore later
  data <- mutate(data,
    original_series = series
  )

  # Minor data cleaning of series column
  data <- mutate(data,
    # Extract everything before trailing ;
    series = regmatches(
      series,
      regexpr(".+(?= ;)", series, perl = TRUE)
    ),
    series = gsub(
      pattern = ">", series,
      replacement = "", perl = TRUE
    ),
    series = fast_str_squish(series)
  )

  # Filter totals if specified
  if (remove_totals) {
    data <- filter(
      data,
      !grepl("total", series, ignore.case = T, perl = TRUE)
    )
  }

  # Determine number of ; separators in series column
  n_seps <- max(stringi::stri_count_fixed(data$series, ";"))

  # Determine number of columns to split series into
  n_columns <- n_seps + 1

  # Create new column names if not specified
  if (is.null(column_names)) {
    column_names <- paste("series", 1:n_columns, sep = "_")
  }

  # Separate columns
  data_separated <- separate(data, series,
    into = column_names,
    sep = ";",
    remove = FALSE,
    fill = "left"
  ) %>%
    mutate_at(.vars = vars(column_names), fast_str_squish)


  # check for columns with NAs
  na_columns <- colnames(data_separated)[colSums(is.na(data_separated)) > 0]

  if (length(na_columns) > 0 & !remove_nas) {
    warning(paste0(na_columns, collapse = ", "),
      " column(s) have NA values.",
      call. = TRUE
    )
  }

  # filter NAs in new series columns

  if (remove_nas & length(na_columns) > 0) {
    remove_na_fn <- function(df, column) {
      col_sym <- dplyr::sym(column)

      df %>%
        filter(!is.na(!!col_sym))
    }

    data_separated <- purrr::map_dfr(
      .x = na_columns,
      .f = remove_na_fn,
      df = data_separated
    )

    message("Rows with NAs in separated series column(s) have been removed.")
  }


  # Replace original series column
  data_separated <- mutate(data_separated,
    series = original_series
  )

  data_separated <- select(
    data_separated,
    -original_series
  )

  return(data_separated)
}

local_path <- "../testdata"
local_filename <- "6202021.xls"



test_that("separate_series() performs as expected", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  wpi_raw <- read_abs("6345.0",
    tables = "1",
    check_local = FALSE,
    retain_files = FALSE,
    path = tempdir()
  )

  wpi <- suppressWarnings(separate_series(wpi_raw))

  expect_equal(17, length(wpi))

  expect_equal(5, length(unique(wpi$series_1)))

  expect_true(all.equal(wpi$series, wpi_raw$series))
})

test_that("separate_series(remove_nas = TRUE) removes NAs", {
  lfs_21 <- read_abs_local(
    filename = local_filename,
    path = local_path
  )

  lfs_21_sep <- suppressWarnings(separate_series(lfs_21))

  non_nas <- lfs_21_sep %>%
    filter(!is.na(series_1) & !is.na(series_2)) %>%
    nrow()

  nas <- lfs_21_sep %>%
    filter(is.na(series_1) | is.na(series_2)) %>%
    nrow()

  expect_warning(separate_series(lfs_21))

  expect_message(separate_series(lfs_21, remove_nas = TRUE))

  expect_equal(nrow(separate_series(lfs_21, remove_nas = TRUE)), non_nas)

  expect_equal(nas + non_nas, nrow(lfs_21_sep))
})


test_that("separate_series works with remove_totals and remove_nas both TRUE", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  awote <- read_abs("6302.0", "10G", retain_files = FALSE, path = tempdir())

  expect_is(
    suppressWarnings(separate_series(awote)),
    "tbl_df"
  )

  expect_is(
    suppressWarnings(separate_series(awote, remove_totals = TRUE)),
    "tbl_df"
  )

  expect_is(
    suppressWarnings(separate_series(awote, remove_nas = TRUE)),
    "tbl_df"
  )

  expect_is(
    suppressWarnings(separate_series(awote, remove_totals = TRUE, remove_nas = TRUE)),
    "tbl_df"
  )
})

#' Helper function to show the files available in a particular catalogue number.
#'
#' @description
#' `r lifecycle::badge("experimental")`
#' To be used in conjunction with \code{download_abs_data_cube()}.
#'
#' This function lists the possible files that are available in a catalogue.
#' The filename (or an unambiguous part of the filename) must be specified
#' as a string as an argument to \code{download_abs_data_cube}.
#'
#' @param catalogue_string character string specifying the catalogue,
#' e.g. "labour-force-australia-detailed".
#' You can use \code{show_available_catalogues()} see all the possible catalogues,
#' or \code{search_catalogues()} to find catalogues that contain a given string.
#'
#' @param refresh logical; `FALSE` by default. If `FALSE`, an internal table
#' of the available ABS catalogues is used. If `TRUE`, this table is refreshed
#' from the ABS website.
#'
#' @details `get_available_files()` is an alias for `show_available_files()`.
#'
#' @return A tibble containing the title of the file, the filename and the complete url.
#'
#' @examples
#' \dontrun{
#' show_available_files("labour-force-australia-detailed")
#' }
#'
#' @importFrom glue glue
#' @importFrom dplyr  %>% filter pull slice tibble
#' @importFrom rvest html_nodes html_attr html_text
#' @importFrom rlang .data
#'
#' @export
#' @family data cube functions

show_available_files <- function(catalogue_string, refresh = FALSE) {
  if (isFALSE(is.logical(refresh))) {
    stop("`refresh` must be `TRUE` or `FALSE`.")
  }

  if (isFALSE(refresh)) {
    table_to_use <- abs_lookup_table
  } else {
    table_to_use <- scrape_abs_catalogues()
  }

  download_url <- filter(
    table_to_use,
    .data$catalogue == catalogue_string
  ) %>%
    pull(url)

  if (length(download_url) == 0) {
    stop("No matching catalogue. Please check against ABS website.")
  }


  # Try to download the page
  tryCatch(
    {
      temp_page_location <- file.path(tempdir(), "temp_readabs.html")
      dl_file(
        url = download_url,
        destfile = temp_page_location
      )
    },
    error = function(cond) {
      message(paste("URL does not seem to exist:", download_url))
      message("Here's the original error message:")
      message(cond)
      # Choose a return value in case of error
      return(NA)
    }
  )

  download_page <- xml2::read_html(temp_page_location)

  # Find the url for the download

  urls <- download_page %>%
    rvest::html_nodes(".file a") %>%
    rvest::html_attr("href")

  urls <- ifelse(stringi::stri_sub(urls, 2L, 11L) == "statistics",
    paste0("https://www.abs.gov.au", urls),
    urls
  )

  labels <- download_page %>%
    rvest::html_nodes(".file-description-link-left, .abs-data-download-left") %>%
    rvest::html_text()

  # fix for some inconsistent CSS on ABS website
  if (length(labels) != length(urls)) {
    labels <- download_page %>%
      rvest::html_nodes("h4") %>%
      rvest::html_text()
  }

  # fix for anything that doesn't fit the above patterns

  if (length(labels) != length(urls)) {
    labels <- rep(NA, length(urls))
  }


  available_downloads <- dplyr::tibble(
    url = urls,
    label = labels
  ) %>%
    mutate(file = stringi::stri_extract_first_regex(url, "[^/]*$")) %>%
    select("label", "file", "url")

  return(available_downloads)
}

#' @rdname show_available_files
#' @family data cube functions
#' @export

get_available_files <- function(catalogue_string, refresh = FALSE) {
  show_available_files(
    catalogue_string = catalogue_string,
    refresh = refresh
  )
}

test_that("get_available_files() works", {
  skip_if_offline()
  skip_on_cran()
  check_abs_connection()

  cats <- show_available_catalogues()

  get_available_files_printname <- function(cat) {
    print(cat)
    get_available_files(cat)
  }

  safely_get_available_files <- purrr::safely(get_available_files_printname)

  results <- purrr::map(cats, safely_get_available_files)

  results <- setNames(results, cats)

  results <- purrr::transpose(results)

  errors <- purrr::compact(results$error)

  error_names <- names(errors)

  expect_equal(length(error_names), 0)
})

#' Helper function for \code{download_abs_data_cube} to scrape the available catalogues from the ABS website.
#'
#' This function downloads a new version of the lookup table used by \code{show_available_catalogues}.
#'
#' @return A tibble containing the catalogues and how they are organised on the ABS website.
#'
#'
#' @importFrom glue glue
#' @importFrom xml2 read_html
#' @importFrom dplyr  %>% filter pull slice
#' @importFrom rvest html_nodes html_attr html_text
#' @importFrom purrr map_dfr
#' @importFrom rlang .data
#'

scrape_abs_catalogues <- function() {
  # Scrape the main ABS statistics page
  stats_page_file <- tempfile(fileext = ".html")
  dl_file(
    url = "https://www.abs.gov.au/statistics",
    destfile = stats_page_file
  )

  abs_stats_page <- xml2::read_html(stats_page_file)

  abs_stats_page_cards <- abs_stats_page %>%
    rvest::html_nodes(".layout__region--content .card")

  headings <- abs_stats_page_cards %>%
    rvest::html_elements("h3") %>%
    rvest::html_text() %>%
    stringi::stri_trim_both()

  url_suffixes <- abs_stats_page_cards %>%
    rvest::html_attr("href") %>%
    stringi::stri_trim_both()

  main_page_data <- dplyr::tibble(
    heading = .env$headings,
    url_suffix = .env$url_suffixes
  )

  # scrape each page

  scrape_sub_page <- function(heading, sub_page_url_suffix) {
    sub_page_file <- tempfile(fileext = ".html")
    dl_file(
      url = glue::glue("https://www.abs.gov.au{sub_page_url_suffix}"),
      destfile = sub_page_file
    )

    sub_page <- xml2::read_html(sub_page_file)

    abs_cards <- sub_page %>%
      rvest::html_elements(".clearfix .card")

    sub_heading <- abs_cards %>%
      rvest::html_elements(".abs-layout-title") %>%
      rvest::html_text() %>%
      stringi::stri_trim_both()

    catalogue <- abs_cards %>%
      rvest::html_attr("href") %>%
      stringi::stri_replace_all_fixed(sub_page_url_suffix, "") %>%
      stringi::stri_replace_all_regex("/[^/]*$", "") %>%
      stringi::stri_replace_all_fixed("/", "")

    sub_page_data <- dplyr::tibble(
      heading = heading,
      sub_heading = sub_heading,
      catalogue = catalogue,
      url = glue::glue("https://www.abs.gov.au{sub_page_url_suffix}/{catalogue}/latest-release")
    )
  }


  new_abs_lookup_table <- purrr::map2_dfr(
    .x = main_page_data$heading,
    .y = main_page_data$url_suffix,
    scrape_sub_page
  )

  new_abs_lookup_table %>%
    filter(.data$catalogue != "")
}

test_that("Lookup table down download_data_cube() is up to date", {
  skip_if_offline()
  skip_on_cran()
  check_abs_connection()

  updated_lookup_table <- scrape_abs_catalogues()

  expect_equal(
    dplyr::arrange(updated_lookup_table, url),
    dplyr::arrange(abs_lookup_table, url)
  )
})

test_that("Data has been updated within last 120 days", {
  skip_on_cran()

  todays_date <- Sys.Date()

  expect_lt(todays_date - data_last_updated, 120)
})

#' @name read_awe
#' @description Convenience function to obtain wage levels from ABS
#' 6302.0, Average Weekly Earnings, Australia.
#' @title read_awe
#' @param wage_measure Character of length 1. Must be one of:
#' \describe{
#'   \item{`awote`}{ Average weekly ordinary time earnings; also known as Full-time adult ordinary time earnings}
#'   \item{`ftawe`}{ Full-time adult total earnings}
#'   \item{`awe`}{ Average weekly total earnings of all employees}
#' }
#' @param sex Character of length 1. Must be one of: `persons`, `males`, or `females`.
#' @param sector Character of length 1. Must be one of: `total`, `private`, or
#' `public`. Note that you cannot get sector-by-state data; if `state` is not
#' `all` then `sector` must be `total`.
#' @param state Character of length 1. Must be one of: `all`, `nsw`, `vic`, `qld`,
#' `sa`, `wa`, `nt`, or `act`. Note that you cannot get sector-by-state data;
#' if `sector` is not `total` then `state` must be `all`.
#' @param na.rm Logical. `FALSE` by default. If `FALSE`, a consistent quarterly
#' series is returned, with `NA` values for quarters in which there is no data.
#' If `TRUE`, only dates with data are included in the returned data frame.
#' @param path See `?read_abs`
#' @param show_progress_bars See `?read_abs`
#' @param check_local See `?read_abs`
#' @details
#' The latest AWE data is available using `read_abs(cat_no = "6302.0", tables = 2)`.
#' However, this time series only goes back to 2012, when the ABS switched
#' from quarterly to biannual collection and release of the AWE data. The
#' `read_awe()` function assembles on time series back to November 1983 quarter;
#' it is quarterly to 2012 and biannual from then. Note that the data
#' returned with this function is consistently quarterly; any quarters for
#' which there are no observations are recorded as `NA` unless `na.rm` = `TRUE`.
#' @return
#' A `tbl_df` with four columns: `date`, `sex`, `wage_measure` and `value`.
#' The data is nominal and seasonally adjusted.
#'
#' @examples
#' \dontrun{
#' read_awe("awote", "persons")
#' }
#'
#' @export
read_awe <- function(wage_measure = c(
                       "awote",
                       "ftawe",
                       "awe"
                     ),
                     sex = c(
                       "persons",
                       "males",
                       "females"
                     ),
                     sector = c(
                       "total",
                       "private",
                       "public"
                     ),
                     state = c(
                       "all",
                       "nsw",
                       "vic",
                       "qld",
                       "sa",
                       "wa",
                       "tas",
                       "nt",
                       "act"
                     ),
                     na.rm = FALSE,
                     path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
                     show_progress_bars = FALSE,
                     check_local = FALSE) {
  .wage_measure <- match.arg(wage_measure)
  .sex <- match.arg(sex)
  .sector <- match.arg(sector)
  .state <- match.arg(state)

  check_abs_connection()

  if (.sector != "total" &
    .state != "all") {
    stop(
      'You cannot get sector-by-state data. Either set sector to "total"',
      ' or state to "all".'
    )
  }

  if (.state == "all") {
    tables <- switch(.sector,
      "total" = "2",
      "private" = "5",
      "public" = "8"
    )

    if (.sector == "total") {
      crosstab_name <- ""
    } else {
      crosstab_name <- "sector"
    }
  } else {
    tables <- switch(.state,
      "nsw" = "12a",
      "vic" = "12b",
      "qld" = "12c",
      "sa" = "12d",
      "wa" = "12e",
      "tas" = "12f",
      "nt" = "12g",
      "act" = "12h"
    )

    crosstab_name <- "state"
  }

  awe_latest <- suppressMessages(read_abs(
    cat_no = "6302.0",
    tables = tables,
    path = path,
    show_progress_bars = show_progress_bars,
    check_local = check_local
  ))

  awe_latest <- tidy_awe(df = awe_latest)

  # awe_old is an internal data object created in /data-raw
  awe_old_table <- bind_rows(awe_old[tables])

  awe_old_table <- awe_old_table %>%
    dplyr::filter(!.data$date %in% awe_latest$date)

  awe <- dplyr::bind_rows(awe_old_table, awe_latest)

  awe <- awe %>%
    filter(
      .data$sex == .sex,
      .data$wage_measure == .wage_measure
    )

  if (isFALSE(na.rm)) {
    # Pad to ensure data is quarterly
    missing_dates <- expand.grid(
      month = unique(format(awe$date, "%m")),
      year = unique(format(awe$date, "%Y")),
      day = 15,
      sex = .sex,
      wage_measure = .wage_measure,
      stringsAsFactors = FALSE
    ) %>%
      dplyr::mutate(date = as.Date(paste(.data$year,
        .data$month,
        .data$day,
        sep = "-"
      ))) %>%
      dplyr::filter(!date %in% awe$date &
        date > min(awe$date) &
        date < max(awe$date)) %>%
      mutate(value = NA_real_) %>%
      dplyr::select(dplyr::any_of(c("date", "sex", "wage_measure", "value", "crosstab")))

    if (!is.null(awe[["crosstab"]])) {
      missing_dates$crosstab <- unique(awe$crosstab)
    }

    awe <- missing_dates %>%
      dplyr::bind_rows(awe) %>%
      dplyr::as_tibble()
  } else {
    awe <- awe %>%
      dplyr::filter(!is.na(.data$value))
  }

  names(awe)[names(awe) == "crosstab"] <- crosstab_name

  if (!is.null(awe[["state"]])) {
    awe <- awe %>%
      dplyr::mutate(state = dplyr::case_when(
        state == "new south wales" ~ "nsw",
        state == "victoria" ~ "vic",
        state == "queensland" ~ "qld",
        state == "south australia" ~ "sa",
        state == "western australia" ~ "wa",
        state == "tasmania" ~ "tas",
        state == "northern territory" ~ "nt",
        state == "australian capital territory" ~ "act"
      ))
  }

  awe <- awe %>%
    dplyr::arrange(.data$date)

  awe
}


#' Internal function to tidy a dataframe from ABS 6302
#' @param df Data frame containing table 2 from ABS 6302, imported using `read_abs()`
#' @keywords internal
tidy_awe <- function(df) {
  df <- df %>%
    dplyr::select("series", "date", "value")

  df <- df %>%
    dplyr::mutate(
      series = fast_str_squish(.data$series),
      series = stringi::stri_replace_all_fixed(.data$series, " ;", ";")
    )

  # Usually the cross tab (eg. state, sector) is at the end of the series string;
  # For fun, sometimes the ABS puts it as the second element!
  df <- df %>%
    tidyr::separate(.data$series,
      into = c("earnse_crosstab1", "series"),
      sep = "(?=Males|Females|Persons;)",
      extra = "merge",
      fill = "right"
    )

  df <- df %>%
    tidyr::separate(.data$earnse_crosstab1,
      into = c("earnse", "crosstab1"),
      sep = ";",
      extra = "merge",
      fill = "right"
    )

  df <- df %>%
    tidyr::separate(.data$series,
      into = c(
        "sex_measure",
        "crosstab2"
      ),
      sep = "earnings;",
      extra = "merge",
      fill = "right"
    )

  df <- df %>%
    dplyr::mutate(
      crosstab = paste0(.data$crosstab1, .data$crosstab2),
      crosstab = fast_str_squish(.data$crosstab),
      crosstab = dplyr::if_else(.data$crosstab == "",
        NA_character_,
        .data$crosstab
      )
    )

  df <- df %>%
    mutate(crosstab = stringi::stri_replace_all_fixed(.data$crosstab, " sector", ""))

  df <- df %>%
    # Drop the crosstab column if it's full of NAs
    dplyr::select_if(~ all(!is.na(.))) %>%
    dplyr::select(-"crosstab1", -"crosstab2")

  df <- df %>%
    tidyr::separate(.data$sex_measure,
      into = c("sex", "measure"),
      sep = ";",
      extra = "merge",
      fill = "right"
    )

  fix_col <- function(col) {
    col <- gsub(";", "", col, fixed = TRUE)
    col <- tolower(col)
    col <- fast_str_squish(col)
    col
  }

  df <- df %>%
    dplyr::mutate(dplyr::across(
      dplyr::any_of(c("earnse", "sex", "measure", "crosstab")),
      fix_col
    ))

  # Some sheets contain standard errors of estimates; we want to drop these
  df <- df %>%
    dplyr::filter(.data$earnse == "earnings") %>%
    dplyr::select(-"earnse")

  df <- df %>%
    dplyr::mutate(
      wage_measure = dplyr::case_when(
        .data$measure == "full time adult ordinary time" ~ "awote",
        .data$measure == "full time adult total" ~ "ftawe",
        .data$measure == "total" ~ "awe",
        TRUE ~ NA_character_
      )
    )

  df <- df %>%
    dplyr::select(dplyr::any_of(c("date", "sex", "wage_measure", "value", "crosstab")))

  df
}

test_that("tidy_awe() returns tidied data frame", {
  awe <- readxl::read_excel(file.path("..", "testdata", "6302002.xls"),
    sheet = "Data1"
  )

  awe <- tidy_abs(awe)

  tidied_awe <- tidy_awe(awe)

  expect_is(tidied_awe, "data.frame")
  expect_equal(nrow(tidied_awe), 927)
  expect_true(all(
    c("date", "sex", "wage_measure", "value") %in% names(tidied_awe)
  ))
  expect_false(any(is.na(tidied_awe$value)))
})

test_that("read_awe() returns expected output", {
  skip_if_offline()
  skip_on_cran()

  no_params <- read_awe()

  expect_is(no_params, "tbl_df")
  expect_identical(unique(no_params$sex), "persons")
  expect_identical(unique(no_params$wage_measure), "awote")
  expect_identical(min(no_params$date), as.Date("1983-11-15"))
  expect_gt(max(no_params$date), as.Date("2020-05-14"))
  expect_is(no_params$value, "numeric")
  expect_gt(
    max(no_params$value, na.rm = T) /
      min(no_params$value, na.rm = T),
    4.5
  )

  params_df <- expand.grid(
    sex = c("persons", "males", "females"),
    wage_measure = c("awote", "ftawe", "awe"),
    sector = c("total", "private", "public"),
    state = c(
      "all",
      "nsw",
      "vic",
      "qld",
      "sa",
      "wa",
      "tas",
      "nt",
      "act"
    ),
    stringsAsFactors = FALSE
  )

  params_df <- params_df %>%
    dplyr::filter(sector == "total" | state == "all")


  for (i in seq_len(nrow(params_df))) {
    awe_data <- read_awe(
      wage_measure = params_df$wage_measure[i],
      sex = params_df$sex[i],
      sector = params_df$sector[i],
      state = params_df$state[i]
    )

    expect_is(awe_data, "tbl_df")

    expect_gt(length(awe_data), 3)

    if (params_df$sector[i] != "total") {
      expect_length(unique(awe_data$sector), 1)
      expect_true(unique(awe_data$sector) != "total")
      expect_true(unique(awe_data$sector) == params_df$sector[i])
      expect_length(awe_data, 5)
      expect_true("sector" %in% names(awe_data))
    }

    if (params_df$state[i] != "all") {
      expect_length(unique(awe_data$state), 1)
      expect_true(unique(awe_data$state) != "all")
      expect_true(unique(awe_data$state) == params_df$state[i])
      expect_length(awe_data, 5)
      expect_true("state" %in% names(awe_data))
    }
  }



  expect_gt(nrow(read_awe()), nrow(read_awe(na.rm = TRUE)))
})

#' Download a tidy tibble containing the Consumer Price Index from the ABS
#'
#' \code{read_cpi()} uses the \code{read_abs()} function to download, import,
#' and tidy the quarterly Consumer Price Index from the ABS. It returns a tibble
#' containing two columns: the date and the CPI index value that corresponds
#' to that date. This makes joining the CPI to another dataframe easy.
#' \code{read_cpi()} returns the original (ie. not seasonally adjusted)
#' all groups CPI for Australia. If you want the analytical series
#' (eg. seasonally adjusted CPI, or trimmed mean CPI) or monthly series, you can use
#' \code{read_abs()}.
#'
#' By default, `read_abs()` retrieves the quarterly CPI. It can use the
#' `frequency` argument to change to monthly.
#'
#' @param path character; default is "data/ABS". Only used if
#' retain_files is set to TRUE. Local directory in which to save
#' downloaded ABS time series spreadsheets.
#' @param frequency Either `"quarterly"` (the default) or `"monthly"`.
#' @param ... Arguments passed to `read_abs()`. See `?read_abs` for details.
#'
#' @examples
#'
#' # Create a tibble called 'cpi' that contains the CPI index
#' # numbers for each quarter
#' \donttest{
#' cpi <- read_cpi()
#' }
#'
#' # This tibble can now be joined to another to help streamline the process of
#' # deflating nominal values.
#' @importFrom dplyr filter select
#'
#' @export

read_cpi <- function(frequency = c("quarterly", "monthly"),
                     path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
                     ...) {

  frequency <- match.arg(frequency)

  cpi_series_id <- switch (frequency,
    "quarterly" = "A2325846C",
    "monthly" = "A130393720C"
  )


  cpi_raw <- read_abs_series(
    series_id = cpi_series_id,
    ...
  )

  series_name <- unique(cpi_raw$series)

  stopifnot(series_name == "Index Numbers ;  All groups CPI ;  Australia ;")

  x <- cpi_raw %>%
    dplyr::select("date",
      "cpi" = "value"
    )

  x
}

#' @importFrom rlang .data .env
#' @import dplyr
#' @noRd
#' @param url URL for an XML file in the ABS Time Series Directory
# given a catalogue number, download the catalogue metadata via XML, then find
# unique filenames in the latest release and return those

get_abs_xml_metadata <- function(url) {
  if (!is.character(url)) {
    stop("`url` argument to get_abs_xml_metadata() must be a string.")
  }

  first_page_df <- get_first_xml_page(url)

  num_pages <- get_numpages(url)

  xml_dfs <- dplyr::tibble()

  # If there's more than one page of XML corresponding to request, get all of them
  if (num_pages > 1) {
    all_pages <- 2:num_pages
    # create list of URLs of XML metadata to scrape
    full_urls <- paste0(url, "&pg=", all_pages)
    xml_dfs <- get_xml_dfs(full_urls)
  }

  xml_dfs <- dplyr::bind_rows(first_page_df, xml_dfs)

  xml_dfs <- xml_dfs %>%
    dplyr::mutate(ProductIssue = as.Date(paste0("01 ", .data$ProductIssue),
      format = "%d %b %Y"
    )) %>%
    dplyr::mutate_at(c("ProductReleaseDate", "SeriesStart", "SeriesEnd"),
      as.Date,
      format = "%d/%m/%Y"
    )

  xml_dfs <- xml_dfs %>%
    dplyr::group_by(.data$TableTitle) %>%
    dplyr::filter(.data$ProductReleaseDate == max(.data$ProductReleaseDate)) %>%
    dplyr::ungroup()

  xml_dfs <- dplyr::mutate(xml_dfs, TableOrder = as.numeric(.data$TableOrder))
  xml_dfs <- dplyr::arrange(xml_dfs, .data$TableOrder)

  xml_dfs
}

get_numpages <- function(url) {
  temp_xml <- tempfile(fileext = ".xml")

  dl_file(
    url = url,
    destfile = temp_xml
  )

  out <- temp_xml %>%
    xml2::read_xml() %>%
    xml2::xml_find_all(xpath = "//NumPages") %>%
    xml2::xml_double()

  if (length(out) == 0) {
    return(1)
  } else {
    return(out)
  }
}

get_first_xml_page <- function(url) {
  get_specific_xml_page(
    url = url,
    page = 1
  )
}

get_last_xml_page <- function(url) {
  num_pages <- get_numpages(url)
  get_specific_xml_page(
    url = url,
    page = num_pages
  )
}

get_specific_xml_page <- function(url, page) {
  if (length(page) > 0) {
    first_url <- paste0(
      url,
      "&pg=",
      page
    )
  } else {
    first_url <- url
  }

  # Some tables in the ABS TSD start with a leading zero, as in
  # Table 01 rather than Table 1; the 0 needs to be included. Here we first test
  # for a readable XML file using the table number supplied (eg. "1"); if that
  # doesn't work then we try with a leading zero ("01").

  safely_get_xml_dfs <- purrr::safely(get_xml_dfs)
  first_page <- safely_get_xml_dfs(first_url)

  first_url_works <- if (!is.null(first_page$result)) {
    TRUE
  } else {
    FALSE
  }

  if (!first_url_works) {
    if (!grepl("ttitle", first_url)) { # this is the case when tables == "all"
      stop(paste0(
        "Cannot find valid entry",
        " in the ABS Time Series Directory.",
        "Check that the cat. no. is correct, and that it contains ",
        "time series spreadsheets (not data cubes)."
      ))
    }

    # now try prepending a 0 on the ttitle

    first_url <- gsub("ttitle=", "ttitle=0", first_url)

    first_page <- safely_get_xml_dfs(first_url)

    first_url_works <- if (!is.null(first_page$result)) {
      TRUE
    } else {
      FALSE
    }

    if (first_url_works) {
      url <- gsub("ttitle=", "ttitle=0", url)
    } else {
      stop(
        "Cannot find valid entry for requested data",
        "in the ABS Time Series Directory"
      )
    }
  }

  first_page <- first_page$result
  first_page
}

# Only used with read_abs_local, where we can't get the titles from
# the Time Series Directory
#' @importFrom readxl read_excel

extract_abs_tabletitles <- function(filename, path) {
  filename <- paste0(path, "/", filename)

  tabletitle <- readxl::read_excel(filename,
    range = "Index!B6:B6",
    col_names = "tabletitle"
  )

  tabletitle <- as.character(tabletitle)

  tabletitle
}

#' Tidy ABS time series data.
#'
#' @param df A data frame containing ABS time series data
#' that has been extracted using \code{extract_abs_sheets}.
#'
#' @param metadata logical. If `TRUE` (the default), a tidy data frame
#' including ABS metadata (series name, table name, etc.) is
#' included in the output. If `FALSE`, metadata is dropped.
#'
#' @return data frame (tibble) in long format.
#'
#' @examples
#'
#' # First extract the data from the local spreadsheet
#' \dontrun{
#' wpi <- extract_abs_sheets("634501.xls")
#' }
#'
#' # Then tidy the data extracted from the spreadsheet. Note that
#' # \code{extract_abs_sheets()} returns a list of data frames, so we need to
#' # subset the list.
#' \dontrun{
#' tidy_wpi <- tidy_abs(wpi[[1]])
#' }
#'
#' @importFrom purrr map_chr
#' @import dplyr
#' @importFrom tidyr gather pivot_longer
#' @export

tidy_abs <- function(df, metadata = TRUE) {
  unit <- series <- value <- X__1 <- series_id <- NULL

  if (!"data.frame" %in% class(df)) {
    stop("Object does not appear to be a data frame; it cannot be tidied.")
  }

  # return an error if the df has <= 1 column
  if (ncol(df) <= 1) {
    stop(
      "The data frame appears to have fewer than 2 columns.",
      " This is unexpected from an ABS time series.",
      " Please check the spreadsheet."
    )
  }

  if (!is.logical(metadata)) {
    stop("`metadata` argument to tidy_abs() must be either `TRUE` or `FALSE`.")
  }

  # newer versions of tibble() and readxl() want to either leave the first
  # colname blank or coerce it to something other than X__1
  # so we set it manually
  colnames(df)[1] <- "X__1"

  # return an error if the sheet is not formatted as we expect
  # from an ABS time series

  if (df[9, 1] != "Series ID") {
    stop(
      "The data frame appears not to be formatted as we expect",
      " from an ABS time series. There should be 9 rows of metadata",
      " after the column names (eg. 'Series ID').",
      " Please check the spreadsheet."
    )
  }

  # a bunch of columns have duplicate names which causes problems; temporarily
  # append the series ID to the colname so they're unique

  # new_col_names <- purrr::map_chr(
  #   .x = 2:ncol(df),
  #   .f = ~ paste0(colnames(df)[.], "_", df[9, .])
  # )

  cols <- 2:ncol(df)
  new_col_names <- paste(colnames(df)[cols], df[9, cols], sep = "_")

  colnames(df)[2:ncol(df)] <- new_col_names

  if (isTRUE(metadata)) {
    df <- df %>%
      tidyr::pivot_longer(cols = !one_of("X__1"), names_to = "series") %>%
      filter(
        !is.na(X__1),
        # This filtering is necessary for cases where the ABS adds notes
        # to the bottom of the data for some reason
        !stringi::stri_detect_fixed(X__1, "Trend Break"),
        !stringi::stri_detect_fixed(X__1, "see")
      ) %>%
      dplyr::group_by(series) %>%
      dplyr::mutate(
        series_type = value[X__1 == "Series Type"],
        data_type = value[X__1 == "Data Type"],
        collection_month = value[X__1 == "Collection Month"],
        frequency = value[X__1 == "Frequency"],
        series_id = value[X__1 == "Series ID"],
        unit = value[X__1 == "Unit"]
      ) %>%
      dplyr::filter(dplyr::row_number() >= 10) %>%
      dplyr::ungroup()

    df <- df %>%
      dplyr::rename(date = X__1) %>%
      dplyr::mutate(
        date = suppressWarnings(as.numeric(date)),
        date = as.Date(date, origin = "1899-12-30"),
        unit = as.character(unit),
        value = as.numeric(value)
      ) %>%
      dplyr::filter(!is.na(date))


    df <- df %>%
      # now remove appended series ID from the end of 'series'
      dplyr::mutate(series = stringi::stri_replace_all_regex(series,
        pattern = "_[^_]+$",
        replacement = ""
      ))
  }

  if (metadata == FALSE) {
    colnames(df) <- df[9, ]
    df <- df[-c(1:9), ]
    colnames(df)[1] <- "date"

    df <- df %>%
      tidyr::gather(
        key = series_id,
        value = value,
        -date
      ) %>%
      dplyr::mutate(
        date = as.Date(as.numeric(date), origin = "1899-12-30"),
        value = as.numeric(value)
      )
  }

  df
}

library(readabs)

local_path <- "../testdata"
local_filename <- "6202021.xls"

test_that("Local file can be read and is in expected form", {
  lfs_21 <- extract_abs_sheets(local_filename, path = local_path)
  lfs_21_df <- lfs_21[[1]]

  expect_is(lfs_21, "list")
  expect_is(lfs_21_df, "data.frame")
  expect_is(lfs_21_df, "tbl")

  expect_equal(as.character(lfs_21_df[1, 1]), "Unit")
  expect_equal(as.character(lfs_21_df[10, 1]), "31017")
})

test_that("Local file can be tidied and looks how we expect", {
  lfs_21 <- extract_abs_sheets(local_filename, path = local_path)
  lfs_21_tidy <- tidy_abs_list(lfs_21)

  expect_length(lfs_21_tidy, 12)
  expect_is(lfs_21_tidy$date, "Date")
  expect_is(lfs_21_tidy$value, "numeric")
  expect_equal(as.character(lfs_21_tidy[1, 1]), "6202021")
})

test_that("Local file can be tidied without metadata", {
  no_meta <- read_abs_local(filenames = local_filename, path = local_path, metadata = FALSE)

  expect_is(no_meta, "data.frame")
  expect_length(no_meta, 6)
})



test_that("tidy_abs_list() gives appropriate errors", {
  expect_error(tidy_abs_list(data.frame(x = 1)))

  expect_error(tidy_abs_list(c(1, 2)))
})

test_that("tidy_abs() gives appropriate errors", {
  expect_error(tidy_abs(df = c(1, 2)))

  expect_error(tidy_abs(data.frame(x = 1)))

  expect_error(tidy_abs(df = data.frame(x = 1, y = 2), metadata = 1))

  expect_error(tidy_abs(df = data.frame(x = c(1:10), y = c(1:10))))
})

test_that("Local file can be read", {
  lfs_21 <- read_abs_local(filenames = local_filename, path = local_path)

  expect_length(lfs_21, 12)
  expect_is(lfs_21$date, "Date")
  expect_is(lfs_21$value, "numeric")
  expect_equal(as.character(lfs_21[1, 1]), "6202021")
  expect_gt(nrow(lfs_21), 951)
})


test_that("read_abs_local() returns appropriate errors and messages", {
  expect_error(read_abs_local(filenames = NULL, path = NULL))

  expect_error(read_abs_local(filenames = 1), path = "../testdata")

  expect_error(read_abs_local(path = "../empty"))

  expect_error(read_abs_local(
    filenames = local_filename, path = local_path,
    metadata = "a"
  ))
})

test_that("read_abs_local() reads an XLSX file", {
  lfs_21_xlsx <- read_abs_local(
    filenames = "6202021.xlsx",
    path = local_path
  )

  expect_s3_class(lfs_21_xlsx, "tbl_df")
  expect_length(lfs_21_xlsx, 12)
})

#' Very slightly faster version of stringr's str_squish()
#' @param string A string to squish (remove whitespace)
#' @keywords internal
fast_str_squish <- function(string) {
  stringi::stri_trim_both(gsub("\\s+", " ", string, perl = TRUE))
}

#' Search for ABS catalogues that match a string
#' @description
#' `r lifecycle::badge("experimental")`
#' Helper function to use with `download_abs_data_cube()`.
#'
#' `download_abs_data_cube()` requires that you specify a `catalogue`.
#' `search_catalogues()` helps you find the catalogue you want, by searching for
#' a given string in the catalogue names, product title, and broad topic.
#'
#' @param string Character. A word or phrase you want to search for, such as "labour" or
#' "union". Not case sensitive.
#' @param refresh Logical. `FALSE` by default. If `TRUE`, will re-scrape the ABS
#' website to ensure that the list of catalogues is up-to-date.
#'
#' @return A data frame (tibble) containing the topic (`heading`), product title
#' (`sub_heading`), catalogue (`catalogue`) and URL (`URL`) of any catalogues
#' that match the provided string.
#' @examples
#'
#' search_catalogues("labour")
#' @export
#' @family data cube functions

search_catalogues <- function(string, refresh = FALSE) {
  stopifnot(!missing(string))

  if (isFALSE(refresh)) {
    df <- abs_lookup_table
  } else {
    df <- scrape_abs_catalogues()
  }

  # Using dplyr::if_any() is fast and clean but requires dplyr >= 1.0.4
  if (getNamespaceVersion("dplyr") > "1.0.4") {
    out <- df %>%
      dplyr::filter(dplyr::if_any(
        .cols = dplyr::everything(),
        ~ grepl(string, .x,
          perl = TRUE, ignore.case = TRUE
        )
      ))
  } else {
    matches <- purrr::map_dfr(df,
      grepl,
      pattern = string, perl = TRUE, ignore.case = TRUE
    ) %>%
      rowSums()

    out <- dplyr::tibble(sum_true = matches) %>%
      dplyr::bind_cols(df) %>%
      dplyr::filter(.data$sum_true >= 1) %>%
      dplyr::select(-"sum_true")
  }

  out <- out %>%
    dplyr::mutate(url = as.character(url))

  return(out)
}

test_that("search_catalogues() returns appropriate output", {
  expect_error(search_catalogues())

  lab <- search_catalogues("labour")
  lab_cap <- search_catalogues("LABOUR")
  lab_norefresh <- search_catalogues("labour", refresh = FALSE)

  expect_identical(lab, lab_norefresh)
  expect_identical(lab, lab_cap)
  expect_is(lab, "tbl")
  expect_length(lab, 4)
  expect_gt(nrow(lab), 20)
  expect_true(all(purrr::map_lgl(lab, ~ inherits(.x, "character"))))

  expect_true(nrow(search_catalogues("foobar")) == 0)
})

test_that("search_catalogues() refreshes table when requested", {
  skip_if_offline()
  skip_on_cran()
  skip_if_not(check_abs_connection())

  ref <- search_catalogues("price", refresh = TRUE)

  expect_is(ref, "tbl")
  expect_length(ref, 4)
  expect_gt(nrow(ref), 4)
})

#' Download and tidy ABS payroll jobs and wages data
#'
#' Import a tidy tibble of ABS Payroll Jobss data.
#'
#' @details The ABS [Payroll Jobs](https://www.abs.gov.au/statistics/labour/jobs/payroll-jobs/latest-release)
#' dataset draws upon data collected
#' by the Australian Taxation Office as part of its Single-Touch Payroll
#' initiative and supplements the monthly Labour Force Survey. Unfortunately,
#' the data as published by the ABS (1) is not in a standard time series
#' spreadsheet; and (2) is messy in various ways that make it hard to
#' read in R. This convenience function uses `download_abs_data_cube()` to
#' import the payrolls data, and then tidies it up.
#'
#' @param series Character. Must be one of:
#' \describe{
#'  \item{"industry_jobs"}{ Payroll jobs by industry division, state, and age
#'  group (Table 1)}
#'  \item{"subindustry_jobs"}{ Payroll jobs by industry sub-division and
#'  industry division (Table 2)}
#'  \item{"empsize_jobs"}{ Payroll jobs by size of employer (number of
#'  employees) and state/territory (Table 3)}
#' }
#' The default is "industry_jobs".
#' @return A tidy (long) `tbl_df`. The number of columns differs based on the `series`.
#'
#' @details
#' Note that this ABS release used to be called Weekly Payroll Jobs and Wages Australia.
#' The total wages series were removed from this release in mid-2023 and it
#' was renamed to Weekly Payroll Jobs. The ability to read total wages
#' indexes using this function was therefore also removed. It was then renamed
#' Payroll Jobs and the frequency was reduced, with further modifications to
#' the data released.
#' @examples
#' \dontrun{
#' # Fetch payroll jobs by industry and state (the default, "industry_jobs")
#' read_payrolls()
#'
#' # Payroll jobs by employer size
#' read_payrolls("empsize_jobs")
#' }
#'
#' @param path Local directory in which downloaded ABS time series
#' spreadsheets should be stored. By default, `path` takes the value set in the
#' environment variable "R_READABS_PATH". If this variable is not set,
#' any files downloaded by read_abs()  will be stored in a temporary directory
#' (\code{tempdir()}).
#'
#' @importFrom rlang .data
#' @export

read_payrolls <- function(series = c(
                            "industry_jobs",
                            "subindustry_jobs",
                            "empsize_jobs"
                          ),
                          path = Sys.getenv("R_READABS_PATH",
                            unset = tempdir()
                          )) {
  check_abs_connection()

  series <- match.arg(series)

  cube_name <- switch(series,
    "industry_jobs" = "DO001",
    "subindustry_jobs" = "DO002",
    "empsize_jobs" = "DO003"
  )

  safely_download_cube <- purrr::safely(.f = ~ download_abs_data_cube(
    catalogue_string = "payroll-jobs",
    cube = cube_name,
    path = path
  ))

  cube_path <- safely_download_cube()

  if (!is.null(cube_path$error)) {
    # Attempt to download requested cube from PREVIOUS payrolls release
    # Necessary because (for the time being) not all tables are included
    # in each release

    attempted_prev_payrolls <- download_previous_payrolls(
      cube_name,
      path
    )

    if (is.null(attempted_prev_payrolls$error)) {
      print("Using table from previous payrolls release")
      cube_path <- attempted_prev_payrolls
    } else {
      stop("Could not download ABS payrolls data.")
    }
  }

  cube_path <- cube_path$result

  sheet_name <- switch(series,
    "industry_jobs" = "Payroll jobs index",
    "sa4_jobs" = "SA4",
    "sa3_jobs" = "SA3",
    "subindustry_jobs" = "Payroll jobs index-Subdivision",
    "empsize_jobs" = "Employment size",
    "gccsa_jobs" = "GCCSA",
    "sex_age_jobs" = "5 year age groups"
  )

  cube <- read_payrolls_local(
    cube_path = cube_path,
    sheet_name = sheet_name
  )

  cube
}

#' @keywords internal
#' @param cube_path Path + filename (incl. extension) to ABS payrolls data cube
#' @param sheet_name Name of the sheet on the Excel cube to import
#' @noRd

read_payrolls_local <- function(cube_path, sheet_name) {
  sheets_present <- readxl::excel_sheets(cube_path)
  sheets_present <- sheets_present[!sheets_present == "Contents"]
  series <- "jobs"

  safely_read_excel <- purrr::safely(read_excel)

  read_attempt <- safely_read_excel(cube_path,
    sheet = sheets_present,
    col_types = "text",
    skip = 5
  )

  if (is.null(read_attempt$error)) {
    cube <- read_attempt$result
  } else {
    stop(
      "Could not find a sheet called '", sheet_name, "' in the payrolls ",
      "workbook ", cube_path, ". Sheets present are: ",
      paste0(sheets_present, collapse = ", ")
    )
  }


  to_snake <- function(x) {
    x <- gsub(" ", "_", x)
    tolower(x)
  }

  cube <- cube %>%
    dplyr::rename_with(.fn = ~ dplyr::case_when(
      .x == "State or Territory" ~ "state",
      .x == "Industry division" ~ "industry",
      .x == "Sub-division" ~ "industry_subdivision",
      .x == "Employment size" ~ "emp_size",
      .x == "Sex" ~ "sex",
      .x == "Age group" ~ "age",
      .x == "Statistical Area 4" ~ "sa4",
      .x == "Statistical Area 3" ~ "sa3",
      TRUE ~ to_snake(.x)
    ))

  cube <- cube[!is.na(cube[[1]]), ]

  cube <- cube %>%
    tidyr::pivot_longer(
      cols = dplyr::starts_with("4"),
      names_to = "date",
      values_to = "value"
    )

  cube <- cube %>%
    # Note that some NA values are hardcoded as "N/A" or similar by the ABS
    dplyr::mutate(value = suppressWarnings(as.numeric(.data$value))) %>%
    dplyr::filter(.data$value != "NA")

  cube <- cube %>%
    dplyr::mutate(date = as.Date(as.numeric(.data$date), origin = "1899-12-30"))

  cube <- cube %>%
    dplyr::mutate_if(
      .predicate = is.character,
      .funs = gsub,
      pattern = ".*\\. ",
      replacement = "",
      perl = TRUE
    )

  cube$series <- "jobs"

  cube
}


#' This function is temporarily necessary while the readabs maintainer
#' attempts to resolve an issue with the ABS. The ABS as at late March 2021
#' stopped including Table 5 of the Weekly Payrolls release with each new
#' release of the data. This function finds the link from the previous
#' release and attemps to download it. This function will no longer be required
#' if/when the ABS reverts to the previous release arrangements. The function
#' is internal and is called by `read_payrolls()`.
#' @param cube_name eg. DO004 for table 4
#' @param path Directory in which to download payrolls cube
#' @keywords internal
#' @return A list containing two elements: `result` (will contain path + filename
#' to downloaded file if download was successful); and `error` (NULL if file
#' downloaded successfully; character otherwise).
download_previous_payrolls <- function(cube_name,
                                       path) {
  latest_payrolls_url <- "https://www.abs.gov.au/statistics/labour/jobs/payroll-jobs/latest-release"
  prev_payrolls_css <- "#release-date-section > div.field.field--name-dynamic-block-fieldnode-previous-releases.field--type-ds.field--label-hidden > div > div > ul > li:nth-child(1) > a"

  temp_page_location <- file.path(tempdir(), "temp_readabs.html")
  dl_file(latest_payrolls_url, temp_page_location)

  prev_payrolls_url <- rvest::read_html(temp_page_location) %>%
    rvest::html_element(prev_payrolls_css) %>%
    rvest::html_attr("href")

  prev_payrolls_url <- paste0("https://www.abs.gov.au/", prev_payrolls_url)

  prev_page_location <- file.path(tempdir(), "temp_readabs_prev.html")
  dl_file(prev_payrolls_url, prev_page_location)

  prev_payrolls_page <- rvest::read_html(prev_page_location)

  prev_payrolls_excel_links <- prev_payrolls_page %>%
    rvest::html_elements(".file--x-office-spreadsheet a") %>%
    rvest::html_attr("href")

  table_link <- prev_payrolls_excel_links[grepl(
    cube_name,
    prev_payrolls_excel_links
  )]

  if (length(table_link) == 0) {
    stop("Could not find URL for requested cube in previous payrolls release")
  } else if (length(table_link) > 1) {
    stop("Found multiple patching URLs for the requested cube in previous payrolls release")
  }

  if (!grepl("abs.gov.au", table_link)) {
    table_link <- paste0("https://www.abs.gov.au", table_link)
  }

  safely_download <- purrr::safely(dl_file)

  full_path <- file.path(path, basename(table_link))

  dl_result <- safely_download(
    url = table_link,
    destfile = full_path
  )

  out <- list(
    result = NULL,
    error = NULL
  )


  if (is.null(dl_result$error)) {
    out$result <- full_path
  } else {
    out$error <- "Could not download payrolls"
  }

  return(out)
}

test_that("read_payrolls() works", {
  skip_if_offline()
  skip_on_cran()
  check_abs_connection()

  check_payrolls <- function(series) {
    print(series)
    df <- read_payrolls(series)
    expect_is(df, "tbl_df")
    expect_is(df$date, "Date")
    expect_is(df$value, "numeric")
    expect_gt(ncol(df), 4)
    expect_false(any(grepl(" ", names(df))))
  }

  purrr::map(
    c(
      "industry_jobs",
      "subindustry_jobs",
      "empsize_jobs"
    ),
    ~ check_payrolls(.x)
  )
})

#' Experimental helper function to download ABS data cubes that are not compatible with read_abs.
#'
#' @description
#' `r lifecycle::badge("experimental")`
#' \code{download_abs_data_cube()} downloads the latest ABS data cubes based on the catalogue name (from the website url) and cube.
#' The function downloads the file to disk.
#'
#' Unlike \code{read_abs()}, this function doesn't import or tidy the data.
#' Convenience functions are provided to import and tidy key data cubes; see
#' \code{?read_payrolls()} and \code{?read_lfs_grossflows()}.
#'
#' @param catalogue_string ABS catalogue name as a string from the ABS website.
#' For example, Labour Force, Australia, Detailed is "labour-force-australia-detailed".
#' The possible catalogues can be obtained using the helper function \code{show_available_catalogues()};
#' or search these catalogues using \code{search_catalogues()},
#'
#' @param cube character. A character string that is either the complete filename or (uniquely) in the filename of the data cube you want to
#' download, e.g. "EQ09". The available filenames can be obtained using the helper function \code{get_available_files()}
#'
#' @param path Local directory in which downloaded files should be stored. By default, `path`
#'  takes the value set in the environment variable "R_READABS_PATH".
#'  If this variable is not set, any files downloaded
#'  will be stored in a temporary directory (\code{tempdir()}).
#'  See \code{Details} below for  more information.
#'
#' @examples
#' \dontrun{
#' download_abs_data_cube(
#'   catalogue_string = "labour-force-australia-detailed",
#'   cube = "EQ09"
#' )
#' }
#'
#' @details `download_abs_data_cube()` downloads an Excel spreadsheet from the ABS.
#'
#' The file need to be saved somewhere on your disk.
#' This local directory can be controlled using the `path` argument to
#' `read_abs()`. If the `path` argument is not set, `read_abs()` will store
#' the files in a directory set in the "R_READABS_PATH" environment variable.
#' If this variable isn't set, files will be saved in a temporary directory.
#'
#' To check the value of the "R_READABS_PATH" variable, run
#' \code{Sys.getenv("R_READABS_PATH")}. You can set the value of this variable
#' for a single session using \code{Sys.setenv(R_READABS_PATH = <path>)}.
#' If you would like to change this variable for all future R sessions, edit
#' your `.Renviron` file and add \code{R_READABS_PATH = <path>} line.
#' The easiest way to edit this file is using \code{usethis::edit_r_environ()}.
#'
#' The filepath is returned invisibly which enables piping to \code{unzip()} or \code{readxl::read_excel}.
#'
#' @importFrom dplyr %>%
#' @importFrom glue glue
#' @importFrom dplyr filter pull slice
#' @importFrom rvest html_nodes html_attr html_text
#' @importFrom httr GET
#'
#' @export
#' @family data cube functions
#'
download_abs_data_cube <- function(catalogue_string,
                                   cube,
                                   path = Sys.getenv("R_READABS_PATH", unset = tempdir())) {
  # check if path is valid
  if (!dir.exists(path)) {
    stop("path does not exist. Please create a folder.")
  }

  check_abs_connection()

  available_cubes <- get_available_files(catalogue_string)

  cube <- if_else(cube == "MRM1",
    "MRM1|MRM%201",
    cube
  )

  file_download_url <- available_cubes %>%
    dplyr::filter(grepl(cube, file, ignore.case = TRUE)) %>%
    dplyr::slice(1) %>% # this gets the first result which is typically the .xlsx file rather than the zip
    dplyr::pull(url)


  # Check that there is a match

  if (length(file_download_url) == 0) {
    stop(glue("No matching cube. Please check against ABS website at {download_url}."))
  }

  # build file path

  filename <- basename(file_download_url)

  filepath <- file.path(path, filename)

  # download file

  dl_file(
    file_download_url,
    filepath
  )

  message("File downloaded in ", filepath)

  return(invisible(filepath))
}

read_lfs_mrm_table <- function(file, sheet, variable_name) {
  df <- file %>%
    readxl::read_excel(
      sheet = sheet,
      skip = 4
    ) %>%
    tidyr::pivot_longer(!tidyr::matches("SA4"), names_to = "date", values_to = "value") %>%
    dplyr::mutate(date = as.numeric(.data$date)) %>%
    dplyr::filter(!is.na({{ variable_name }})) %>%
    dplyr::mutate(SA4_code = substr(.data$SA4, 1, 3)) %>%
    dplyr::mutate(SA4_name = substr(.data$SA4, 5, nchar(.data$SA4))) %>%
    dplyr::mutate(
      variable = variable_name,
      date = as.Date(.data$date, origin = "1899-12-30")
    )
  df[, c("SA4_code", "SA4_name", "variable", "date", "value")]
}

read_lfs_mrm <- function(file) {
  bind_rows(
    read_lfs_mrm_table(file, "Table 1", "employed_persons_000s"),
    read_lfs_mrm_table(file, "Table 2", "unemployed_persons_000s"),
    read_lfs_mrm_table(file, "Table 3", "nilf_persons_000s"),
    read_lfs_mrm_table(file, "Table 4", "emp_to_pop_ratio"),
    read_lfs_mrm_table(file, "Table 5", "unemployment_rate"),
    read_lfs_mrm_table(file, "Table 6", "participation_rate")
  )
}

#' Convenience function to download and tidy data cubes from
#' ABS Labour Force, Australia, Detailed.
#' @param cube character. A character string that is either the complete filename
#' or (uniquely) in the filename of the data cube you want to download. Use
#' `get_available_lfs_cubes()` to see a dataframe of options.
#' @param path Local directory in which downloaded files should be stored.
#' @return A tibble with the data from the data cube. Columns names are
#' tidied and dates are converted to Date class.
#' @examples
#' read_lfs_datacube("EQ02")
#' @export
read_lfs_datacube <- function(cube,
                              path = Sys.getenv("R_READABS_PATH", unset = tempdir())) {
  options(timeout = 180)
  file <- download_abs_data_cube(
    catalogue_string = "labour-force-australia-detailed",
    cube = cube,
    path = path
  )

  if (cube == "MRM" || cube == "MRM1") {
    df <- read_lfs_mrm(file)
  } else {
    df <- file %>%
      readxl::read_excel(
        sheet = "Data 1",
        skip = 3
      ) %>%
      rename(date = 1) %>%
      mutate(date = as.Date(date))

    colnames(df) <- tolower(colnames(df))
    colnames(df) <- gsub(" |-|:", "_", colnames(df))
    colnames(df) <- gsub("\\(|\\)|\\'", "", colnames(df))
  }

  df
}

#' Show the available Labour Force, Australia, detailed data cubes that can be
#' downloaded
#' @export
#' @details Intended to be used with \code{read_lfs_datacube()}. Call
#' \code{read_lfs_datacube()} interactively, find the table of interest
#' (eg. "LM1"), then use `read_lfs_datacube()`.
#' @examples
#'
#' get_available_lfs_cubes()
#'
get_available_lfs_cubes <- function() {
  all_files <- get_available_files("labour-force-australia-detailed")

  all_files %>%
    dplyr::filter(
      substr(file, 1, 1) != "6",
      !grepl("zip", file)
    )
}

test_that("read_lfs_datacube", {
  skip_on_cran()
  avail_cubes <- get_available_lfs_cubes()
  expect_s3_class(avail_cubes, "tbl_df")
  expect_length(avail_cubes, 3)
  expect_gt(nrow(avail_cubes), 1)

  lm1 <- read_lfs_datacube("LM1")
  expect_s3_class(lm1, "tbl_df")
  expect_length(lm1, 10)
  expect_gt(nrow(lm1), 300000)
})

# Test for MRM, which has a different format, and therefore parsing logic
test_that("read_lfs_datacube - MRM", {
  mrm1 <- read_lfs_datacube("MRM1")
  expect_s3_class(mrm1, "tbl_df")
  expect_gt(nrow(mrm1), 30000)
})

test_that("File is downloaded", {
  # Download latest EQ09

  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  temp_path <- tempdir()

  download_abs_data_cube(
    catalogue_string = "labour-force-australia-detailed-quarterly",
    cube = "EQ09",
    path = temp_path
  )


  # Check file exists
  expect_true(file.exists(file.path(temp_path, "EQ09.xlsx")))


  # Check file can be read
  eq09 <- readxl::read_excel(file.path(temp_path, "EQ09.xlsx"),
    range = "A3",
    col_names = FALSE
  ) %>%
    dplyr::pull()
  expect_equal(TRUE, grepl("Released at 11.30 am", eq09))

  unlink(temp_path)
})

#' This function checks to see if the ABS website is available.
#' If available, it invisibly returns `TRUE`. If unavailable, it will
#' stop with an error.
#'
#' Note that it tries first using `httr::HEAD()` via the `url_exists()` function
#' (below). It is given one second to execute `url_exists()`. If this either
#' fails or takes >1 second, an alternative method is tried.
#'
#' We use two methods as (1) some networks do not play well with curl and
#' functions that rely on it, like `HEAD()`, but (2) the second method of
#' checking the website is slower than the first, so we don't want to default
#' to it and slow down all users.
#'
#' @noRd

check_abs_connection <- function() {
  abs_url_works <- url_exists("https://www.abs.gov.au")


  if (isFALSE(abs_url_works)) {
    abs_url_works_nocurl <- url_exists_nocurl("https://www.abs.gov.au")

    if (isFALSE(abs_url_works_nocurl)) {
      stop(
        "R cannot access the ABS website.",
        " Please check your internet connection and security settings."
      )
    }
  }

  invisible(TRUE)
}

# Function from: https://stackoverflow.com/a/52915256
#' Internal function to check if URL exists.
#' @param url URL for website to check
#' @return Logical. `TRUE` if URL exists and returns HTTP status code in the
#' 200 range; `FALSE` otherwise.
#' @noRd

url_exists <- function(url = "https://www.abs.gov.au") {
  sHEAD <- purrr::safely(httr::HEAD)
  sGET <- purrr::safely(httr::GET)

  # Try HEAD first since it's lightweight
  res <- sHEAD(url)

  if (is.null(res$result) ||
    ((httr::status_code(res$result) %/% 200) != 1)) {
    res <- sGET(url)

    if (is.null(res$result)) {
      return(FALSE)
    }

    if (((httr::status_code(res$result) %/% 200) != 1)) {
      warning(sprintf("[%s] appears to be online but isn't responding as expected; HTTP status code is not in the 200-299 range", url))
      return(FALSE)
    }

    return(TRUE)
  } else {
    return(TRUE)
  }
}

#' Internal function to check if URL exists. Slower than url_exists. Used
#' for networks that block curl.
#' @param url URL for website to check
#' does not return expected output).
#' @return Logical. `TRUE` if URL exists and returns HTTP status code in the
#' 200 range; `FALSE` otherwise.
#' @noRd
url_exists_nocurl <- function(url = "https://www.abs.gov.au") {
  con <- url(url,
    method = Sys.getenv("R_READABS_DL_METHOD", unset = "default")
  )
  out <- suppressWarnings(tryCatch(readLines(con), error = function(e) e))
  abs_url_works <- all(class(out) != "error")
  close(con)
  return(abs_url_works)
}

test_that("download_abs() fetches files", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  xml_urls <- form_abs_tsd_url(
    cat_no = "6202.0",
    tables = c("11", "11a"),
    series_id = NULL
  )

  xml_dfs <- purrr::map_dfr(xml_urls,
    .f = get_abs_xml_metadata
  )

  urls <- unique(xml_dfs$TableURL)

  table_titles <- unique(xml_dfs$TableTitle)

  .path <- tempdir()

  # A single file ---
  download_abs(urls[1], .path, FALSE)

  expect_true(file.exists(file.path(.path, basename(urls[1]))))

  unlink(file.path(.path, basename(urls[1])))

  # Multiple files using purrr::walk()
  purrr::walk(urls,
    download_abs,
    path = .path,
    show_progress_bars = F
  )

  expect_true(all(file.exists(file.path(.path, basename(urls)))))

  unlink(.path)

  # Vectorised download.file ----
  download_abs(urls, .path)

  expect_true(all(file.exists(file.path(.path, basename(urls)))))

  unlink(.path)
})

test_that("read_abs_url() behaves as expected", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  my_url <- "https://www.abs.gov.au/AUSSTATS/ABS@Archive.nsf/log?openagent&136401500301.xls&1364.0.15.003&Time%20Series%20Spreadsheet&4E7EEA9838FD6744CA2579340011C2D1&0&Jun%20Qtr%202011&26.10.2011&Latest"

  # Test it works with a single URL
  expect_s3_class(
    read_abs_url(my_url),
    "tbl_df"
  )

  # Test the file exists in the expected location after download
  my_path <- file.path(tempdir(), "read_abs_local_test_dir")

  rm(list = list.files(path = my_path))

  read_abs_url(my_url, path = my_path)

  expect_true(file.exists(file.path(my_path, "136401500301.xls")))

  # Test multiple URLs work
  multiple_urls <- read_abs_url(c(
    "https://www.abs.gov.au/AUSSTATS/ABS@Archive.nsf/log?openagent&136401500301.xls&1364.0.15.003&Time%20Series%20Spreadsheet&4E7EEA9838FD6744CA2579340011C2D1&0&Jun%20Qtr%202011&26.10.2011&Latest",
    "https://www.abs.gov.au/statistics/labour/employment-and-unemployment/labour-force-australia/aug-2022/6202001.xlsx"
  ))


  expect_s3_class(multiple_urls, "tbl_df")
  expect_gt(nrow(multiple_urls), 60000)
})

test_that("check_latest_date() returns expected output", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(check_abs_connection())

  test_latest_date <- function(result) {
    print(class(result))
    expect_is(result, "Date")
    expect_length(result, 1)
    expect_gt(result, Sys.Date() - 180)
  }

  whole_cat <- check_latest_date("6345.0")
  single_table <- check_latest_date("6202.0", tables = 1)
  two_tables <- check_latest_date("6345.0", tables = c("1", "5a"))
  single_series <- check_latest_date(series_id = "A2713849C")

  purrr::map(
    c(
      whole_cat,
      single_table,
      two_tables,
      single_series
    ),
    test_latest_date
  )
})

test_that("individual steps of read_abs() work", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  # Set parameters ---
  cat_no <- "6202.0"
  tables <- c("11", "11a")
  series_id <- NULL
  .path <- tempdir()
  metadata <- TRUE
  show_progress_bars <- FALSE
  retain_files <- FALSE
  check_local <- FALSE


  # Form time series URLS ----
  xml_urls <- form_abs_tsd_url(
    cat_no = cat_no,
    tables = tables,
    series_id = series_id
  )

  expect_length(xml_urls, 2)
  expect_type(xml_urls, "character")

  temp_xml_1 <- tempfile(fileext = ".xml")
  temp_xml_2 <- tempfile(fileext = ".xml")

  purrr::walk2(
    .x = xml_urls,
    .y = c(
      temp_xml_1,
      temp_xml_2
    ),
    .f = utils::download.file,
    headers = readabs_header
  )

  expect_true(file.exists(temp_xml_1))
  expect_true(file.exists(temp_xml_2))

  xml_1 <- xml2::read_xml(temp_xml_1)

  expect_s3_class(xml_1, "xml_document")

  product_title <- xml_1 %>%
    xml2::as_list() %>%
    .$TimeSeriesIndex %>%
    .$Series %>%
    .$ProductTitle %>%
    as.character()

  expect_identical(
    product_title,
    "Labour Force, Australia"
  )

  unlink(temp_xml_1)
  unlink(temp_xml_2)

  # Get XML metadata -----

  xml_dfs <- purrr::map_dfr(xml_urls,
    .f = get_abs_xml_metadata
  )

  expect_s3_class(xml_dfs, "data.frame")

  # Download files ----

  urls <- unique(xml_dfs$TableURL)

  table_titles <- unique(xml_dfs$TableTitle)

  purrr::walk(urls, download_abs,
    path = .path,
    show_progress_bars = show_progress_bars
  )


  expect_true(all(file.exists(file.path(.path, basename(urls)))))


  # Extract data sheets from files -----
  sheets <- purrr::map2(basename(urls),
    table_titles,
    .f = extract_abs_sheets,
    path = .path
  )

  expect_length(sheets, 2)
  expect_type(sheets, "list")
  expect_s3_class(sheets[[1]][[1]], "data.frame")

  # Tidy data sheets ----
  sheets <- unlist(sheets, recursive = FALSE)

  # tidy the sheets
  sheet <- tidy_abs_list(sheets, metadata = metadata)

  expect_length(sheet, 12)
  expect_gt(nrow(sheet), 42923)
  expect_s3_class(sheet, "tbl_df")
})

test_that("show_available_catalogues() returns expected vector", {
  cats <- show_available_catalogues()

  expect_type(cats, "character")
  expect_gt(length(cats), 200)
  expect_equal(sort(cats), sort(abs_lookup_table$catalogue))

  filtered_cats <- show_available_catalogues(selected_heading = "Price indexes and inflation")

  expect_type(filtered_cats, "character")
  expect_gt(length(filtered_cats), 5)
})

test_that("show_available_catalogues(refresh = TRUE) returns tibble", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  cats <- show_available_catalogues(refresh = TRUE)

  expect_type(cats, "character")
  expect_gt(length(cats), 200)
  expect_equal(
    sort(cats),
    sort(abs_lookup_table$catalogue)
  )
})

test_that("read_job_mobility() works as expected", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  test_mobility_df <- function(df) {
    expect_length(df, 12)
    expect_gt(nrow(df), 0)
    expect_identical(
      names(df),
      c(
        "table_no",
        "sheet_no",
        "table_title",
        "date",
        "series",
        "value",
        "series_type",
        "data_type",
        "collection_month",
        "frequency",
        "series_id",
        "unit"
      )
    )
    expect_is(df$value, "numeric")
    expect_is(df$date, "Date")
  }

  test_mobility_df(read_job_mobility(1))
  test_mobility_df(read_job_mobility())
})

#' Download and tidy ABS Job Mobility tables
#'
#' @description Import a tidy tibble of ABS Job Mobility data
#' @param tables Either `"all"` (the default) to import all tables, or a
#' vector of table numbers, such as `1` or `c(2, 4)`.
#' @param path Local directory in which downloaded ABS time series spreadsheets should be stored. By default, 'path' takes the value set in the environment variable "R_READABS_PATH". If this variable is not set, any files downloaded by read_abs() will be stored in a temporary directory (tempdir()).
#' @export
#' @examples
#' \dontrun{
#' # Get all tables from the ABS Job Mobility series
#' read_job_mobility()
#'
#' # Get tables 1 and 2
#' read_job_mobility(c(1, 2))
#' }
#'
read_job_mobility <- function(tables = "all",
                              path = Sys.getenv("R_READABS_PATH",
                                unset = tempdir()
                              )) {
  check_abs_connection()

  raw_job_mob_tbls <- show_available_files("job-mobility")

  job_mob_tbls <- raw_job_mob_tbls %>%
    mutate(
      base_file_name = tools::file_path_sans_ext(file),
      label = gsub("62230_", "", .data$base_file_name)
    )

  available_tables <- job_mob_tbls %>%
    dplyr::filter(base::substr(.data$label, 1, 5) == "Table") %>%
    dplyr::filter(tools::file_ext(file) == "xlsx")

  if (tables == "all") {
    selected_tables <- available_tables
  } else {
    selected_tables <- available_tables %>%
      dplyr::filter(grepl(
        paste0(tables, collapse = "|"),
        .data$label
      ))
  }

  cubes <- selected_tables$file

  safely_download_cube <-
    purrr::safely(
      .f = download_abs_data_cube
    )

  cube_results <- purrr::map(cubes,
    safely_download_cube,
    catalogue_string = "job-mobility",
    path = path
  )

  cube_results <- purrr::set_names(cube_results, cubes)

  get_result <- function(result_list, name) {
    if (!is.null(result_list$error)) {
      stop("Could not download table ", name)
    }

    result_list$result
  }

  cube_paths <- purrr::map2_chr(
    cube_results,
    cubes,
    get_result
  )


  read_abs_local(filenames = cube_paths)
}

# Function to create URL(s) to search the ABS Time Series Directory
# given an ABS catalogue number or a series ID

form_abs_tsd_url <- function(cat_no = NULL,
                             tables = "all",
                             series_id = NULL) {
  if (is.null(cat_no) & is.null(series_id)) {
    stop("Must specify either `cat_no` or `series_id`")
  }

  # create the url to search for in the Time Series Directory
  base_url <- "https://ausstats.abs.gov.au/servlet/TSSearchServlet?"

  if (!is.null(series_id)) {
    if (!is.character(series_id)) {
      stop("If you specify `series_id`, it must be a string.")
    }

    xml_urls <- paste0(base_url, "sid=", series_id)
  } else {
    if (tables[1] == "all") {
      tables_url <- ""
    } else {
      tables_url <- paste0("&ttitle=", tables)
    }

    xml_urls <- paste0(
      base_url,
      "catno=", cat_no,
      tables_url
    )
  }


  xml_urls
}

test_match_tables <- function(cat_no, tables) {
  url <- form_abs_tsd_url(cat_no, tables)
  xml_df <- purrr::map_dfr(url, get_abs_xml_metadata)
  matching <- match_tables(xml_df$TableTitle, tables)
  unique(xml_df$TableTitle[matching])
}

test_that("match_tables() returns only requested tables", {
  skip_on_cran()

  t1 <- test_match_tables("5368.0", "1")
  expect_length(t1, 1)
  expect_true(grepl("TABLE 1.", t1))

  t2 <- test_match_tables("5368.0", c("1", "2"))
  expect_length(t2, 2)
  expect_true(all(
    grepl("TABLE 1.|TABLE 2.", t2)
  ))

  t3 <- test_match_tables("5368.0", c("1", "12a"))
  expect_length(t3, 2)

  t6 <- test_match_tables("6291.0.55.001", "1")
  expect_length(t6, 1)

  t7 <- test_match_tables(
    "6202.0",
    c("1", "4", "10a", "11", "11a", "12", "12a", "21")
  )
  expect_length(t7, 8)

  t8 <- test_match_tables("5368.0", c("12a", "13a"))
  expect_length(t8, 2)
  expect_true(all(grepl("TABLE 12a|TABLE 13a", t8)))
})

#' Read and tidy locally-saved ABS time series spreadsheet(s)
#'
#' If you need to download and tidy time series data from the ABS,
#' use \code{read_abs()}. \code{read_abs_local()} imports and tidies data
#' from ABS time series spreadsheets that are already saved to your local drive.
#'
#' Unlike \code{read_abs()}, the `table_title` column in the data frame
#' returned by \code{read_abs_local()} is blank. If you require `table_title`,
#' please use \code{read_abs()} instead.
#'
#' @param cat_no character; a single catalogue number such as "6202.0".
#' When `cat_no` is specified, all local files in `path` corresponding to
#' the specified catalogue number will be imported.
#' For example, if you run `read_abs_local("6202.0")`, it will look in
#' the `6202.0` sub-folder of `path` and attempt to load any
#' .xls and .xlsx files in that location.
#' If `cat_no`` is specified, `filenames` will be ignored.
#'
#' @param filenames character vector of at least one filename of a
#' locally-stored ABS time series spreadsheet. For example, "6202001.xls" or
#' c("6202001.xls", "6202005.xls"). Ignored if a value is supplied to `cat_no`.
#' If `filenames` is blank and `cat_no` is blank, `read_abs_local()` will
#' attempt to read all .xls and .xlsx files in the directory specified with `path`.
#'
#' @param path path to local directory containing ABS time series file(s).
#' Default is `Sys.getenv("R_READABS_PATH", unset = tempdir())`.
#'  If nothing is specified in `filenames` or `cat_no`,
#' `read_abs_local()` will attempt to read all .xls and .xlsx files in the directory
#' specified with `path`.
#'
#' @param use_fst logical. If `TRUE` (the default) then, if an `fst` file of the
#' tidy data frame has already been saved in `path`, it is read immediately.
#'
#' @param metadata logical. If `TRUE` (the default), a tidy data frame including
#' ABS metadata (series name, table name, etc.) is included in the output. If
#' `FALSE`, metadata is dropped.
#'
#' @examples
#'
#' # Load and tidy two specified files from the "data/ABS" subdirectory
#' # of your working directory
#' \dontrun{
#' lfs <- read_abs_local(c("6202001.xls", "6202005.xls"))
#' }
#'
#' @export
#'

read_abs_local <- function(cat_no = NULL,
                           filenames = NULL,
                           path = Sys.getenv("R_READABS_PATH",
                             unset = tempdir()
                           ),
                           use_fst = TRUE,
                           metadata = TRUE) {
  # Error catching
  if (is.null(filenames) && is.null(path)) {
    stop("You must specify a value to filenames and/or path.")
  }

  if (!is.null(filenames) && !is.character(filenames)) {
    stop("if a value is given to `filenames`, it must be specified as a character vector, such as '6202001.xls' or c('6202001.xls', '6202005.xls')")
  }

  if (is.null(path)) {
    warning(paste0(
      "`path` not specified.",
      "\nLooking for ABS time series files in ",
      getwd()
    ))
  }

  if (!is.null(cat_no) && !is.character(cat_no)) {
    stop("If `cat_no` is specified, it must be a string such as '6202.0'")
  }

  if (!is.logical(metadata) || length(metadata) != 1L || is.na(metadata)) {
    stop("`metadata` argument must be either TRUE or FALSE")
  }

  # Retrieve cache if available
  if (is.null(filenames) && isTRUE(use_fst) && fst_available(cat_no, path)) {
    out <- fst::read_fst(path = catno2fst(cat_no = cat_no, path = path))
    return(dplyr::as_tibble(out))
  }

  # If catalogue number is specifid, that takes precedence
  if (!is.null(cat_no)) {
    path <- file.path(path, cat_no)

    filenames <- list.files(path = path, pattern = "\\.(xls|xlsx)$")

    if (length(filenames) == 0) {
      stop(paste0("Could not find any .xls or .xlsx files in path: '", path, "'"))
    }

    # If cat_no isn't specified and filenames isn't specified,
    # get all files in path
  } else if (is.null(filenames)) {
    message(paste0(
      "`filenames` and `cat_no` not specified. ",
      "Looking for ABS time series files in '",
      path,
      "' and attempting to read all files.",
      "\nSpecify `cat_no` or filenames` if you do not wish to",
      " attempt to import all .xls(x) files in ",
      path
    ))

    # Get a list of all filenames in path, as `filenames` is null

    filenames <- list.files(path = path, pattern = "\\.(xls|xlsx)$")

    if (length(filenames) == 0) {
      stop(paste0("Could not find any .xls or .xlsx files in path: '", path, "'"))
    }
  }

  # Create filenames for local ABS time series files

  filenames <- base::basename(filenames)

  # Import data from local spreadsheet(s) to a list
  message("Extracting data from locally-saved spreadsheets")
  # first extract table titles from the time series spreadsheet index
  table_titles <- purrr::map(filenames, extract_abs_tabletitles, path = path)
  # then read in the data
  sheets <- purrr::map2(filenames, table_titles,
    .f = extract_abs_sheets, path = path
  )

  # remove one 'layer' of the list,
  # so that each sheet is its own element in the list
  sheets <- unlist(sheets, recursive = FALSE)

  # tidy the sheets
  sheet <- tidy_abs_list(sheets, metadata = metadata)

  sheet
}

#' Download and import an ABS time series spreadsheet from a given URL
#'
#' @details If you have a specific URL to the time series spreadsheet you wish
#' to download, `read_abs_url()` will download, import and tidy it. This is
#' useful for older vintages of data, or discontinued data.
#'
#' @param url Character vector of url(s) to ABS time series spreadsheet(s).
#' @param path Local directory in which downloaded ABS time series
#' spreadsheets should be stored. By default, `path` takes the value set in the
#' environment variable "R_READABS_PATH". If this variable is not set,
#' any files downloaded by read_abs()  will be stored in a temporary directory
#' (\code{tempdir()}). See \code{?read_abs()} for more.
#' @param show_progress_bars TRUE by default. If set to FALSE, progress bars
#' will not be shown when ABS spreadsheets are downloading.
#' @param ... Additional arguments passed to `read_abs_local()`.
#'
#' @examples
#' \dontrun{
#' url <- paste0(
#'   "https://www.abs.gov.au/statistics/labour/",
#'   "employment-and-unemployment/labour-force-australia/aug-2022/6202001.xlsx"
#' )
#' read_abs_url(url)
#' }
#' @export
read_abs_url <- function(url,
                         path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
                         show_progress_bars = TRUE,
                         ...) {
  get_redirect_url <- function(url) {
    header <- httr::HEAD(url)
    header$url
  }

  real_urls <- purrr::map_chr(url, get_redirect_url)

  files <- file.path(path, basename(real_urls))

  download_abs(
    real_urls,
    path
  )

  read_abs_local(
    filenames = files,
    ...
  )
}

# Documentation: https://api.gov.au/assets/APIs/abs/DataAPI.openapi.html#/Get%20Data/GetData
# More documentation: https://www.abs.gov.au/about/data-services/application-programming-interfaces-apis/data-api-user-guide/using-api
# Online data viewer: https://dataexplorer.abs.gov.au

#' ABS.Stat API functions
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' These experimental functions provide a minimal interface to the ABS.Stat API.
#'
#' More information on the ABS.Stat API can be found
#' on the [ABS website](https://www.abs.gov.au/about/data-services/application-programming-interfaces-apis/data-api-user-guide/using-api)
#'
#' Note that an ABS.Stat 'dataflow' is like a table. A 'datastructure'
#' contains metadata that describes the variables in the dataflow. To load data
#' from the ABS.Stat API, you need to either:
#'
#'   - Using `read_api_dataflows()` you can get information on the available dataflows
#'   - Using `read_api_datastructure()` you can get metadata relating to a
#'   specific dataflow, including the variables available in each dataflow
#'   - Using `read_api()` you can get the data belonging to a given dataflow.
#'   - Using `read_api_url()` you can get the data for a given query url
#'   generated using the [online data viewer](https://dataexplorer.abs.gov.au).
#'
#' @param id A dataflow id. Use `read_api_dataflows()` to obtain a dataframe
#'   listing available dataflows.
#' @param datakey A named list matching filter variables to codes. All variables
#'   with a `position` in the datastructure are filterable. Use
#'   `read_api_datastructure()` to obtain information about the variables in
#'   a dataflow and the values of that variable.
#' @param start_period The start period (used to filter by time). This is
#'   inclusive. The supported formats are:
#'
#'   - `"YYYY"` for annual data (e.g. 2019)
#'   - `"YYYY-S[1-2]"` for semi-annual data (e.g. 2019-S1)
#'   - `"YYYY-Q[1-4]"` for quarterly data (e.g. 2019-Q1)
#'   - `"YYYY-MM[01-12]"` for monthly data (e.g. 2019-01)
#'   - `"YYYY-W[01-53]"` for weekly data (e.g. 2019-W01)
#'   - `"YYYY-MM-DD"` for daily and business data (e.g. 2019-01-01)
#' @param end_period The end period (used to filter on time). This is inclusive.
#'   The supported formats are the same as for `start_period`
#' @param version A version number, if unspecified the latest version of the
#'   dataset is used. Use `read_api_dataflows()` to see
#'   available dataflow versions.
#' @param url A complete query url
#'
#' @details
#' Note that the API enforces a reasonably strict gateway timeout policy. This
#' means that, if you're trying to access a reasonably large dataset, you will
#' need to filter it on the server side using the `datakey`. You might like to
#' review the data manually via the [ABS website](https://dataexplorer.abs.gov.au)
#' to figure out what subset of the data you require.
#'
#' Note, furthermore, that the datastructure contains a complete codebook for
#' the variables appearing in the relevant dataflow. Since some variables are
#' shared across multiple dataflows, this means that the datastructure
#' corresponding to a particular `id` may contain values for a given variable
#' which are not in the corresponding dataflow.
#'
#' @return A data.frame
#'
#' @examples
#' \dontrun{
#' # List available dataflows
#' read_api_dataflows()
#'
#' # Say we want the "Estimated resident population, Country of birth"
#' # data flow, with the id ERP_COB. We load the data like this:
#' # Get full data set for a given flow by providing id and start period:
#' read_api("ERP_COB", start_period = 2020)
#'
#' # In some cases, loading a whole dataflow (as above) won't work.
#' # For eg., the `ABS_C16_T10_SA` dataflow is very large,
#' # so the gateway will timeout if we try to collect the full data set
#' try(read_api("ABS_C16_T10_SA"))
#'
#' # We need to filter the dataflow before downlaoding it.
#' # To figure out how to filter it, we get metadata ('datastructure').
#' ds <- read_api_datastructure("ABS_C16_T10_SA")
#'
#' # The `asgs_2016` code for 'Australia' is 0
#' ds[ds$var == "asgs_2016" & ds$label == "Australia", ]
#'
#' # The `sex_abs` code for 'Persons' (i.e. all persons) is 3
#' ds[ds$var == "sex_abs" & ds$label == "Persons", ]
#'
#' # So we have:
#' x <- read_api("ABS_C16_T10_SA", datakey = list(asgs_2016 = 0, sex_abs = 3))
#' unique(x["asgs_2016"]) # Confirming only 'Australia' level records came through
#' unique(x["sex_abs"]) # Confirming only 'Persons' level records came through
#'
#' # Please note however that not all values in the datastructure necessarily
#' # appear in the data. You get 404s in this case
#' ds[ds$var == "regiontype" & ds$label == "Destination Zones", ]
#' try(read_api("ABS_C16_T10_SA", datakey = list(regiontype = "DZN")))
#'
#' # If you already have a query url, then use `read_api_url()`
#' wpi_url <- "https://data.api.abs.gov.au/rest/data/ABS,WPI/all"
#' read_api_url(wpi_url)
#' }
#' @name abs_api
NULL

#' @export
#' @rdname abs_api
read_api_dataflows <- function() {
  check_abs_connection()
  r <- httr::GET(abs_api_url("rest/dataflow/ABS"))
  r <- httr::content(r)
  out <- purrr::map_dfr(r$references, ~ .[c("id", "name", "version")])
  names(out) <- c("id", "name", "version")

  # Not all of them have descriptions
  desc <- purrr::map(r$references, "description")
  desc <- purrr::map_chr(desc, ~ ifelse(is.null(.), "", .))
  out$desc <- desc

  out$version <- numeric_version(out$version)
  out[c("id", "name", "desc", "version")]
}


#' @export
#' @rdname abs_api
read_api <- function(id, datakey = NULL, start_period = NULL, end_period = NULL, version = NULL) {
  check_abs_connection()

  # Fetch datastructure
  datastructure <- read_api_datastructure(id)
  datastructure <- datastructure[!is.na(datastructure$code), ]

  # Build data query
  dataflow <- paste0("ABS,", id)
  if (!is.null(version)) dataflow <- paste(dataflow, version, sep = ",")
  q <- c("startPeriod" = start_period, "endPeriod" = end_period)
  if (!is.null(datakey)) {
    k <- abs_api_match_key(datakey, datastructure)
  } else {
    k <- "all"
  }

  url <- abs_api_url(c("rest", "data", dataflow, k), q)

  # Fetch data
  df <- abs_api_fetch_data(url)
  abs_api_label_data(df, datastructure)
}

#' @export
#' @rdname abs_api
read_api_url <- function(url) {
  check_abs_connection()

  # Fetch datastructure
  id <- abs_api_id_from_url(url)
  datastructure <- read_api_datastructure(id)
  datastructure <- datastructure[!is.na(datastructure$code), ]

  # Fetch data
  df <- abs_api_fetch_data(url)
  abs_api_label_data(df, datastructure)
}


#' @export
#' @rdname abs_api
read_api_datastructure <- function(id) {
  check_abs_connection()

  r <- httr::GET(abs_api_url(c("rest", "dataflow", "ABS", id, "?references=all")), httr::accept_xml())
  httr::stop_for_status(r)
  r <- httr::content(r)

  r <- xml2::as_list(r)

  codelists <- r$Structure$Structures$Codelists
  names(codelists) <- purrr::map_chr(r$Structure$Structures$Codelists, ~ attr(., "id"))
  components <- r$Structure$Structures$DataStructures$DataStructure$DataStructureComponents

  codelists <- purrr::imap_dfr(codelists, function(codelist, nm) {
    desc <- codelist$Name[[1]]
    purrr::map_dfr(codelist, ~ c(local_id = nm, desc = desc, code = attr(., "id"), label = .$Name))
  })
  codelists <- codelists[!is.na(codelists$code), ]

  components <- purrr::imap_dfr(components, function(x, nm) {
    nm <- tolower(gsub("List", "", nm))
    purrr::map_dfr(x, ~ c(
      role = nm,
      var = attr(., "id"),
      local_id = attr(.$LocalRepresentation$Enumeration$Ref, "id"),
      position = attr(., "position")
    ))
  })

  out <- merge(components, codelists, by = "local_id", all.x = TRUE, all.y = TRUE)
  out <- out[order(out$code), ]
  out <- out[order(out$position), setdiff(names(out), "local_id")]
  rownames(out) <- NULL

  out$var <- tolower(out$var)
  dplyr::tibble(out)
}


# Internal ---------------------------------------------------------------------

#' Construct an endpoint url
#'
#' @param path A character vector, parts of the url path to be concatenated
#'   together separated by "/"
#' @param query A named character vector, parts of the query string to be
#'   concatenated together as 'name=value' separated by '&'
#'
#' @return A string
#' @noRd
#' @keywords internal
#'
#' @examples
#'
#' abs_api_url(c("a", "path"), query = c(nulls = NULL, get = "dropped"))
abs_api_url <- function(path, query = NULL) {
  out <- paste0("https://data.api.abs.gov.au/", paste(path, collapse = "/"))
  if (!is.null(query)) {
    query <- paste(names(query), query, sep = "=")
    out <- paste0(out, "?", paste(query, collapse = "&"))
  }

  out
}

#' Extract dataflow id from url string
#'
#' @param url A url string
#'
#' @return A string
#' @noRd
#' @keywords internal
#'
abs_api_id_from_url <- function(url) {
  stopifnot("`url` must be of length 1" = length(url) == 1)
  if (!grepl("^https://data.api.abs.gov.au/rest/data/ABS,", url)) {
    stop("`url` is not an ABS query url. Query urls must match regex: \n\t",
         "'^https://data.api.abs.gov.au/rest/data/ABS,.*'",
         call. = FALSE
    )
  }
  id <- strsplit(url, "/")[[1]][6]
  id <- strsplit(id, ",")[[1]][2]
  id
}

#' Construct a datakey to filter an ABS.Stat dataflow
#'
#' @param datakey A named list matching variables to codes
#' @param datastructure A datastructure retrieved with `read_api_datastructure()`
#'
#' @return A datakey string
#' @noRd
#' @keywords internal
#'
#' @examples
#' z <- read_api_datastructure("ERP_COB")
#' abs_api_match_key(list(SEX = 1:3, REGION = "AUS"), z)
#' abs_api_match_key(list(SEX = 1:10, REGION = "AUS"), z)
#' try(abs_api_match_key(list(SX = 1:3, REGION = "AUS"), z))
#' try(abs_api_match_key(list(UNIT_MEASURE = "AUD", REGION = "AUS"), z))
abs_api_match_key <- function(datakey, datastructure) {
  stopifnot("`datakey` must be a named list" = is.list(datakey))
  stopifnot("`datakey` must be a named list" = length(names(datakey)) == length(datakey))

  bad_vars <- setdiff(names(datakey), datastructure$var)
  if (length(bad_vars) > 0) {
    stop(
      call. = FALSE,
      "Variable(s) `", paste(bad_vars, collapse = "`, `"),
      "` not found. Please review the datastructure with `read_api_datastructure()`"
    )
  }

  key <- purrr::imap_dfr(datakey, function(vals, var) {
    ds <- datastructure[datastructure$var %in% var, ]

    bad_vals <- setdiff(vals, ds$code)
    if (length(bad_vals) > 0) {
      if (length(bad_vals) > 5) bad_vals <- c(bad_vals[1:5], "...")
      warning(
        call. = FALSE,
        "Variable `", var, "` does not contain codes: ",
        paste(bad_vals, collapse = ", ")
      )
    }

    if (anyNA(ds$position)) {
      stop(
        call. = FALSE,
        "Cannot filter on `", var,
        "`. Please review the datastructure with `read_api_datastructure()`"
      )
    }

    ds[ds$code %in% vals, ]
  })


  pos <- unique(datastructure$position)
  pos <- sort(pos)

  key <- purrr::map_chr(pos, function(pos) {
    keyvals <- key[key$position == pos, ]
    paste0(paste(keyvals$code, collapse = "+"), ".")
  })

  paste(key, collapse = "")
}

#' Fetch data
#'
#' @param url An API data endpoint
#'
#' @return A data.frame
#' @noRd
#' @keywords internal
#'
abs_api_fetch_data <- function(url) {
  as_csv <- httr::accept("application/vnd.sdmx.data+csv")
  r <- httr::GET(url, as_csv)
  httr::stop_for_status(r)

  r <- utils::read.csv(
    text = rawToChar(httr::content(r)),
    colClasses = "character"
  )
  r <- r[setdiff(names(r), "DATAFLOW")]
  names(r) <- tolower(names(r))

  r
}

#' Label fetched data using datastructure
#'
#' @param df A data.frame of data fetched with `abs_api_fetch_data`
#' @param datastructure A data.structure fetched with `read_api_datastructure`
#'
#' @return A labelled data.frame
#' @noRd
#' @keywords internal
abs_api_label_data <- function(df, datastructure) {
  df <- purrr::imap_dfc(df, function(x, var_name) {
    codes <- datastructure[datastructure$var == var_name, ]
    if (nrow(codes) == 0) {
      # For columns not in codes attempt to coerce to numeric: obs_value & time_period
      if (can_numeric(x)) {
        return(as.numeric(x))
      } else {
        return(x)
      }
    }
    labs <- codes$code


    names(labs) <- codes$label
    labelled::labelled(
      x = x,
      labels = labs,
      label = unique(codes$desc)
    )
  })

  df
}

#' Can something be safely coerced to numeric?
#'
#' @param x A vector
#'
#' @return `TRUE` or `FALSE`
#' @noRd
#' @keywords internal
can_numeric <- function(x) {
  if (is.numeric(x) || is.logical(x) || is.factor(x)) {
    TRUE
  } else {
    !any(!(x %in% c(NA, "")) & is.na(suppressWarnings(as.numeric(x))))
  }
}

test_that("read_api helper functions return tibbles with expected structure", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  res <- read_api_dataflows()
  expect_s3_class(res, "tbl_df")
  expect_named(res, c("id", "name", "desc", "version"))

  res <- read_api_datastructure(sample(res$id, 1)) # Using a random dataflow
  expect_s3_class(res, "tbl_df")
  expect_named(res, c("role", "var", "position", "desc", "code", "label"))
})

test_that("read_api throws expected errors and warnings", {
  skip_on_cran()
  check_abs_connection()

  expect_error(read_api("doesnt_exist"), class = "http_404")
  expect_error(read_api_datastructure("doesnt_exist"), class = "http_404")

  expect_error(
    read_api("ERP_COB", datakey = list(not_a_var = 1), start_period = 2020),
    "Variable\\(s\\) .+ not found"
  )
  expect_error(
    read_api("ERP_COB", datakey = list(unit_measure = "PSNS"), start_period = 2020),
    "Cannot filter on .*"
  )
  expect_warning(
    read_api("ERP_COB", datakey = list(age = -99:-90), start_period = 2020),
    "`age` does not contain codes: -99, -98, -97, -96, -95, \\.\\.\\."
  )

  expect_error(
    read_api_url(c("www.google.com", "www.google.com")),
    "must be of length 1"
  )
  expect_error(
    read_api_url("www.google.com"),
    ".* is not an ABS query ur"
  )
})

test_that("read_api filtering works as expected", {
  skip_on_cran()
  check_abs_connection()

  x <- read_api("ABS_C16_T10_SA", datakey = list(asgs_2016 = 0, sex_abs = 3))
  expect_equal(as.integer(unique(x["asgs_2016"])), 0)
  expect_equal(as.integer(unique(x["sex_abs"])), 3)
})

test_that("url queries work", {
  skip_on_cran()
  check_abs_connection()

  wpi_url <- "https://data.api.abs.gov.au/rest/data/ABS,WPI/all"
  res <- read_api_url(wpi_url) %>%
    arrange(measure, index, sector, industry, tsest, region, freq, time_period)

  expect_s3_class(res, "tbl_df")
  expect_true(labelled::is.labelled(res$measure))
  expect_length(res, 13)
  expect_gt(nrow(res), 142)
})

#' Download, import and tidy 'gross flows' data cube
#' from the monthly ABS Labour Force survey.
#'
#' This convenience function downloads, imports and tidies the 'gross flows' data
#' cube from the monthly ABS Labour Force survey. The gross flows data cube (GM1)
#' shows estimates of the number of people who
#' transitioned from one labour force status to another between two months.
#'
#' @param weights either `"current"` or `"previous"`. If `"current"`, figures will
#' use the current month's Labour Force survey weights; if `"previous"`, the
#' previous month's weights are used.
#' @param path Local directory in which downloaded files should be stored.
#' By default, 'path' takes the value set in the environment variable
#' "R_READABS_PATH". If this variable is not set, any files downloaded
#' will be stored in a temporary directory (`tempdir()`).
#' See `Details` in \code{?read_abs} for more information.
#' @return A tibble containing data cube GM1 from the monthly Labour Force survey.
#' @export
#' @examples
#' \dontrun{
#' read_lfs_grossflows()
#' }
#'
read_lfs_grossflows <- function(weights = c(
                                  "current",
                                  "previous"
                                ),
                                path = Sys.getenv("R_READABS_PATH",
                                  unset = tempdir()
                                )) {
  weights <- match.arg(weights)

  weight_desc <- switch(weights,
    "current" = "current month",
    "previous" = "previous month"
  )

  gf_file <- download_abs_data_cube(
    catalogue_string = "labour-force-australia",
    cube = "gm1",
    path = path
  )

  sheet_name <- switch(weights,
    "current" = "Data 1",
    "previous" = "Data 2"
  )

  raw_sheet <- readxl::read_xlsx(
    path = gf_file,
    sheet = sheet_name,
    skip = 4,
    col_names = c(
      "date",
      "sex",
      "age",
      "state",
      "lfs_current",
      "lfs_previous",
      "persons"
    ),
    col_types = c(
      "date",
      "text",
      "text",
      "text",
      "text",
      "text",
      "numeric"
    )
  )

  gf <- raw_sheet %>%
    mutate(
      date = as.Date(date,
        format = "%Y-%m-%d %H-%M-%S"
      ),
      unit = "000s",
      weights = weight_desc
    )

  # Run some minimal checks on the data frame to ensure its contents are as
  # expected
  stopifnot(check_lfs_grossflows(gf))

  gf
}

#' Internal function to check if the data frame returned by read_lfs_grossflows()
#' contains expected unique values in key columns
#' @param df data frame containing gross flows data
#' @keywords internal
check_lfs_grossflows <- function(df) {
  names_match <- identical(
    names(df),
    c(
      "date",
      "sex",
      "age",
      "state",
      "lfs_current",
      "lfs_previous",
      "persons",
      "unit",
      "weights"
    )
  )

  sex_match <- identical(
    unique(df$sex),
    c(
      "Males",
      "Females"
    )
  )

  age_match <- identical(
    unique(df$age),
    c(
      "15-19 years",
      "20-24 years",
      "25-29 years",
      "30-34 years",
      "35-39 years",
      "40-44 years",
      "45-49 years",
      "50-54 years",
      "55-59 years",
      "60-64 years",
      "65 years and over"
    )
  )

  lfs_match <- identical(
    unique(df$lfs_current),
    c(
      "Employed full-time",
      "Employed part-time",
      "Unemployed",
      "Not in the labour force (NILF)",
      "Unmatched in common sample (responded in previous month but not in current)",
      "Outgoing rotation group"
    )
  )

  all(
    names_match,
    sex_match,
    age_match,
    lfs_match
  )
}

test_that("read_lfs_grossflows() downloads and imports LFS cube GM1", {
  skip_if_offline()
  skip_on_cran()

  prev_option <- getOption("timeout")
  options("timeout" = 120)
  gf_cur <- read_lfs_grossflows()
  gf_prev <- read_lfs_grossflows(weights = "previous")

  options("timeout" = prev_option)

  # Current should still be the default
  expect_identical(
    unique(gf_cur$weights),
    "current month"
  )

  # Pass minimal tests
  expect_true(check_lfs_grossflows(gf_cur))
  expect_true(check_lfs_grossflows(gf_prev))

  expect_gt(nrow(gf_cur), 1000000)
  expect_false(identical(gf_cur, gf_prev))
  expect_false(any(is.na(gf_cur$persons)))
  expect_equal(min(gf_cur$persons), 0)
  expect_true(max(gf_cur$persons) > min(gf_cur$persons))
})

#' Download a tidy tibble containing the Estimated Residential Population from the ABS
#'
#' \code{read_erp()} uses the \code{read_abs()} function to download, import,
#' and tidy the Estimated Residential Population from the ABS. It allows the user
#' to specify age, sex and states/territories of interest. It returns a tibble
#' containing five columns: the date, the age range, sex and states that the ERP
#' corresponds to. This makes joining the ERP to another dataframe easy.
#'
#' @param age_range numeric; default is "0:100". A vector containing ages in single
#' years for which an ERP is sought. The ABS top-code ages at 100.
#'
#' @param sex character; default is "Persons". Other values are "Male" and
#' "Female". Multiple values allowed.
#'
#' @param states character; default is "Australia". Other values are the full
#' or abbreviated names of the states and self-governing territories. Multiple
#' values allowed.
#'
#' @param path character; default is "data/ABS". Only used if
#' retain_files is set to TRUE. Local directory in which to save
#' downloaded ABS time series spreadsheets.
#'
#' @param show_progress_bars logical; TRUE by default. If set to FALSE, progress
#' bars will not be shown when ABS spreadsheets are downloading.
#'
#' @param check_local logical; FALSE by default. See \code{?read_abs}.
#'
#' @param retain_files logical; FALSE by default. When TRUE, the spreadsheets
#' downloaded from the ABS website will be saved in the
#' directory specified with 'path'.
#'
#' @examples
#' \donttest{
#' # Create a tibble called 'erp' that contains the ERP index
#' # numbers for 30 June each year for Australia.
#'
#' erp <- read_erp()
#' }
#'
#' @importFrom dplyr filter select tibble mutate group_by summarise arrange
#' @importFrom tools toTitleCase
#'
#' @export

read_erp <- function(age_range = 0:100,
                     sex = "Persons",
                     states = c(
                       "Australia", "New South Wales", "Victoria", "Queensland",
                       "South Australia", "Western Australia", "Tasmania",
                       "Northern Territory", "Australian Capital Territory"
                     ),
                     path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
                     show_progress_bars = TRUE,
                     check_local = FALSE,
                     retain_files = FALSE) {
  if (!is.numeric(age_range)) {
    stop("The age_range argument must be a numeric vector.")
  }

  # Check if all elements are integers
  if (any(age_range != as.integer(age_range))) {
    stop("All elements in age_range must be integers.")
  }

  # Check if all elements are between 0 and 100
  if (any(age_range < 0 | age_range > 100)) {
    stop("All elements in age_range must be between 0 and 100.")
  }

  # Restrict the 'sex' argument to the valid choices
  sex <- match.arg(sex,
    c("Persons", "Male", "Female"),
    several.ok = TRUE
  )

  # Restrict the states argument to valid choices but include abbreviations
  # Always return the full name if an abbreviation has been used.
  stes <- purrr::map_chr(states, validate_state_erp)

  if (!is.logical(retain_files)) {
    stop("`retain_files` must be either `TRUE` or `FALSE`.")
  }

  if (!is.logical(show_progress_bars)) {
    stop("`show_progress_bars` must be either `TRUE` or `FALSE`")
  }

  if (retain_files == TRUE & !is.character(path)) {
    stop(
      "If `retain_files` is `TRUE`,",
      " you must specify a valid file path in `path`."
    )
  }

  state_lookup <- erp_state_lookup()
  ste_tbls <- state_lookup$tbl_n[state_lookup$full_name %in% stes]

  erp_raw <- read_abs(
    cat_no = "3101.0",
    tables = ste_tbls,
    retain_files = retain_files,
    check_local = check_local,
    show_progress_bars = show_progress_bars,
    path = path
  )

  x <- tidy_erp(
    erp_raw,
    age_range,
    sex
  )

  x
}

#' @keywords internal
# Tidy a table of ERP data downloaded with read_abs(cat_no = "3101.0")
# x <- read_abs(cat_no = "3101.0")
# tidy_erp(x)
tidy_erp <- function(erp_raw,
                     age_range,
                     sex) {
  erp_raw %>%
    dplyr::mutate(
      age = gsub("[^0-9]", "", .data$series),
      age = as.integer(.data$age),
      series_sex = gsub(".*;\\s*(Male|Female|Persons)\\s*;.*", "\\1", .data$series),
      state = trimws(gsub(".*,(\\s*[^,]+)$", "\\1", .data$table_title))
    ) %>%
    dplyr::filter(
      .data$age %in% age_range,
      .data$series_sex %in% sex
    ) %>%
    dplyr::group_by(.data$date, .data$series_sex, .data$state, .data$age) %>%
    dplyr::summarise(erp = sum(.data$value)) %>%
    dplyr::ungroup() %>%
    dplyr::arrange(.data$state, .data$series_sex, .data$age, .data$date) %>%
    dplyr::select("date", "state",
      sex = "series_sex",
      "erp", "age"
    )
}

#' @keywords internal
#' @noRd
erp_state_lookup <- function() {
  # Lookup table of state/territory names, abbreviations and ABS ERP table numbers
  state_lookup <- tibble(
    full_name = c(
      "Australia", "New South Wales", "Victoria", "Queensland",
      "South Australia", "Western Australia", "Tasmania",
      "Northern Territory", "Australian Capital Territory"
    ),
    abbrev = c("Aus", "NSW", "Vic", "Qld", "SA", "WA", "Tas", "NT", "ACT"),
    tbl_n = c(59, 51:58)
  )
}

# Ensure that user specified state/territory names are acceptable. Allow full
# names or abbreviations. Throw an error if they are incorrect.
# Always return the full name.
#' @keywords internal
#' @noRd
#'

validate_state_erp <- function(state) {
  state_lookup <- erp_state_lookup()

  # Define valid states and their abbreviations
  valid_states <- state_lookup$full_name
  valid_abbreviations <- state_lookup$abbrev

  # Allow trailing periods in abbreviations
  valid_abbreviations_with_period <- paste0(valid_abbreviations, ".")

  # Combine all valid inputs
  all_valid_inputs <- c(
    valid_states,
    valid_abbreviations,
    valid_abbreviations_with_period,
    toupper(valid_abbreviations),
    tolower(valid_abbreviations)
  )

  # Create a lookup table (map all variants to full state names)
  state_map <- stats::setNames(
    rep(valid_states, times = 5),
    all_valid_inputs
  )

  # Standardize input to lowercase and remove trailing periods in case
  # people do things like "Vic." or "Tas."
  state_cleaned <- gsub("\\.$", "", state) # Remove trailing period
  state_cleaned <- tolower(state_cleaned) # Convert to lowercase

  # Create a cleaned version of the map for case-insensitive matching
  names(state_map) <- tolower(names(state_map))

  # Check if the input matches a valid state or abbreviation
  if (!state_cleaned %in% names(state_map)) {
    stop("Invalid state. Please use a valid state name or abbreviation.")
  }

  # Return the standard full state name
  return(unname(tools::toTitleCase(state_map[state_cleaned])))
}

test_that("read_erp() works", {
  x <- read_erp()

  expect_is(x, "tbl_df")
  expect_length(x, 5)
  expect_gt(nrow(x), 40000)
  expect_is(x$date, "Date")
  expect_is(x$state, "character")
  expect_is(x$sex, "character")
  expect_is(x$erp, "numeric")
  expect_is(x$age, "integer")
  expect_equal(length(unique(x$state)), 9)
  expect_equal(unique(x$sex), "Persons")
  expect_equal(unique(x$age), 0:100)
})

test_that("tidy_erp() works with all permutations of parameters", {
  raw_erp <- read_abs("3101.0")

  test_tidy_erp <- function(age_range,
                            sex,
                            erp_raw = raw_erp) {
    x <- tidy_erp(
      erp_raw,
      age_range,
      sex
    )

    expect_is(x, "tbl_df")
    expect_length(x, 5)
    expect_equal(sort(unique(x$age)), sort(age_range))
    expect_equal(sort(unique(x$sex)), sort(sex))
  }

  test_tidy_erp(1:10, "Male")
  test_tidy_erp(15, "Female")
  test_tidy_erp(c(10, 50, 90), "Persons")
  test_tidy_erp(
    0:100,
    c("Male", "Female", "Persons")
  )
  test_tidy_erp(
    1,
    c("Female", "Male")
  )
})

#' Helper function for \code{download_abs_data_cube} to show the available catalogues.
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' This function lists the possible catalogues that are available on the ABS website.
#' These catalogues must be specified as a string as an argument to \code{download_abs_data_cube}.
#'
#' @param selected_heading optional character string specifying the heading on the \href{https://www.abs.gov.au/statistics}{ABS statistics webpage}.
#' e.g. "Earnings and work hours"
#' @param refresh logical; `FALSE` by default. If `FALSE`, an internal table of the available ABS catalogues is used. If `TRUE`, this table is refreshed from the ABS website.
#'
#' @examples
#' show_available_catalogues("Earnings and work hours")
#' @importFrom dplyr %>% pull filter
#' @importFrom rlang .data
#'
#' @return a character vector of catalogues.
#' @export
#' @family data cube functions
show_available_catalogues <- function(selected_heading = NULL, refresh = FALSE) {
  if (isFALSE(is.logical(refresh))) {
    stop("`refresh` must be `TRUE` or `FALSE`.")
  }

  if (isFALSE(refresh)) {
    table_to_use <- abs_lookup_table
  } else {
    table_to_use <- scrape_abs_catalogues()
  }

  if (!is.null(selected_heading)) {
    available_catalogues <- filter(table_to_use, .data$heading == selected_heading)
  } else {
    available_catalogues <- table_to_use
  }

  available_catalogues <- pull(available_catalogues, .data$catalogue)

  nonempty_available_catalogues <- available_catalogues[
    nzchar(available_catalogues)
  ]

  return(nonempty_available_catalogues)
}

#' Get date of most recent observation(s) in ABS time series
#'
#' This function returns the most recent observation date for a specified ABS
#' time series catalogue number (as a whole), individual tables, or series IDs.
#'
#' @param cat_no ABS catalogue number, as a string, including the extension.
#' For example, "6202.0".
#'
#' @param tables numeric. Time series tables in `cat_no`` to download and
#' extract. Default is "all", which will read all time series in `cat_no`.
#' Specify `tables` to download and import specific tables(s) -
#' eg. `tables = 1` or `tables = c(1, 5)`.
#'
#' @param series_id (optional) character. Supply an ABS unique time series
#' identifier (such as "A2325807L") to get only that series.
#' This is an alternative to specifying `cat_no`.
#'
#' @details Where the individual time series in your request
#' have multiple dates, only the most recent will be returned.
#'
#' @return Date vector of length one. Date corresponds to the most recent
#' observation date for any of the time series in the table(s) requested.
#' observation date for any of the time series in the table(s) requested.
#'
#' @examples
#' \dontrun{
#'
#' # Check a whole catalogue number; return the latest release date for any
#' # time series in the number
#'
#' check_latest_date("6345.0")
#'
#' # Return latest release date for a table within a catalogue number  - note
#' # the function will return the release date
#' # of the most-recently-updated series within the tables
#' check_latest_date("6345.0", tables = 1)
#'
#' # Or for multiple tables - note the function will return the release date
#' # of the most-recently-updated series within the tables
#' check_latest_date("6345.0", tables = c("1", "5a"))
#'
#' # Or for an individual time series
#' check_latest_date(series_id = "A2713849C")
#' }
#'
#' @export

check_latest_date <- function(cat_no = NULL,
                              tables = "all",
                              series_id = NULL) {
  # check that R has access to the internet
  check_abs_connection()

  # Create URLs to query the ABS Time Series Directory
  xml_urls <- form_abs_tsd_url(
    cat_no = cat_no,
    tables = tables,
    series_id = series_id
  )

  xml_df <- purrr::map(xml_urls, get_last_xml_page) %>%
    purrr::pluck(1)

  max_date <- as.Date(xml_df$SeriesEnd,
    format = "%d/%m/%Y"
  ) %>%
    max()

  max_date
}

#' Extract data sheets from an ABS timeseries workbook saved locally as an
#' Excel file.
#'
#' Note that this function will not tidy the data for you.
#' Use `read_abs_local()`to import and tidy data from local ABS time series
#' spreadsheets or `read_abs()` to download, import and tidy ABS time series.
#'
#' @param filename Filename for an ABS time series spreadsheet (as string)
#'
#' @param table_title String giving the full title of the ABS table, such as
#' "Table 1. Employed persons, Australia"
#'
#' @param path Local directory in which an ABS time series is stored. Default is
#' `Sys.getenv("R_READABS_PATH", unset = tempdir())`.
#'
#' @importFrom readxl excel_sheets read_excel
#' @importFrom dplyr filter "%>%"
#' @importFrom purrr map set_names
#' @importFrom tools file_path_sans_ext
#'
#' @export

extract_abs_sheets <- function(filename,
                               table_title = NULL,
                               path = Sys.getenv("R_READABS_PATH",
                                 unset = tempdir()
                               )) {
  filename <- file.path(path, filename)

  sheets <- readxl::excel_sheets(path = filename)
  sheets <- sheets[!sheets %in% c(
    "Index", "Inquiries", "Enquiries",
    "Contents", "Technical note", "Technical notes", "Note", "Notes"
  )]
  sheets <- sheets[!grepl("Table ", sheets, ignore.case = T)]

  if (length(sheets) < 1) {
    stop(sprintf(
      "The Excel workbook %s appears to have no data sheets.",
      filename
    ))
  }

  data_sheets <- purrr::map(
    .x = sheets,
    .f = readxl::read_excel,
    path = filename,
    trim_ws = TRUE,
    .name_repair = "minimal"
  )

  data_sheets <- data_sheets %>%
    purrr::set_names(paste0(
      tools::file_path_sans_ext(basename(filename)), "=",
      sheets, "=",
      table_title
    ))

  data_sheets
}

#' Tidy multiple dataframes of ABS time series data contained in a list.
#'
#' @param list_of_dfs A list of dataframes containing extracted
#' ABS time series data.
#'
#' @param metadata logical. If `TRUE` (the default), a tidy data frame including
#' ABS metadata (series name, table name, etc.) is included in the output. If
#' `FALSE`, metadata is dropped.
#'
#' @importFrom purrr map_dfr
#' @importFrom tidyr separate
#' @importFrom dplyr "%>%"
#' @export

tidy_abs_list <- function(list_of_dfs, metadata = TRUE) {
  table_no <- sheet_no <- table_title <- NULL

  if (!"list" %in% class(list_of_dfs)) {
    if ("data.frame" %in% class(list_of_dfs)) {
      stop(
        "abs_tidy_list() works with lists of data frames,",
        " not individual data frames. Use abs_tidy() to tidy an",
        " individual data frame that is not contained in a list."
      )
    }
    stop("The object is not a list of data frames.")
  }

  message("Tidying data from imported ABS spreadsheets")

  # apply abs_tidy to each element of list_of_dfs and combine into 1 df
  # with new column "sheet_id"
  x <- purrr::map(list_of_dfs,
    tidy_abs,
    metadata = metadata
  )

  sheet_ids <- names(list_of_dfs)

  list_of_split_sheet_ids <- stringi::stri_split_fixed(sheet_ids, "=", 3)

  for (i in seq_along(x)) {
    x[[i]]$table_no <- list_of_split_sheet_ids[[i]][1]
    x[[i]]$sheet_no <- list_of_split_sheet_ids[[i]][2]
    x[[i]]$table_title <- list_of_split_sheet_ids[[i]][3]
  }

  x <- bind_rows(x)

  x <- select(x, table_no, sheet_no, table_title, everything())

  x
}

download_abs <- function(urls,
                         path,
                         show_progress_bars = TRUE) {
  # create a subdirectory of the working directory with value given by
  # the path argument and define the filename that will be given to the
  # downloaded table
  if (!dir.exists(path)) {
    dir.create(path = path, recursive = TRUE, showWarnings = FALSE)
  }

  filenames <- file.path(path, basename(urls))

  purrr::walk2(
    .x = urls,
    .y = filenames,
    .f = dl_file,
    quiet = !show_progress_bars
  )

  return(TRUE)
}

dl_file <- function(url,
                    destfile,
                    quiet = TRUE,
                    method = Sys.getenv("R_READABS_DL_METHOD", unset = "auto")) {
  suppressWarnings(
    utils::download.file(
      url = url,
      destfile = destfile,
      mode = "wb",
      quiet = quiet,
      headers = readabs_header,
      cacheOK = FALSE,
      method = method
    )
  )
}

safely_download_abs <- purrr::safely(download_abs)

#' Download, extract, and tidy ABS time series spreadsheets
#' @description
#' `r lifecycle::badge("stable")`
#'
#' \code{read_abs()} downloads ABS time series spreadsheets,
#' then extracts the data from those spreadsheets,
#' then tidies the data. The result is a single
#' data frame (tibble) containing tidied data.
#'
#' @param cat_no ABS catalogue number, as a string, including the extension.
#' For example, "6202.0".
#'
#' @param tables numeric. Time series tables in `cat_no`` to download and
#' extract. Default is "all", which will read all time series in `cat_no`.
#' Specify `tables` to download and import specific tables(s) -
#' eg. `tables = 1` or `tables = c(1, 5)`.
#'
#' @param series_id (optional) character. Supply an ABS unique time series
#' identifier (such as "A2325807L") to get only that series.
#' This is an alternative to specifying `cat_no`.
#'
#' @param path Local directory in which downloaded ABS time series
#' spreadsheets should be stored. By default, `path` takes the value set in the
#' environment variable "R_READABS_PATH". If this variable is not set,
#' any files downloaded by read_abs()  will be stored in a temporary directory
#' (\code{tempdir()}). See \code{Details} below for
#' more information.
#'
#' @param metadata logical. If `TRUE` (the default), a tidy data frame including
#' ABS metadata (series name, table name, etc.) is included in the output. If
#' `FALSE`, metadata is dropped.
#'
#' @param show_progress_bars TRUE by default. If set to FALSE, progress bars
#' will not be shown when ABS spreadsheets are downloading.
#'
#' @param retain_files when TRUE (the default), the spreadsheets downloaded
#' from the ABS website will be saved in the directory specified with `path`.
#' If set to `FALSE`, the files will be stored in a temporary directory.
#'
#' @param ... Arguments to `read_abs_series()` are passed to `read_abs()`.
#'
#' @param check_local If `TRUE`, the default, local `fst` files are used,
#' if present.
#'
#' @param release_date Either `"latest"` or a string coercible to a date, such as
#' `"2022-02-01"`. If `"latest"`, the latest release of the requested data will
#' be returned. If a date, (eg. `"2022-02-01"`) `read_abs()` will
#' attempt to download the data from that month's release. Note that this only
#' works consistently as expected for monthly data. See `Details`.
#'
#' @return A data frame (tibble) containing the tidied data from the ABS time
#' series table(s).
#'
#' @details
#' `read_abs_series()` is a wrapper around `read_abs()`, with `series_id` as
#' the first argument.
#'
#' `read_abs()` downloads spreadsheet(s) from the ABS containing time
#' series data. These files need to be saved somewhere on your disk.
#' This local directory can be controlled using the `path` argument to
#' `read_abs()`. If the `path` argument is not set, `read_abs()` will store
#' the files in a directory set in the "R_READABS_PATH" environment variable.
#' If this variable isn't set, files will be saved in a temporary directory.
#'
#' To check the value of the "R_READABS_PATH" variable, run
#' \code{Sys.getenv("R_READABS_PATH")}. You can set the value of this variable
#' for a single session using \code{Sys.setenv(R_READABS_PATH = <path>)}.
#' If you would like to change this variable for all future R sessions, edit
#' your `.Renviron` file and add \code{R_READABS_PATH = <path>} line.
#' The easiest way to edit this file is using \code{usethis::edit_r_environ()}.
#'
#' Certain corporate networks restrict your ability to download files in an R
#' session. On some of these networks, the `"wininet"` method must be used when
#' downloading files. Users can now specify the method that will be used to
#' download files by setting the `"R_READABS_DL_METHOD"` environment variable.
#'
#' For example, the following code sets the environment variable for your
#' current session: s`Sys.setenv("R_READABS_DL_METHOD" = "wininet")`
#' You can add `R_READABS_DL_METHOD = "wininet"` to your .Renviron to have
#' this persist across sessions.
#'
#' The `release_date` argument allows you to download table(s) other than the
#' latest release. This is useful for examining revisions to time series, or
#' for obtaining the version of series that were available on a given date.
#' Note that you cannot supply more than one date to `release_date`. Note also
#' that any dates prior to mid-2019 (the exact date varies by series) will fail.
#' Specifying `release_date` only reliably works for monthly, and some
#' quarterly, data. It does not work for annual data.
#'
#' @rdname read_abs
#' @examples
#'
#' # Download and tidy all time series spreadsheets
#' # from the Wage Price Index (6345.0)
#' \dontrun{
#' wpi <- read_abs("6345.0")
#' }
#'
#' # Download table 1 from the Wage Price Index
#' \dontrun{
#' wpi_t1 <- read_abs("6345.0", tables = "1")
#' }
#'
#' # Or table 1 as in the Sep 2019 release of the WPI:
#' \dontrun{
#' wpi_t1_sep2019 <- read_abs("6345.0", tables = "1", release_date = "2019-09-01")
#' }
#'
#' # Or tables 1 and 2a from the WPI
#' \dontrun{
#' wpi_t1_t2a <- read_abs("6345.0", tables = c("1", "2a"))
#' }
#'
#'
#' # Get two specific time series, based on their time series IDs
#' \dontrun{
#' cpi <- read_abs(series_id = c("A2325806K", "A2325807L"))
#' }
#'
#' # Get series IDs using the `read_abs_series()` wrapper function
#' \dontrun{
#' cpi <- read_abs_series(c("A2325806K", "A2325807L"))
#' }

#' @importFrom purrr walk walk2 map map_dfr map2
#' @importFrom dplyr group_by filter
#' @name read_abs
#' @export

read_abs <- function(cat_no = NULL,
                     tables = "all",
                     series_id = NULL,
                     path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
                     metadata = TRUE,
                     show_progress_bars = TRUE,
                     retain_files = TRUE,
                     check_local = TRUE,
                     release_date = "latest") {
  if (isTRUE(check_local) &&
    fst_available(cat_no = cat_no, path = path)) {
    if (!identical(tables, "all")) {
      warning(
        "`tables` was provided",
        "yet `check_local = TRUE` and fst files are available ",
        "so `tables` will be ignored."
      )
    }
    out <- fst::read_fst(path = catno2fst(cat_no = cat_no, path = path))
    out <- dplyr::as_tibble(out)
    if (is.null(series_id)) {
      return(out)
    }
    if (series_id %in% out[["series_id"]]) {
      users_series_id <- series_id
      out <- dplyr::filter(out, series_id %in% users_series_id)
    } else {
      warning(
        "`series_id` was provided,",
        "but was not present in the local table and will be ignored."
      )
    }
    return(out)
  }

  if (!is.logical(retain_files)) {
    stop("The `retain_files` argument to `read_abs()` must be TRUE or FALSE.")
  }

  if (is.null(cat_no) & is.null(series_id)) {
    stop(
      "read_abs() requires either an ABS catalogue number,",
      "such as '6202.0' or '6401.0',",
      "or an ABS time series ID like 'A84423127L'."
    )
  }

  if (!is.null(cat_no) & !is.null(series_id)) {
    stop("Please specify either the cat_no OR the series_id, not both.")
  }

  if (!is.null(cat_no)) {
    if (nchar(cat_no) < 6) {
      message(paste0(
        "Please ensure you include the cat_no extension.\n",
        "`read_abs()` will assume you meant \"", cat_no,
        ".0\"", " rather than ", cat_no
      ))
      cat_no <- paste0(cat_no, ".0")
    }
  }

  if (!is.null(cat_no) & is.null(tables)) {
    message(paste0(
      "`tables` not specified;",
      "attempting to fetch all tables from ",
      cat_no
    ))
    tables <- "all"
  }

  if (!is.logical(metadata)) {
    stop("`metadata` argument must be either TRUE or FALSE")
  }

  if (length(release_date) != 1) {
    stop("`release_date` argument must be length 1.")
  }

  # satisfy CRAN
  ProductReleaseDate <- SeriesID <- NULL

  # create a subdirectory of 'path' corresponding to the catalogue number
  # if specified
  if (retain_files && !is.null(cat_no)) {
    .path <- file.path(path, cat_no)
  } else {
    # create temp directory to temporarily store
    # spreadsheets if retain_files == FALSE
    if (!retain_files) {
      .path <- tempdir()
    } else {
      .path <- path
    }
  }

  # check that R has access to the internet
  check_abs_connection()

  # Create URLs to query the ABS Time Series Directory
  xml_urls <- form_abs_tsd_url(
    cat_no = cat_no,
    tables = tables,
    series_id = series_id
  )

  # find spreadsheet URLs from cat_no in the Time Series Directory
  download_message <- ifelse(!is.null(cat_no),
    paste0("catalogue ", cat_no),
    paste0("series ID", series_id)
  )

  message(paste0(
    "Finding URLs for tables corresponding to ABS ",
    download_message
  ))

  xml_dfs <- purrr::map_dfr(xml_urls,
    .f = get_abs_xml_metadata,
    .progress = TRUE
  )

  # Ensure we're not getting spurious matches of table numbers
  if (tables[1] != "all" && is.null(series_id)) {
    xml_dfs <- xml_dfs[match_tables(xml_dfs$TableTitle, tables), ]
  }

  # the same Series ID can appear in multiple spreadsheets;
  # we just want one (the latest)
  if (!is.null(series_id)) {
    xml_dfs <- xml_dfs %>%
      dplyr::group_by(SeriesID) %>%
      dplyr::filter(ProductReleaseDate == max(ProductReleaseDate)) %>%
      dplyr::filter(row_number() == 1) %>%
      dplyr::ungroup()
  } else {
    xml_dfs <- xml_dfs %>%
      dplyr::group_by(.data$TableURL) %>%
      dplyr::filter(.data$ProductReleaseDate == max(.data$ProductReleaseDate)) %>%
      dplyr::filter(.data$ProductReleaseDate == max(.data$ProductReleaseDate)) %>%
      dplyr::ungroup()
  }

  urls <- unique(xml_dfs$TableURL)
  # Remove spaces from URLs
  urls <- gsub(" ", "+", urls)

  if (as.character(release_date) != "latest") {
    requested_date <- format(as.Date(release_date), "%b-%Y")
    urls <- gsub(
      "latest-release",
      tolower(requested_date),
      urls
    )
  }

  table_titles <- unique(xml_dfs$TableTitle)

  # download tables corresponding to URLs
  message(paste0(
    "Attempting to download files from ", download_message,
    ", ", xml_dfs$ProductTitle[1]
  ))

  dl_result <- safely_download_abs(
    urls = urls,
    path = .path,
    show_progress_bars = show_progress_bars
  )

  if (is.null(dl_result$result)) {
    urls <- gsub(".xlsx", ".xls", urls)
    dl_result_xls <- safely_download_abs(
      urls = urls,
      path = .path,
      show_progress_bars = show_progress_bars
    )

    if (!is.null(dl_result_xls$error)) {
      stop("URL ", url, " does not appear to be valid.")
    }
  }

  # extract the sheets to a list
  filenames <- base::basename(urls)
  message("Extracting data from downloaded spreadsheets")
  sheets <- purrr::map2(filenames, table_titles,
    .f = extract_abs_sheets, path = .path,
    .progress = TRUE
  )

  # remove one 'layer' of the list,
  # so that each sheet is its own element in the list
  sheets <- unlist(sheets, recursive = FALSE)

  # tidy the sheets
  sheet <- tidy_abs_list(sheets, metadata = metadata)

  # remove spreadsheets from disk if `retain_files` == FALSE
  if (!retain_files) {
    # delete downloaded files
    file.remove(file.path(.path, filenames))
  }

  # if series_id is specified, remove all other series_ids

  if (!is.null(series_id)) {
    users_series_id <- series_id

    sheet <- sheet %>%
      dplyr::filter(series_id %in% users_series_id)
  }

  # if fst is available, and what has been requested is the full data,
  #  write the result to the <path>/fst/ file
  if (retain_files &&
    is.null(series_id) &&
    identical(tables, "all") &&
    requireNamespace("fst", quietly = TRUE)) {
    fst::write_fst(
      sheet,
      catno2fst(
        cat_no = cat_no,
        path = path
      )
    )
  }

  # return a data frame
  sheet
}

#' @rdname read_abs
#' @export

read_abs_series <- function(series_id, ...) {
  read_abs(
    series_id = series_id,
    ...
  )
}

match_tables <- function(table_list, requested_tables) {
  requested <- paste0(requested_tables, collapse = "|")
  # Looking for table number preceded by a space or a 0, and
  # followed my a full stop or a space
  regex_pattern <- paste0(
    "(?<=\\s|0)",
    "(", requested, ")",
    "(?=\\.|\\s|\\:)"
  )

  predot_matches <- regexpr(".*\\.|.*\\:", table_list)
  table_list_predot <- regmatches(table_list, m = predot_matches)

  grepl(regex_pattern, table_list_predot, perl = TRUE, ignore.case = TRUE)
}

library(readabs)

wpi_url <- "http://ausstats.abs.gov.au/servlet/TSSearchServlet?catno=6345.0&pg=1&ttitle=1"

test_that("WPI XML page is a data.frame with expected column names", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  wpi_1_xml <- get_xml_dfs(wpi_url)

  expect_is(wpi_1_xml, "data.frame")
  expect_equal(colnames(wpi_1_xml)[1], "ProductNumber")
  expect_length(wpi_1_xml, 18)
  expect_gt(nrow(wpi_1_xml), 26)
})

test_that("read_abs() downloads, imports, and tidies a data frame", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  wpi_1 <- read_abs("6345.0",
    tables = "1", retain_files = FALSE,
    check_local = FALSE, path = tempdir()
  )

  expect_is(wpi_1, "tbl")
  expect_length(wpi_1, 12)
  expect_gt(nrow(wpi_1), 2400)
  expect_is(wpi_1$date, "Date")
  expect_is(wpi_1$series, "character")
  expect_is(wpi_1$value, "numeric")
  expect_lt(Sys.Date() - max(wpi_1$date), 180)
})

test_that("read_abs() gets a whole catalogue number", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  wpi <- read_abs("6345.0", retain_files = FALSE, check_local = FALSE, path = tempdir())

  expect_is(wpi, "tbl")
  expect_length(wpi, 12)
  expect_gt(nrow(wpi), 34000)
})

test_that("read_abs() works when retain_files = FALSE", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  wpi_7 <- read_abs("6345.0",
    tables = "7a", retain_files = FALSE, check_local = F,
    path = tempdir()
  )

  expect_is(wpi_7, "tbl")
})

test_that("read_abs() works with series ID(s)", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  lfs_2 <- read_abs(series_id = c("A84423071L", "A84423141J"), retain_files = FALSE, check_local = F, path = tempdir())

  expect_is(lfs_2, "tbl")
  expect_length(unique(lfs_2$series), 2)
  expect_match(unique(lfs_2$series)[1], "Employed total ;  > Males ;")
  expect_match(unique(lfs_2$series)[2], "Employed total ;  > Females ;")
  expect_length(lfs_2, 12)

  lfs_wrapper <- read_abs_series(series_id = c("A84423071L", "A84423141J"), retain_files = FALSE, check_local = F, path = tempdir())

  expect_identical(lfs_2, lfs_wrapper)
})



test_that("read_abs() returns appropriate errors and messages when given invalid input", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  expect_error(read_abs(cat_no = NULL))
  expect_error(read_abs("6202.0", 1, retain_files = NULL, check_local = FALSE))
  expect_error(read_abs(cat_no = "6345.0", metadata = 1, check_local = FALSE))
  expect_error(read_abs(cat_no = "6345.0", series_id = "foo", check_local = FALSE),
    regexp = "either.*cat_no.*series_id"
  )
})

test_that("read_abs() works with 'table 01' as well as 'table 1' filename structures", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  const_df <- read_abs("8755.0", 1, retain_files = FALSE, check_local = FALSE, path = tempdir())
  const_df_2 <- read_abs("8755.0", "1", retain_files = FALSE, check_local = FALSE, path = tempdir())

  expect_is(const_df, "data.frame")
  expect_equal(const_df, const_df_2)
  expect_length(const_df, 12)
})

test_that("read_abs() returns an error when requesting non-existing cat_no", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  expect_error(read_abs("9999.0"))
})


test_that("read_cpi() function downloads CPI index numbers", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  cpi <- read_cpi()

  expect_is(cpi, "tbl")
  expect_length(cpi, 2)
  expect_is(cpi$date, "Date")
  expect_is(cpi$cpi, "numeric")

  latest_cpi_date <- cpi$date[cpi$date == max(cpi$date)]


  # Test that inflation over the most recent year is within expected bounds
  date_12m_before_latest <- as.POSIXlt(latest_cpi_date)
  date_12m_before_latest$year <- date_12m_before_latest$year - 1
  date_12m_before_latest <- as.Date(date_12m_before_latest)

  latest_annual_inflation <- (cpi$cpi[cpi$date == latest_cpi_date] /
    cpi$cpi[cpi$date == date_12m_before_latest]) - 1
  expect_gt(latest_annual_inflation, -.02)
  expect_lt(latest_annual_inflation, 0.1)

  # Test that default is quarterly
  expect_equal(min(cpi$date), as.Date("1948-09-01"))

  cpi_m <- read_cpi("monthly")
  expect_equal(min(cpi_m$date), as.Date("2024-04-01"))

})

test_that("read_cpi() returns appropriate errors", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  expect_error(read_cpi(retain_files = NULL))
  expect_error(read_cpi(show_progress_bars = NULL))
})

test_that("3401.0 table 1 can be loaded", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()
  expect_s3_class(read_abs("3401.0", "1"), "tbl_df")
})

test_that("previous vintages of time series can be loaded", {
  skip_on_cran()
  skip_if_offline()
  check_abs_connection()

  expect_s3_class(
    read_abs("6345.0", 1, release_date = "2020-03-01", check_local = F),
    "tbl_df"
  )

  expect_error(read_abs("6345.0", 1,
    release_date = "2020-03",
    check_local = F
  ))

  expect_error(read_abs("6345.0", 1,
    release_date = c("2020-03-01", "2020-06-01"),
    check_local = F
  ))
})

# Function by Jaron Lee, United States Studies Centre

#' Extracts ABS time series data from local Excel spreadsheets and converts to
#' long format.
#'
#' `read_abs_data()` is soft deprecated and will be removed in a future version.
#' Please use `read_abs_local()` to import and tidy locally-stored
#' ABS time series spreadsheets, or `read_abs()` to download, import,
#' and tidy time series spreadsheets from the ABS website.
#'
#' @param path Filepath to Excel spreadsheet.
#' @param sheet Sheet name or number.
#'
#' @return Long-format dataframe
#' @export
#'
read_abs_data <- function(path, sheet) {
  Date <- series <- value <- NULL
  df <- readxl::read_excel(path = path, sheet = sheet)
  dat <- df[-(1:9), ]
  colnames(dat)[1] <- "Date"
  dat$Date <- as.Date(as.integer(dat$`Date`), origin = "1899-12-30")
  dat <- tidyr::gather(dat, `series`, `value`, -`Date`)
  dat$value <- as.numeric(dat$`value`)

  .Deprecated(
    new = "read_abs_local()",
    old = "read_abs_data()",
    msg = "read_abs_data() is deprecated and will be removed in a future version. Please use read_abs_local()."
  )

  return(dat)
}


#' Extracts ABS series metadata directly from Excel spreadsheets and
#' converts to long-form.
#'
#' @param path Filepath to Excel spreadsheet.
#' @param sheet Sheet name or number.
#'
#' @return Long-form dataframe
#' @export
#'
read_abs_metadata <- function(path, sheet) {
  df <- readxl::read_excel(path = path, sheet = sheet, col_names = FALSE)
  dat <- df[(1:9), ]
  final_dat <- (t(dat[, 2:ncol(dat)]))
  colnames(final_dat) <- t(dat[, 1])
  rownames(final_dat) <- NULL
  final_dat <- as.data.frame(final_dat, stringsAsFactors = FALSE)
  final_dat$`Series Start` <- as.Date(as.integer(final_dat$`Series Start`),
    origin = "1899-12-30"
  )
  final_dat$`Series End` <- as.Date(as.integer(final_dat$`Series End`),
    origin = "1899-12-30"
  )
  colnames(final_dat)[1] <- "Series"

  .Deprecated(
    new = "read_abs_local()",
    old = "read_abs_metadata()",
    msg = "read_abs_metadata() is deprecated and will be removed in a future version. Please use read_abs_local()
              to read tidy versions of locally-stored time series spreadsheets, or read_abs() to get them from the ABS website."
  )


  return(final_dat)
}

local_path <- "../testdata"
local_filename <- "6202021.xls"


# These functions are deprecated; they date to the pre-0.3.0 version of the package

test_that("Old read_abs_data() function imports a spreadsheet", {
  filepath <- paste0(local_path, "/", local_filename)

  expect_warning(read_abs_data(filepath, "Data1"))

  local_file <- suppressWarnings(read_abs_data(filepath, "Data1"))

  expect_is(local_file, "data.frame")

  expect_equal(length(colnames(local_file)), 3)
})

test_that("Old read_abs_metadata() function imports a spreadsheet", {
  filepath <- paste0(local_path, "/", local_filename)

  expect_warning(read_abs_metadata(filepath, "Data1"))

  local_file <- suppressWarnings(read_abs_metadata(filepath, "Data1"))

  expect_is(local_file, "data.frame")

  expect_equal(length(colnames(local_file)), 9)
})

---
title: "Using readabs"
output: 
  rmarkdown::html_vignette:
    toc: true
author: "Matt Cowgill"
vignette: >
  %\VignetteIndexEntry{Using readabs}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r setup, echo = FALSE, message = FALSE}
library(knitr)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.path = "VIGNETTE-"
)

set.seed(42)
```

# About {readabs}

The `{readabs}` package helps you work with two types of data released by the Australian Bureau of Statistics: **time series spreadsheets** and **data cubes**.

All ABS time series spreadsheets look like this -- with the series in the first row, then 9 further rows of metadata, then the data. 

```{r out.width = "100%", echo = FALSE}
include_graphics("VIGNETTE-ts-example.png")
```

Data cubes can look like anything -- if it's not formatted like the screenshot above, it's a data cube.

This vignette provides information about how to use `{readabs}` to work with each of these data types.

# Working with ABS time series data

One key function -- `read_abs()` -- downloads, imports, and tidies ABS time series data.

## ABS time series data is messy

If you want to visualise or analyse data in R, you will often need to tidy it first. In [tidy data](http://vita.had.co.nz/papers/tidy-data.html):

1.  Each variable forms a column.
2.  Each observation forms a row.
3.  Each type of observational unit forms a table.

ABS time series data is not tidy. Tidying it requires a bit of work. This screenshot of an ABS time series spreadsheet shows some of the problems, namely:

-   metadata and data are in the same columns;
-   in some cases, the data are spread across multiple worksheets;
-   each time series has its own column (it's 'wide', not 'long'); and
-   dates are in an Excel format (eg. `Feb-1978` is stored as `28522`), which is a pain to convert to an actual date.

```{r out.width = "100%", echo = FALSE}
include_graphics("VIGNETTE-spreadsheet-screenshot.png")
```

## Download, import, and tidy time series data

`read_abs()` is the main function of the `{readabs}` package. First, let's load `{readabs}` and a couple of other packages we'll use for demonstration purposes:
```{r library, message=FALSE}
library(readabs)
library(dplyr)
library(ggplot2)
```

### Getting a whole catalogue number worth of time series data

If you give `read_abs()` an ABS catalogue number, it will download, import and tidy all the time series spreadsheets from that catalogue number. Easy!

For example, to get all the spreadsheets from the Wage Price Index, catalogue number 6345.0, we'd do:

```{r read-wpi-all, eval = FALSE}
wpi <- read_abs("6345.0")
```

```{r create-local-wpi, include=FALSE, eval=FALSE}
wpi <- read_abs("6345.0")
wpi <- head(wpi)
saveRDS(wpi, "wpi.rds")
```

```{r load-local-wpi, include = FALSE}
wpi <- readRDS("wpi.rds")
```

Cool! Now we've got a data frame (a tibble, to be precise) that contains all the time series from the Wage Price Index, converted to long and stacked on top of each other. Here's what it looks like:

```{r glimpse-wpi}
head(wpi)
```

It's over 54 000 rows long, and 12 variables wide. Some catalogue numbers are much bigger - for example, if you get the entire monthly Labour Force release (catalogue number 6202.0), you'll have a data frame with well over 2 million rows.

All the metadata from the time series spreadsheets is included in the data frame:

-   `table_title` is the title of the table;
-   `date` is the date of the observation in that row;
-   `series` is the name of the individual time series - in the ABS spreadsheet this is in the first row;
-   `value` is the observation, the actual data;
-   `series_type` can be 'Original', 'Seasonally Adjusted', or 'Trend';
-   `data_type` tells us whether this is an index number, a 'stock', a 'flow', expressed as a 'percent', etc.;
-   `collection_month` tells us (for quarterly or annual data) which month the data was collected;
-   `frequency` tells us the frequency of the time series;
-   `series_id` is a unique identifier given by the ABS to each time series; and
-   `unit` tells us the unit of measurement, such as '000s', 'Index Numbers', 'Percent' or '000 hours'.

The `table_no` and `sheet_no` columns will help you if you need to cross-check information on the ABS spreadsheet - `table_no` matches the filename of the spreadsheet (eg. '634501.xls') and `sheet_no` is the name of the Excel worksheet within the file that contains the time series.

### Get individual table(s)

Unless you tell it otherwise, `read_abs()` will get all the time series spreadsheets from a given catalogue number.

Generally this will be overkill. Maybe you don't want all ~2.1 million rows of Labour Force data; perhaps you know that the time series you need is in table 1. In that case you can use the `tables` argument to `read_abs()` to specify the table(s) you want:

```{r create-local-lfs, include=FALSE, eval=FALSE}
lfs_1 <- read_abs("6202.0", tables = 1)
lfs_1 <- head(lfs_1)
saveRDS(lfs_1, "lfs_1.rds")

lfs_5 <- read_abs("6202.0", tables = 5)
# lfs_5 <- head(lfs_5)
saveRDS(lfs_5, "lfs_5.rds")
```

```{r read-local-lfs_1, include=FALSE}
lfs_1 <- readRDS("lfs_1.rds")
```

```{r read-lfs-1, eval = FALSE}
lfs_1 <- read_abs("6202.0", tables = 1)
```

```{r glimpse_lfs_1}
head(lfs_1)
```

If you want more than one table, but not the whole catalogue number, you can specify multiple tables:

```{r read-local-lfs_1_t, include=FALSE}
lfs_5 <- readRDS("lfs_5.rds")
lfs_1_5 <- bind_rows(lfs_1, lfs_5)
```

```{r read-lfs-1-5, eval=FALSE}
lfs_1_5 <- read_abs("6202.0", tables = c(1, 5))
```

```{r glimpse_lfs_1_5}
head(lfs_1_5)
```

### Get time series using their IDs

Every ABS time series has a unique identifier. If you know the ID for the time series you need, you can supply it to the `series_id` argument to `read_abs()`. You can give it a single ID, or multiple IDs.

For example, the time series ID for the number of employed people in Australia (trend) is `"A84423127L"`. We can get a data frame that only contains that series as follows:

```{r create-local-seriesid, eval=FALSE, include=FALSE}
employed <- read_abs(series_id = "A84423127L")
employed <- head(employed)
saveRDS(employed, "employed.rds")
```

```{r read-local-seriesid, include = FALSE}
employed <- readRDS("employed.rds")
```

```{r read-abs-seriesid, eval = FALSE}
employed <- read_abs(series_id = "A84423127L")
```

```{r glimpse-seriesid}
head(employed)

unique(employed$series)
```

A benefit of this approach is that the filtering is done for you - the data frame only contains the series you're interested in. 

## Use `separate_series()` to work with your time series data

The `series` column of a table imported using `read_abs()` often contains a long, multi-part string, as this is how the ABS labels its series. This is where the `separate_series()` function comes in handy.

In this example, we'll work with data from the Labour Force survey that we downloaded and tidied earlier using `read_abs()`.

Now let's have a look at the time series from table 1 of the Labour Force survey:

```{r examine-lfs}
unique(lfs_1$series)
```

There's a bunch of data in here. We can see that the `series` column contains a first part, like "Participation rate" or "Unemployment rate", and a second part, which is "Persons", "Males", or "Females". You will often want to filter using one or both of these components separately - and sometimes there will be more than two components.

The `separate_series` function helps you by separating the `series` column into multiple components and removing extraneous symbols.

```{r separate-series}
lfs_1_sep <- lfs_1 %>%
  separate_series()

lfs_1_sep %>%
  group_by(series_1, series_2) %>%
  summarise()
```

The `remove_totals` and `remove_nas` arguments to `separate_series()` can be very useful when you are tidying data. They're both set to `FALSE` by default.

We've now got a new tibble that still contains the original `series` column, but also contains new columns `series_1` and `series_2`, with the two parts of `series` split into these new columns.

Let's make a data frame that just contains the male and female unemployment rates over time.

```{r create-unemp-df}
unemp <- lfs_1_sep %>%
  filter(series_1 == "Unemployment rate")

unique(unemp$series_1)

unique(unemp$series_2)
```

Now we have a data frame, `unemp`, that contains various unemployment rate series. Let's filter to only Males or Females.

```{r filter-male-female}
unemp <- unemp %>%
  filter(series_2 %in% c("Males", "Females"))

unique(unemp$series_2)
```

Now our data frame only contains the male and female unemployment rates, which is what we want. Let's graph it, filtering once more to show only the seasonally adjusted series and adding a 'sex' column:

```{r graph-unemp, dpi = 200}
unemp %>%
  filter(series_type == "Seasonally Adjusted") %>%
  mutate(sex = series_2) %>%
  ggplot(aes(x = date, y = value, col = sex)) +
  geom_line() +
  theme_minimal() +
  theme(
    legend.position = "bottom",
    axis.title = element_blank(),
    legend.title = element_blank(),
    text = element_text(size = 5)
  ) +
  labs(
    title = "The male and female unemployment rates have converged",
    subtitle = "Unemployment rates for Australian men and women (aged 15+), 1978-2018 (per cent)",
    caption = "Source: ABS 6202.0"
  )
```

Ta-da! Now we've got a nice little graph - and you didn't need to go to the ABS website or click around in Excel.


## Importing and tidying local time series spreadsheets

If you already have ABS time series spreadsheets saved locally that you want to read, the `read_abs_local()` function is what you want.

If you've downloaded files using `read_abs()`, you can import them using `read_abs_local()` by specifying the catalogue number. This will look in the subdirectory of `path` that corresponds to `cat_no`:

```{r read-lfs-local-catno, eval = FALSE}
lfs_local_1 <- read_abs_local("6202.0")
```

The data frame you'll get will look the same as if you'd used `read_abs()` to get the spreadsheet(s) from the ABS website.

# Working with ABS data cubes

All ABS time series spreadsheets are alike. Each ABS data cube is formatted in its own unique way.

There is no data cube equivalent of the `read_abs()` function that will download, import, and tidy any data cube for you. But `download_abs_data_cube()` can download any data cube, for you to then manually import and tidy. Use `show_available_catalogues()` to see all the 
possible catalogues, `search_catalogues()` to find catalogues that contain a given string (in the catalogue name itself, or the topic). Similarly, use `show_available_files()` to find all the filenames in a given catalogue, or `search_files()` to look for filenames within a catalogue that contain a particular string.

Convenience functions - `read_payrolls()` and `read_lfs_grossflows()` are included to download, import, and tidy particular commonly-used data cubes.

## Download data cubes

Any ABS data cube can be downloaded using the `download_abs_data_cube()` function. You need to specify the name of the catalogue (`catalogue_string`), and the filename (in whole or unique part) of the file you want to download (`cube`).

Let's say I want to download cube GM1, the gross flows spreadsheet from the monthly Labour Force release.

Let's look for the name of the catalogue:

```{r, eval = FALSE}
search_catalogues("labour force")
```
```{r, eval = FALSE, include = FALSE}
cats <- search_catalogues("labour force")
saveRDS(cats, "cats.rds")
```
```{r, echo = FALSE}
readRDS("cats.rds")
```

It looks like `labour-force-australia` is the catalogue name for the monthly Labour Force survey.

Now I need to find the filename for the gross flows spreadsheet, GM1.

```{r, eval = FALSE}
search_files("GM1", "labour-force-australia")
```
```{r, echo = FALSE}
x <- "GM1.xlsx"
x
```

The GM1 table is, conveniently enough, called `GM1.xlsx`. 

So we can download it like this:

```{r, eval=FALSE}
gm1_path <- download_abs_data_cube("labour-force-australia", "GM1")

print(gm1_path)
```
```{r, include=FALSE}
print("/var/folders/_4/ngvkm2811nbd8b_v66wytw1r0000gn/T//RtmpZT2ffU/GM1.xlsx")
```

Now we have an Excel sheet saved at that path, which you can import using `readxl::read_excel()` or the Excel-reading function of your choice.

## Download, import and tidy key data cubes
For most data cubes, you'll need to write your own code to import and tidy the data. But for some key cubes, convenience functions are provided.

### Labour force gross flows

The labour force gross flows cube is one of the cubes for which `readabs` includes a convenience function, `read_lfs_grossflows()`.

```{r, eval=FALSE, include=FALSE}
gf <- read_lfs_grossflows()
gf <- head(gf)
saveRDS(gf, "gf.rds")
```
```{r, include=FALSE}
gf <- readRDS("gf.rds")
```
```{r, eval=FALSE}
gf <- read_lfs_grossflows()
```

Now we have a tidy tibble of the Labour Force gross flows data:

```{r}
head(gf)
```

### ABS Weekly Payroll Jobs and Wages

The ABS Weekly Payroll Jobs and Wages release contains a range of useful information. Even though this release contains time series data, the ABS does not release it in its standard time series spreadsheet format, and `read_abs()` therefore cannot import it. The `read_payrolls()` function can be used instead.

Let's say you want to download the jobs figures, as index numbers, for each geographical area (SA3) over time. That's straightforward with `read_payrolls()`:

```{r, eval=FALSE, include=FALSE}
payrolls <- read_payrolls("sa3_jobs")
payrolls <- head(payrolls)
saveRDS(payrolls, "payrolls.rds")
```
```{r, include=FALSE}
payrolls <- readRDS("payrolls.rds")
```
```{r, eval=FALSE}
payrolls <- read_payrolls()
```

Here's what we've got:

```{r}
head(payrolls)
```


# Choosing where ABS data is saved

The `read_abs()` function downloads spreadsheets from the ABS website to your machine, then loads them into R, then tidies them.

By default, the spreadsheets will be saved in the directory specified by the environment variable `R_READABS_PATH`. If this variable isn't set, the spreadsheets are saved to a temporary directory. You can change this location for a specific instance by using the `path` argument to `read_abs()`.

Keeping the spreadsheets is often a good idea. But if you don't want to retain them, no problem. Just specify `retain_files = FALSE` when you use `read_abs()`. The spreadsheets will still be downloaded, but they'll be stored in a temporary directory that is flushed when your R session ends.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_erp.R
\name{read_erp}
\alias{read_erp}
\title{Download a tidy tibble containing the Estimated Residential Population from the ABS}
\usage{
read_erp(
  age_range = 0:100,
  sex = "Persons",
  states = c("Australia", "New South Wales", "Victoria", "Queensland", "South Australia",
    "Western Australia", "Tasmania", "Northern Territory",
    "Australian Capital Territory"),
  path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
  show_progress_bars = TRUE,
  check_local = FALSE,
  retain_files = FALSE
)
}
\arguments{
\item{age_range}{numeric; default is "0:100". A vector containing ages in single
years for which an ERP is sought. The ABS top-code ages at 100.}

\item{sex}{character; default is "Persons". Other values are "Male" and
"Female". Multiple values allowed.}

\item{states}{character; default is "Australia". Other values are the full
or abbreviated names of the states and self-governing territories. Multiple
values allowed.}

\item{path}{character; default is "data/ABS". Only used if
retain_files is set to TRUE. Local directory in which to save
downloaded ABS time series spreadsheets.}

\item{show_progress_bars}{logical; TRUE by default. If set to FALSE, progress
bars will not be shown when ABS spreadsheets are downloading.}

\item{check_local}{logical; FALSE by default. See \code{?read_abs}.}

\item{retain_files}{logical; FALSE by default. When TRUE, the spreadsheets
downloaded from the ABS website will be saved in the
directory specified with 'path'.}
}
\description{
\code{read_erp()} uses the \code{read_abs()} function to download, import,
and tidy the Estimated Residential Population from the ABS. It allows the user
to specify age, sex and states/territories of interest. It returns a tibble
containing five columns: the date, the age range, sex and states that the ERP
corresponds to. This makes joining the ERP to another dataframe easy.
}
\examples{
\donttest{
# Create a tibble called 'erp' that contains the ERP index
# numbers for 30 June each year for Australia.

erp <- read_erp()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidy_abs_list.R
\name{tidy_abs_list}
\alias{tidy_abs_list}
\title{Tidy multiple dataframes of ABS time series data contained in a list.}
\usage{
tidy_abs_list(list_of_dfs, metadata = TRUE)
}
\arguments{
\item{list_of_dfs}{A list of dataframes containing extracted
ABS time series data.}

\item{metadata}{logical. If \code{TRUE} (the default), a tidy data frame including
ABS metadata (series name, table name, etc.) is included in the output. If
\code{FALSE}, metadata is dropped.}
}
\description{
Tidy multiple dataframes of ABS time series data contained in a list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scrape_abs_catalogues.r
\name{scrape_abs_catalogues}
\alias{scrape_abs_catalogues}
\title{Helper function for \code{download_abs_data_cube} to scrape the available catalogues from the ABS website.}
\usage{
scrape_abs_catalogues()
}
\value{
A tibble containing the catalogues and how they are organised on the ABS website.
}
\description{
This function downloads a new version of the lookup table used by \code{show_available_catalogues}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_lfs_grossflows.R
\name{check_lfs_grossflows}
\alias{check_lfs_grossflows}
\title{Internal function to check if the data frame returned by read_lfs_grossflows()
contains expected unique values in key columns}
\usage{
check_lfs_grossflows(df)
}
\arguments{
\item{df}{data frame containing gross flows data}
}
\description{
Internal function to check if the data frame returned by read_lfs_grossflows()
contains expected unique values in key columns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_abs_sheets.R
\name{extract_abs_sheets}
\alias{extract_abs_sheets}
\title{Extract data sheets from an ABS timeseries workbook saved locally as an
Excel file.}
\usage{
extract_abs_sheets(
  filename,
  table_title = NULL,
  path = Sys.getenv("R_READABS_PATH", unset = tempdir())
)
}
\arguments{
\item{filename}{Filename for an ABS time series spreadsheet (as string)}

\item{table_title}{String giving the full title of the ABS table, such as
"Table 1. Employed persons, Australia"}

\item{path}{Local directory in which an ABS time series is stored. Default is
\code{Sys.getenv("R_READABS_PATH", unset = tempdir())}.}
}
\description{
Note that this function will not tidy the data for you.
Use \code{read_abs_local()}to import and tidy data from local ABS time series
spreadsheets or \code{read_abs()} to download, import and tidy ABS time series.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_cpi.R
\name{read_cpi}
\alias{read_cpi}
\title{Download a tidy tibble containing the Consumer Price Index from the ABS}
\usage{
read_cpi(
  frequency = c("quarterly", "monthly"),
  path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
  ...
)
}
\arguments{
\item{frequency}{Either \code{"quarterly"} (the default) or \code{"monthly"}.}

\item{path}{character; default is "data/ABS". Only used if
retain_files is set to TRUE. Local directory in which to save
downloaded ABS time series spreadsheets.}

\item{...}{Arguments passed to \code{read_abs()}. See \code{?read_abs} for details.}
}
\description{
\code{read_cpi()} uses the \code{read_abs()} function to download, import,
and tidy the quarterly Consumer Price Index from the ABS. It returns a tibble
containing two columns: the date and the CPI index value that corresponds
to that date. This makes joining the CPI to another dataframe easy.
\code{read_cpi()} returns the original (ie. not seasonally adjusted)
all groups CPI for Australia. If you want the analytical series
(eg. seasonally adjusted CPI, or trimmed mean CPI) or monthly series, you can use
\code{read_abs()}.
}
\details{
By default, \code{read_abs()} retrieves the quarterly CPI. It can use the
\code{frequency} argument to change to monthly.
}
\examples{

# Create a tibble called 'cpi' that contains the CPI index
# numbers for each quarter
\donttest{
cpi <- read_cpi()
}

# This tibble can now be joined to another to help streamline the process of
# deflating nominal values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_api.R
\name{abs_api}
\alias{abs_api}
\alias{read_api_dataflows}
\alias{read_api}
\alias{read_api_url}
\alias{read_api_datastructure}
\title{ABS.Stat API functions}
\usage{
read_api_dataflows()

read_api(
  id,
  datakey = NULL,
  start_period = NULL,
  end_period = NULL,
  version = NULL
)

read_api_url(url)

read_api_datastructure(id)
}
\arguments{
\item{id}{A dataflow id. Use \code{read_api_dataflows()} to obtain a dataframe
listing available dataflows.}

\item{datakey}{A named list matching filter variables to codes. All variables
with a \code{position} in the datastructure are filterable. Use
\code{read_api_datastructure()} to obtain information about the variables in
a dataflow and the values of that variable.}

\item{start_period}{The start period (used to filter by time). This is
inclusive. The supported formats are:
\itemize{
\item \code{"YYYY"} for annual data (e.g. 2019)
\item \code{"YYYY-S[1-2]"} for semi-annual data (e.g. 2019-S1)
\item \code{"YYYY-Q[1-4]"} for quarterly data (e.g. 2019-Q1)
\item \code{"YYYY-MM[01-12]"} for monthly data (e.g. 2019-01)
\item \code{"YYYY-W[01-53]"} for weekly data (e.g. 2019-W01)
\item \code{"YYYY-MM-DD"} for daily and business data (e.g. 2019-01-01)
}}

\item{end_period}{The end period (used to filter on time). This is inclusive.
The supported formats are the same as for \code{start_period}}

\item{version}{A version number, if unspecified the latest version of the
dataset is used. Use \code{read_api_dataflows()} to see
available dataflow versions.}

\item{url}{A complete query url}
}
\value{
A data.frame
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

These experimental functions provide a minimal interface to the ABS.Stat API.

More information on the ABS.Stat API can be found
on the \href{https://www.abs.gov.au/about/data-services/application-programming-interfaces-apis/data-api-user-guide/using-api}{ABS website}

Note that an ABS.Stat 'dataflow' is like a table. A 'datastructure'
contains metadata that describes the variables in the dataflow. To load data
from the ABS.Stat API, you need to either:
\itemize{
\item Using \code{read_api_dataflows()} you can get information on the available dataflows
\item Using \code{read_api_datastructure()} you can get metadata relating to a
specific dataflow, including the variables available in each dataflow
\item Using \code{read_api()} you can get the data belonging to a given dataflow.
\item Using \code{read_api_url()} you can get the data for a given query url
generated using the \href{https://dataexplorer.abs.gov.au}{online data viewer}.
}
}
\details{
Note that the API enforces a reasonably strict gateway timeout policy. This
means that, if you're trying to access a reasonably large dataset, you will
need to filter it on the server side using the \code{datakey}. You might like to
review the data manually via the \href{https://dataexplorer.abs.gov.au}{ABS website}
to figure out what subset of the data you require.

Note, furthermore, that the datastructure contains a complete codebook for
the variables appearing in the relevant dataflow. Since some variables are
shared across multiple dataflows, this means that the datastructure
corresponding to a particular \code{id} may contain values for a given variable
which are not in the corresponding dataflow.
}
\examples{
\dontrun{
# List available dataflows
read_api_dataflows()

# Say we want the "Estimated resident population, Country of birth"
# data flow, with the id ERP_COB. We load the data like this:
# Get full data set for a given flow by providing id and start period:
read_api("ERP_COB", start_period = 2020)

# In some cases, loading a whole dataflow (as above) won't work.
# For eg., the `ABS_C16_T10_SA` dataflow is very large,
# so the gateway will timeout if we try to collect the full data set
try(read_api("ABS_C16_T10_SA"))

# We need to filter the dataflow before downlaoding it.
# To figure out how to filter it, we get metadata ('datastructure').
ds <- read_api_datastructure("ABS_C16_T10_SA")

# The `asgs_2016` code for 'Australia' is 0
ds[ds$var == "asgs_2016" & ds$label == "Australia", ]

# The `sex_abs` code for 'Persons' (i.e. all persons) is 3
ds[ds$var == "sex_abs" & ds$label == "Persons", ]

# So we have:
x <- read_api("ABS_C16_T10_SA", datakey = list(asgs_2016 = 0, sex_abs = 3))
unique(x["asgs_2016"]) # Confirming only 'Australia' level records came through
unique(x["sex_abs"]) # Confirming only 'Persons' level records came through

# Please note however that not all values in the datastructure necessarily
# appear in the data. You get 404s in this case
ds[ds$var == "regiontype" & ds$label == "Destination Zones", ]
try(read_api("ABS_C16_T10_SA", datakey = list(regiontype = "DZN")))

# If you already have a query url, then use `read_api_url()`
wpi_url <- "https://data.api.abs.gov.au/rest/data/ABS,WPI/all"
read_api_url(wpi_url)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_abs_data.R
\name{read_abs_metadata}
\alias{read_abs_metadata}
\title{Extracts ABS series metadata directly from Excel spreadsheets and
converts to long-form.}
\usage{
read_abs_metadata(path, sheet)
}
\arguments{
\item{path}{Filepath to Excel spreadsheet.}

\item{sheet}{Sheet name or number.}
}
\value{
Long-form dataframe
}
\description{
Extracts ABS series metadata directly from Excel spreadsheets and
converts to long-form.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search_files.R
\name{search_files}
\alias{search_files}
\title{Search for a file within an ABS catalogue}
\usage{
search_files(string, catalogue, refresh = FALSE)
}
\arguments{
\item{string}{String to search for among filenames in a catalogue}

\item{catalogue}{Name of catalogue}

\item{refresh}{logical; \code{FALSE} by default. When \code{TRUE}, will re-scrape the
list of files within the catalogue.}
}
\description{
Search for a file within an ABS catalogue
}
\examples{
\dontrun{
search_files("GM1", "labour-force-australia")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search_catalogues.R
\name{search_catalogues}
\alias{search_catalogues}
\title{Search for ABS catalogues that match a string}
\usage{
search_catalogues(string, refresh = FALSE)
}
\arguments{
\item{string}{Character. A word or phrase you want to search for, such as "labour" or
"union". Not case sensitive.}

\item{refresh}{Logical. \code{FALSE} by default. If \code{TRUE}, will re-scrape the ABS
website to ensure that the list of catalogues is up-to-date.}
}
\value{
A data frame (tibble) containing the topic (\code{heading}), product title
(\code{sub_heading}), catalogue (\code{catalogue}) and URL (\code{URL}) of any catalogues
that match the provided string.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Helper function to use with \code{download_abs_data_cube()}.

\code{download_abs_data_cube()} requires that you specify a \code{catalogue}.
\code{search_catalogues()} helps you find the catalogue you want, by searching for
a given string in the catalogue names, product title, and broad topic.
}
\examples{

search_catalogues("labour")
}
\seealso{
Other data cube functions:
\code{\link[=download_abs_data_cube]{download_abs_data_cube()}},
\code{\link[=show_available_catalogues]{show_available_catalogues()}},
\code{\link[=show_available_files]{show_available_files()}}
}
\concept{data cube functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_lfs_grossflows.R
\name{read_lfs_grossflows}
\alias{read_lfs_grossflows}
\title{Download, import and tidy 'gross flows' data cube
from the monthly ABS Labour Force survey.}
\usage{
read_lfs_grossflows(
  weights = c("current", "previous"),
  path = Sys.getenv("R_READABS_PATH", unset = tempdir())
)
}
\arguments{
\item{weights}{either \code{"current"} or \code{"previous"}. If \code{"current"}, figures will
use the current month's Labour Force survey weights; if \code{"previous"}, the
previous month's weights are used.}

\item{path}{Local directory in which downloaded files should be stored.
By default, 'path' takes the value set in the environment variable
"R_READABS_PATH". If this variable is not set, any files downloaded
will be stored in a temporary directory (\code{tempdir()}).
See \code{Details} in \code{?read_abs} for more information.}
}
\value{
A tibble containing data cube GM1 from the monthly Labour Force survey.
}
\description{
This convenience function downloads, imports and tidies the 'gross flows' data
cube from the monthly ABS Labour Force survey. The gross flows data cube (GM1)
shows estimates of the number of people who
transitioned from one labour force status to another between two months.
}
\examples{
\dontrun{
read_lfs_grossflows()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/show_available_catalogues.r
\name{show_available_catalogues}
\alias{show_available_catalogues}
\title{Helper function for \code{download_abs_data_cube} to show the available catalogues.}
\usage{
show_available_catalogues(selected_heading = NULL, refresh = FALSE)
}
\arguments{
\item{selected_heading}{optional character string specifying the heading on the \href{https://www.abs.gov.au/statistics}{ABS statistics webpage}.
e.g. "Earnings and work hours"}

\item{refresh}{logical; \code{FALSE} by default. If \code{FALSE}, an internal table of the available ABS catalogues is used. If \code{TRUE}, this table is refreshed from the ABS website.}
}
\value{
a character vector of catalogues.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

This function lists the possible catalogues that are available on the ABS website.
These catalogues must be specified as a string as an argument to \code{download_abs_data_cube}.
}
\examples{
show_available_catalogues("Earnings and work hours")
}
\seealso{
Other data cube functions:
\code{\link[=download_abs_data_cube]{download_abs_data_cube()}},
\code{\link[=search_catalogues]{search_catalogues()}},
\code{\link[=show_available_files]{show_available_files()}}
}
\concept{data cube functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_abs_local.R
\name{read_abs_local}
\alias{read_abs_local}
\title{Read and tidy locally-saved ABS time series spreadsheet(s)}
\usage{
read_abs_local(
  cat_no = NULL,
  filenames = NULL,
  path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
  use_fst = TRUE,
  metadata = TRUE
)
}
\arguments{
\item{cat_no}{character; a single catalogue number such as "6202.0".
When \code{cat_no} is specified, all local files in \code{path} corresponding to
the specified catalogue number will be imported.
For example, if you run \code{read_abs_local("6202.0")}, it will look in
the \code{6202.0} sub-folder of \code{path} and attempt to load any
.xls and .xlsx files in that location.
If \verb{cat_no`` is specified, }filenames` will be ignored.}

\item{filenames}{character vector of at least one filename of a
locally-stored ABS time series spreadsheet. For example, "6202001.xls" or
c("6202001.xls", "6202005.xls"). Ignored if a value is supplied to \code{cat_no}.
If \code{filenames} is blank and \code{cat_no} is blank, \code{read_abs_local()} will
attempt to read all .xls and .xlsx files in the directory specified with \code{path}.}

\item{path}{path to local directory containing ABS time series file(s).
Default is \code{Sys.getenv("R_READABS_PATH", unset = tempdir())}.
If nothing is specified in \code{filenames} or \code{cat_no},
\code{read_abs_local()} will attempt to read all .xls and .xlsx files in the directory
specified with \code{path}.}

\item{use_fst}{logical. If \code{TRUE} (the default) then, if an \code{fst} file of the
tidy data frame has already been saved in \code{path}, it is read immediately.}

\item{metadata}{logical. If \code{TRUE} (the default), a tidy data frame including
ABS metadata (series name, table name, etc.) is included in the output. If
\code{FALSE}, metadata is dropped.}
}
\description{
If you need to download and tidy time series data from the ABS,
use \code{read_abs()}. \code{read_abs_local()} imports and tidies data
from ABS time series spreadsheets that are already saved to your local drive.
}
\details{
Unlike \code{read_abs()}, the \code{table_title} column in the data frame
returned by \code{read_abs_local()} is blank. If you require \code{table_title},
please use \code{read_abs()} instead.
}
\examples{

# Load and tidy two specified files from the "data/ABS" subdirectory
# of your working directory
\dontrun{
lfs <- read_abs_local(c("6202001.xls", "6202005.xls"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_payrolls.R
\name{read_payrolls}
\alias{read_payrolls}
\title{Download and tidy ABS payroll jobs and wages data}
\usage{
read_payrolls(
  series = c("industry_jobs", "subindustry_jobs", "empsize_jobs"),
  path = Sys.getenv("R_READABS_PATH", unset = tempdir())
)
}
\arguments{
\item{series}{Character. Must be one of:
\describe{
\item{"industry_jobs"}{ Payroll jobs by industry division, state, and age
group (Table 1)}
\item{"subindustry_jobs"}{ Payroll jobs by industry sub-division and
industry division (Table 2)}
\item{"empsize_jobs"}{ Payroll jobs by size of employer (number of
employees) and state/territory (Table 3)}
}
The default is "industry_jobs".}

\item{path}{Local directory in which downloaded ABS time series
spreadsheets should be stored. By default, \code{path} takes the value set in the
environment variable "R_READABS_PATH". If this variable is not set,
any files downloaded by read_abs()  will be stored in a temporary directory
(\code{tempdir()}).}
}
\value{
A tidy (long) \code{tbl_df}. The number of columns differs based on the \code{series}.
}
\description{
Import a tidy tibble of ABS Payroll Jobss data.
}
\details{
The ABS \href{https://www.abs.gov.au/statistics/labour/jobs/payroll-jobs/latest-release}{Payroll Jobs}
dataset draws upon data collected
by the Australian Taxation Office as part of its Single-Touch Payroll
initiative and supplements the monthly Labour Force Survey. Unfortunately,
the data as published by the ABS (1) is not in a standard time series
spreadsheet; and (2) is messy in various ways that make it hard to
read in R. This convenience function uses \code{download_abs_data_cube()} to
import the payrolls data, and then tidies it up.

Note that this ABS release used to be called Weekly Payroll Jobs and Wages Australia.
The total wages series were removed from this release in mid-2023 and it
was renamed to Weekly Payroll Jobs. The ability to read total wages
indexes using this function was therefore also removed. It was then renamed
Payroll Jobs and the frequency was reduced, with further modifications to
the data released.
}
\examples{
\dontrun{
# Fetch payroll jobs by industry and state (the default, "industry_jobs")
read_payrolls()

# Payroll jobs by employer size
read_payrolls("empsize_jobs")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_cube.r
\name{download_abs_data_cube}
\alias{download_abs_data_cube}
\title{Experimental helper function to download ABS data cubes that are not compatible with read_abs.}
\usage{
download_abs_data_cube(
  catalogue_string,
  cube,
  path = Sys.getenv("R_READABS_PATH", unset = tempdir())
)
}
\arguments{
\item{catalogue_string}{ABS catalogue name as a string from the ABS website.
For example, Labour Force, Australia, Detailed is "labour-force-australia-detailed".
The possible catalogues can be obtained using the helper function \code{show_available_catalogues()};
or search these catalogues using \code{search_catalogues()},}

\item{cube}{character. A character string that is either the complete filename or (uniquely) in the filename of the data cube you want to
download, e.g. "EQ09". The available filenames can be obtained using the helper function \code{get_available_files()}}

\item{path}{Local directory in which downloaded files should be stored. By default, \code{path}
takes the value set in the environment variable "R_READABS_PATH".
If this variable is not set, any files downloaded
will be stored in a temporary directory (\code{tempdir()}).
See \code{Details} below for  more information.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
\code{download_abs_data_cube()} downloads the latest ABS data cubes based on the catalogue name (from the website url) and cube.
The function downloads the file to disk.

Unlike \code{read_abs()}, this function doesn't import or tidy the data.
Convenience functions are provided to import and tidy key data cubes; see
\code{?read_payrolls()} and \code{?read_lfs_grossflows()}.
}
\details{
\code{download_abs_data_cube()} downloads an Excel spreadsheet from the ABS.

The file need to be saved somewhere on your disk.
This local directory can be controlled using the \code{path} argument to
\code{read_abs()}. If the \code{path} argument is not set, \code{read_abs()} will store
the files in a directory set in the "R_READABS_PATH" environment variable.
If this variable isn't set, files will be saved in a temporary directory.

To check the value of the "R_READABS_PATH" variable, run
\code{Sys.getenv("R_READABS_PATH")}. You can set the value of this variable
for a single session using \code{Sys.setenv(R_READABS_PATH = <path>)}.
If you would like to change this variable for all future R sessions, edit
your \code{.Renviron} file and add \code{R_READABS_PATH = <path>} line.
The easiest way to edit this file is using \code{usethis::edit_r_environ()}.

The filepath is returned invisibly which enables piping to \code{unzip()} or \code{readxl::read_excel}.
}
\examples{
\dontrun{
download_abs_data_cube(
  catalogue_string = "labour-force-australia-detailed",
  cube = "EQ09"
)
}

}
\seealso{
Other data cube functions:
\code{\link[=search_catalogues]{search_catalogues()}},
\code{\link[=show_available_catalogues]{show_available_catalogues()}},
\code{\link[=show_available_files]{show_available_files()}}
}
\concept{data cube functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_abs_local.R
\name{read_abs_url}
\alias{read_abs_url}
\title{Download and import an ABS time series spreadsheet from a given URL}
\usage{
read_abs_url(
  url,
  path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
  show_progress_bars = TRUE,
  ...
)
}
\arguments{
\item{url}{Character vector of url(s) to ABS time series spreadsheet(s).}

\item{path}{Local directory in which downloaded ABS time series
spreadsheets should be stored. By default, \code{path} takes the value set in the
environment variable "R_READABS_PATH". If this variable is not set,
any files downloaded by read_abs()  will be stored in a temporary directory
(\code{tempdir()}). See \code{?read_abs()} for more.}

\item{show_progress_bars}{TRUE by default. If set to FALSE, progress bars
will not be shown when ABS spreadsheets are downloading.}

\item{...}{Additional arguments passed to \code{read_abs_local()}.}
}
\description{
Download and import an ABS time series spreadsheet from a given URL
}
\details{
If you have a specific URL to the time series spreadsheet you wish
to download, \code{read_abs_url()} will download, import and tidy it. This is
useful for older vintages of data, or discontinued data.
}
\examples{
\dontrun{
url <- paste0(
  "https://www.abs.gov.au/statistics/labour/",
  "employment-and-unemployment/labour-force-australia/aug-2022/6202001.xlsx"
)
read_abs_url(url)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidy_abs.R
\name{tidy_abs}
\alias{tidy_abs}
\title{Tidy ABS time series data.}
\usage{
tidy_abs(df, metadata = TRUE)
}
\arguments{
\item{df}{A data frame containing ABS time series data
that has been extracted using \code{extract_abs_sheets}.}

\item{metadata}{logical. If \code{TRUE} (the default), a tidy data frame
including ABS metadata (series name, table name, etc.) is
included in the output. If \code{FALSE}, metadata is dropped.}
}
\value{
data frame (tibble) in long format.
}
\description{
Tidy ABS time series data.
}
\examples{

# First extract the data from the local spreadsheet
\dontrun{
wpi <- extract_abs_sheets("634501.xls")
}

# Then tidy the data extracted from the spreadsheet. Note that
# \code{extract_abs_sheets()} returns a list of data frames, so we need to
# subset the list.
\dontrun{
tidy_wpi <- tidy_abs(wpi[[1]])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_job_mobility.R
\name{read_job_mobility}
\alias{read_job_mobility}
\title{Download and tidy ABS Job Mobility tables}
\usage{
read_job_mobility(
  tables = "all",
  path = Sys.getenv("R_READABS_PATH", unset = tempdir())
)
}
\arguments{
\item{tables}{Either \code{"all"} (the default) to import all tables, or a
vector of table numbers, such as \code{1} or \code{c(2, 4)}.}

\item{path}{Local directory in which downloaded ABS time series spreadsheets should be stored. By default, 'path' takes the value set in the environment variable "R_READABS_PATH". If this variable is not set, any files downloaded by read_abs() will be stored in a temporary directory (tempdir()).}
}
\description{
Import a tidy tibble of ABS Job Mobility data
}
\examples{
\dontrun{
# Get all tables from the ABS Job Mobility series
read_job_mobility()

# Get tables 1 and 2
read_job_mobility(c(1, 2))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_cube.r
\name{read_lfs_datacube}
\alias{read_lfs_datacube}
\title{Convenience function to download and tidy data cubes from
ABS Labour Force, Australia, Detailed.}
\usage{
read_lfs_datacube(cube, path = Sys.getenv("R_READABS_PATH", unset = tempdir()))
}
\arguments{
\item{cube}{character. A character string that is either the complete filename
or (uniquely) in the filename of the data cube you want to download. Use
\code{get_available_lfs_cubes()} to see a dataframe of options.}

\item{path}{Local directory in which downloaded files should be stored.}
}
\value{
A tibble with the data from the data cube. Columns names are
tidied and dates are converted to Date class.
}
\description{
Convenience function to download and tidy data cubes from
ABS Labour Force, Australia, Detailed.
}
\examples{
read_lfs_datacube("EQ02")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_abs.R
\name{read_abs}
\alias{read_abs}
\alias{read_abs_series}
\title{Download, extract, and tidy ABS time series spreadsheets}
\usage{
read_abs(
  cat_no = NULL,
  tables = "all",
  series_id = NULL,
  path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
  metadata = TRUE,
  show_progress_bars = TRUE,
  retain_files = TRUE,
  check_local = TRUE,
  release_date = "latest"
)

read_abs_series(series_id, ...)
}
\arguments{
\item{cat_no}{ABS catalogue number, as a string, including the extension.
For example, "6202.0".}

\item{tables}{numeric. Time series tables in \verb{cat_no`` to download and extract. Default is "all", which will read all time series in }cat_no\verb{. Specify }tables\verb{to download and import specific tables(s) - eg.}tables = 1\code{or}tables = c(1, 5)`.}

\item{series_id}{(optional) character. Supply an ABS unique time series
identifier (such as "A2325807L") to get only that series.
This is an alternative to specifying \code{cat_no}.}

\item{path}{Local directory in which downloaded ABS time series
spreadsheets should be stored. By default, \code{path} takes the value set in the
environment variable "R_READABS_PATH". If this variable is not set,
any files downloaded by read_abs()  will be stored in a temporary directory
(\code{tempdir()}). See \code{Details} below for
more information.}

\item{metadata}{logical. If \code{TRUE} (the default), a tidy data frame including
ABS metadata (series name, table name, etc.) is included in the output. If
\code{FALSE}, metadata is dropped.}

\item{show_progress_bars}{TRUE by default. If set to FALSE, progress bars
will not be shown when ABS spreadsheets are downloading.}

\item{retain_files}{when TRUE (the default), the spreadsheets downloaded
from the ABS website will be saved in the directory specified with \code{path}.
If set to \code{FALSE}, the files will be stored in a temporary directory.}

\item{check_local}{If \code{TRUE}, the default, local \code{fst} files are used,
if present.}

\item{release_date}{Either \code{"latest"} or a string coercible to a date, such as
\code{"2022-02-01"}. If \code{"latest"}, the latest release of the requested data will
be returned. If a date, (eg. \code{"2022-02-01"}) \code{read_abs()} will
attempt to download the data from that month's release. Note that this only
works consistently as expected for monthly data. See \code{Details}.}

\item{...}{Arguments to \code{read_abs_series()} are passed to \code{read_abs()}.}
}
\value{
A data frame (tibble) containing the tidied data from the ABS time
series table(s).
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

\code{read_abs()} downloads ABS time series spreadsheets,
then extracts the data from those spreadsheets,
then tidies the data. The result is a single
data frame (tibble) containing tidied data.
}
\details{
\code{read_abs_series()} is a wrapper around \code{read_abs()}, with \code{series_id} as
the first argument.

\code{read_abs()} downloads spreadsheet(s) from the ABS containing time
series data. These files need to be saved somewhere on your disk.
This local directory can be controlled using the \code{path} argument to
\code{read_abs()}. If the \code{path} argument is not set, \code{read_abs()} will store
the files in a directory set in the "R_READABS_PATH" environment variable.
If this variable isn't set, files will be saved in a temporary directory.

To check the value of the "R_READABS_PATH" variable, run
\code{Sys.getenv("R_READABS_PATH")}. You can set the value of this variable
for a single session using \code{Sys.setenv(R_READABS_PATH = <path>)}.
If you would like to change this variable for all future R sessions, edit
your \code{.Renviron} file and add \code{R_READABS_PATH = <path>} line.
The easiest way to edit this file is using \code{usethis::edit_r_environ()}.

Certain corporate networks restrict your ability to download files in an R
session. On some of these networks, the \code{"wininet"} method must be used when
downloading files. Users can now specify the method that will be used to
download files by setting the \code{"R_READABS_DL_METHOD"} environment variable.

For example, the following code sets the environment variable for your
current session: s\code{Sys.setenv("R_READABS_DL_METHOD" = "wininet")}
You can add \code{R_READABS_DL_METHOD = "wininet"} to your .Renviron to have
this persist across sessions.

The \code{release_date} argument allows you to download table(s) other than the
latest release. This is useful for examining revisions to time series, or
for obtaining the version of series that were available on a given date.
Note that you cannot supply more than one date to \code{release_date}. Note also
that any dates prior to mid-2019 (the exact date varies by series) will fail.
Specifying \code{release_date} only reliably works for monthly, and some
quarterly, data. It does not work for annual data.
}
\examples{

# Download and tidy all time series spreadsheets
# from the Wage Price Index (6345.0)
\dontrun{
wpi <- read_abs("6345.0")
}

# Download table 1 from the Wage Price Index
\dontrun{
wpi_t1 <- read_abs("6345.0", tables = "1")
}

# Or table 1 as in the Sep 2019 release of the WPI:
\dontrun{
wpi_t1_sep2019 <- read_abs("6345.0", tables = "1", release_date = "2019-09-01")
}

# Or tables 1 and 2a from the WPI
\dontrun{
wpi_t1_t2a <- read_abs("6345.0", tables = c("1", "2a"))
}


# Get two specific time series, based on their time series IDs
\dontrun{
cpi <- read_abs(series_id = c("A2325806K", "A2325807L"))
}

# Get series IDs using the `read_abs_series()` wrapper function
\dontrun{
cpi <- read_abs_series(c("A2325806K", "A2325807L"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_cube.r
\name{get_available_lfs_cubes}
\alias{get_available_lfs_cubes}
\title{Show the available Labour Force, Australia, detailed data cubes that can be
downloaded}
\usage{
get_available_lfs_cubes()
}
\description{
Show the available Labour Force, Australia, detailed data cubes that can be
downloaded
}
\details{
Intended to be used with \code{read_lfs_datacube()}. Call
\code{read_lfs_datacube()} interactively, find the table of interest
(eg. "LM1"), then use \code{read_lfs_datacube()}.
}
\examples{

get_available_lfs_cubes()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_awe.R
\name{tidy_awe}
\alias{tidy_awe}
\title{Internal function to tidy a dataframe from ABS 6302}
\usage{
tidy_awe(df)
}
\arguments{
\item{df}{Data frame containing table 2 from ABS 6302, imported using \code{read_abs()}}
}
\description{
Internal function to tidy a dataframe from ABS 6302
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_latest_date.R
\name{check_latest_date}
\alias{check_latest_date}
\title{Get date of most recent observation(s) in ABS time series}
\usage{
check_latest_date(cat_no = NULL, tables = "all", series_id = NULL)
}
\arguments{
\item{cat_no}{ABS catalogue number, as a string, including the extension.
For example, "6202.0".}

\item{tables}{numeric. Time series tables in \verb{cat_no`` to download and extract. Default is "all", which will read all time series in }cat_no\verb{. Specify }tables\verb{to download and import specific tables(s) - eg.}tables = 1\code{or}tables = c(1, 5)`.}

\item{series_id}{(optional) character. Supply an ABS unique time series
identifier (such as "A2325807L") to get only that series.
This is an alternative to specifying \code{cat_no}.}
}
\value{
Date vector of length one. Date corresponds to the most recent
observation date for any of the time series in the table(s) requested.
observation date for any of the time series in the table(s) requested.
}
\description{
This function returns the most recent observation date for a specified ABS
time series catalogue number (as a whole), individual tables, or series IDs.
}
\details{
Where the individual time series in your request
have multiple dates, only the most recent will be returned.
}
\examples{
\dontrun{

# Check a whole catalogue number; return the latest release date for any
# time series in the number

check_latest_date("6345.0")

# Return latest release date for a table within a catalogue number  - note
# the function will return the release date
# of the most-recently-updated series within the tables
check_latest_date("6345.0", tables = 1)

# Or for multiple tables - note the function will return the release date
# of the most-recently-updated series within the tables
check_latest_date("6345.0", tables = c("1", "5a"))

# Or for an individual time series
check_latest_date(series_id = "A2713849C")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/separate_series.R
\name{separate_series}
\alias{separate_series}
\title{Separate the series column in a tidy ABS time series data frame}
\usage{
separate_series(
  data,
  column_names = NULL,
  remove_totals = FALSE,
  remove_nas = FALSE
)
}
\arguments{
\item{data}{A data frame (tibble) containing tidied data from the ABS time
series table(s).}

\item{column_names}{(optional) character vector. Supply a vector of column
names, such as \code{c("group_name", "variable","gender")}. If not
supplied, columns will be named "series_1" etc.}

\item{remove_totals}{logical. FALSE by default.
If set to TRUE, any series rows that contain the word "total"
will be removed.}

\item{remove_nas}{locical. FALSE by default. If set to TRUE, any rows
containining an NA in at least one of the separated series columns
will be removed.}
}
\value{
A data frame (tibble) containing the tidied data from the ABS time
series table(s).
}
\description{
Separate the 'series' column in a data frame (tibble)
downloaded using \code{read_abs()} into multiple columns using the ";"
separator.
}
\examples{
\dontrun{
wpi <- read_abs("6345.0", 1) \%>\%
  separate_series()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_payrolls.R
\name{download_previous_payrolls}
\alias{download_previous_payrolls}
\title{This function is temporarily necessary while the readabs maintainer
attempts to resolve an issue with the ABS. The ABS as at late March 2021
stopped including Table 5 of the Weekly Payrolls release with each new
release of the data. This function finds the link from the previous
release and attemps to download it. This function will no longer be required
if/when the ABS reverts to the previous release arrangements. The function
is internal and is called by \code{read_payrolls()}.}
\usage{
download_previous_payrolls(cube_name, path)
}
\arguments{
\item{cube_name}{eg. DO004 for table 4}

\item{path}{Directory in which to download payrolls cube}
}
\value{
A list containing two elements: \code{result} (will contain path + filename
to downloaded file if download was successful); and \code{error} (NULL if file
downloaded successfully; character otherwise).
}
\description{
This function is temporarily necessary while the readabs maintainer
attempts to resolve an issue with the ABS. The ABS as at late March 2021
stopped including Table 5 of the Weekly Payrolls release with each new
release of the data. This function finds the link from the previous
release and attemps to download it. This function will no longer be required
if/when the ABS reverts to the previous release arrangements. The function
is internal and is called by \code{read_payrolls()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_awe.R
\name{read_awe}
\alias{read_awe}
\title{read_awe}
\usage{
read_awe(
  wage_measure = c("awote", "ftawe", "awe"),
  sex = c("persons", "males", "females"),
  sector = c("total", "private", "public"),
  state = c("all", "nsw", "vic", "qld", "sa", "wa", "tas", "nt", "act"),
  na.rm = FALSE,
  path = Sys.getenv("R_READABS_PATH", unset = tempdir()),
  show_progress_bars = FALSE,
  check_local = FALSE
)
}
\arguments{
\item{wage_measure}{Character of length 1. Must be one of:
\describe{
\item{\code{awote}}{ Average weekly ordinary time earnings; also known as Full-time adult ordinary time earnings}
\item{\code{ftawe}}{ Full-time adult total earnings}
\item{\code{awe}}{ Average weekly total earnings of all employees}
}}

\item{sex}{Character of length 1. Must be one of: \code{persons}, \code{males}, or \code{females}.}

\item{sector}{Character of length 1. Must be one of: \code{total}, \code{private}, or
\code{public}. Note that you cannot get sector-by-state data; if \code{state} is not
\code{all} then \code{sector} must be \code{total}.}

\item{state}{Character of length 1. Must be one of: \code{all}, \code{nsw}, \code{vic}, \code{qld},
\code{sa}, \code{wa}, \code{nt}, or \code{act}. Note that you cannot get sector-by-state data;
if \code{sector} is not \code{total} then \code{state} must be \code{all}.}

\item{na.rm}{Logical. \code{FALSE} by default. If \code{FALSE}, a consistent quarterly
series is returned, with \code{NA} values for quarters in which there is no data.
If \code{TRUE}, only dates with data are included in the returned data frame.}

\item{path}{See \code{?read_abs}}

\item{show_progress_bars}{See \code{?read_abs}}

\item{check_local}{See \code{?read_abs}}
}
\value{
A \code{tbl_df} with four columns: \code{date}, \code{sex}, \code{wage_measure} and \code{value}.
The data is nominal and seasonally adjusted.
}
\description{
Convenience function to obtain wage levels from ABS
6302.0, Average Weekly Earnings, Australia.
}
\details{
The latest AWE data is available using \code{read_abs(cat_no = "6302.0", tables = 2)}.
However, this time series only goes back to 2012, when the ABS switched
from quarterly to biannual collection and release of the AWE data. The
\code{read_awe()} function assembles on time series back to November 1983 quarter;
it is quarterly to 2012 and biannual from then. Note that the data
returned with this function is consistently quarterly; any quarters for
which there are no observations are recorded as \code{NA} unless \code{na.rm} = \code{TRUE}.
}
\examples{
\dontrun{
read_awe("awote", "persons")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_abs_data.R
\name{read_abs_data}
\alias{read_abs_data}
\title{Extracts ABS time series data from local Excel spreadsheets and converts to
long format.}
\usage{
read_abs_data(path, sheet)
}
\arguments{
\item{path}{Filepath to Excel spreadsheet.}

\item{sheet}{Sheet name or number.}
}
\value{
Long-format dataframe
}
\description{
\code{read_abs_data()} is soft deprecated and will be removed in a future version.
Please use \code{read_abs_local()} to import and tidy locally-stored
ABS time series spreadsheets, or \code{read_abs()} to download, import,
and tidy time series spreadsheets from the ABS website.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/show_available_files.r
\name{show_available_files}
\alias{show_available_files}
\alias{get_available_files}
\title{Helper function to show the files available in a particular catalogue number.}
\usage{
show_available_files(catalogue_string, refresh = FALSE)

get_available_files(catalogue_string, refresh = FALSE)
}
\arguments{
\item{catalogue_string}{character string specifying the catalogue,
e.g. "labour-force-australia-detailed".
You can use \code{show_available_catalogues()} see all the possible catalogues,
or \code{search_catalogues()} to find catalogues that contain a given string.}

\item{refresh}{logical; \code{FALSE} by default. If \code{FALSE}, an internal table
of the available ABS catalogues is used. If \code{TRUE}, this table is refreshed
from the ABS website.}
}
\value{
A tibble containing the title of the file, the filename and the complete url.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
To be used in conjunction with \code{download_abs_data_cube()}.

This function lists the possible files that are available in a catalogue.
The filename (or an unambiguous part of the filename) must be specified
as a string as an argument to \code{download_abs_data_cube}.
}
\details{
\code{get_available_files()} is an alias for \code{show_available_files()}.
}
\examples{
\dontrun{
show_available_files("labour-force-australia-detailed")
}

}
\seealso{
Other data cube functions:
\code{\link[=download_abs_data_cube]{download_abs_data_cube()}},
\code{\link[=search_catalogues]{search_catalogues()}},
\code{\link[=show_available_catalogues]{show_available_catalogues()}}
}
\concept{data cube functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{fast_str_squish}
\alias{fast_str_squish}
\title{Very slightly faster version of stringr's str_squish()}
\usage{
fast_str_squish(string)
}
\arguments{
\item{string}{A string to squish (remove whitespace)}
}
\description{
Very slightly faster version of stringr's str_squish()
}
\keyword{internal}
